unit Filters;

(*

  Source : https://github.com/ddiakopoulos/MoogLadders

  --

  Copyright (c) 2015, Dimitri Diakopoulos All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  -- --

  Huovilainen developed an improved and physically correct model of the Moog
  Ladder filter that builds upon the work done by Smith and Stilson. This model
  inserts nonlinearities inside each of the 4 one-pole sections on account of the
  smoothly saturating function of analog transistors. The base-emitter voltages of
  the transistors are considered with an experimental value of 1.22070313 which
  maintains the characteristic sound of the analog Moog. This model also permits
  self-oscillation for resonances greater than 1. The model depends on five
  hyperbolic tangent functions (tanh) for each sample, and an oversampling factor
  of two (preferably higher, if possible). Although a more faithful
  representation of the Moog ladder, these dependencies increase the processing
  time of the filter significantly. Lastly, a half-sample delay is introduced for
  phase compensation at the final stage of the filter.

  References: Huovilainen (2004), Huovilainen (2010), DAFX - Zolzer (ed) (2nd ed)
  Original implementation: Victor Lazzarini for CSound5

  Considerations for oversampling:
  http://music.columbia.edu/pipermail/music-dsp/2005-February/062778.html
  http://www.synthmaker.co.uk/dokuwiki/doku.php?id=tutorials:oversampling

  --

  2020-04-30 : Adapted for use in Wren by Blue Hell, also made it polyphonic.

  available models :

    +-----------+--------------+--------------------+-------------------------+------------------------+
    |Implemented| Name         | License            | Original souerce        | Cloded-source friendly |
    +-----------+--------------+--------------------+-------------------------+------------------------+
    | N         | Simplified   | Custom             | DAFX                    |  No                    |
    | Y         | Huovilainen  | LGPLv3             | CSound                  |  If dynamically linked |
    |           | Stilson      | Unlicense          | Moog~ by D. Lowenfels   |  Yes                   |
    | Y         | Microtracker | Unlicense          | Magnus Jonsson          |  Yes                   |
    | Y         | Krajeski     | Unlicense          | Aaron Krajeski          |  Yes                   |
    | Y         | MusicDSP     | Suggested CC-BY-SA | MusicDSP.org            |  Yes                   |
    | Y         | Oberheim     | Custom             | Will Pirkle             |  Yes                   |
    | Y         | Improved     | ISC                | Via Author              |  Yes                   |
    | Y         | RKSimulation | BSD                | Bob~ by Miller Puckette |  Yes                   |
    +-----------+--------------+--------------------+-------------------------+------------------------+
*)


interface

uses

  System.Math, System.SysUtils,

  KnobsUtils, KnobsConversions;


type

  TFilterType = (
    FT_LOWPASS   = 0,
    FT_HIGHPASS     ,
    FT_BANDPASS     ,
    FT_ALLPASS      ,
    FT_NOTCH        ,
    FT_PEAK         ,
    FT_LOW_SHELF    ,
    FT_HIGH_SHELF
  );


  TMoogFilterType = (
    MFT_HuovilainenMoog  = 0 ,   // The current default moog filter is Huovilainen
    MFT_ImprovedMoog         ,
    MFT_KrajeskiMoog         ,
    MFT_MicroTrackerMoog     ,
    MFT_MusicDSPMoog         ,
    MFT_OberheimVariationMoog,
    MFT_RKSimulationMoog     ,
    MFT_StilsonMoog
  );


  TCoefs2 = array[ 0 .. 1] of TSignal;
  TCoefs3 = array[ 0 .. 2] of TSignal;
  TCoefs4 = array[ 0 .. 3] of TSignal;
  TCoefs5 = array[ 0 .. 4] of TSignal;


  TSimpleFilter = class
  private
    FPolyphony  : Integer;
    FSampleRate : TSignal;
  protected
    procedure   Initialize;                                                                                     virtual;
  public
    constructor Create( aPolyphony: Integer; aSampleRate: TSignal);
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                              virtual;
    procedure   Process( InSample: TSignalArray; var OutSample: TSignalArray);                                  virtual;
  public
    property    Polyphony  : Integer read FPolyphony;
    property    SampleRate : TSignal read FSampleRate;
  end;


  TBiQuadBase = class( TSimpleFilter)
  private type
    TState = record
      BCoef : TCoefs3; // b0 b1 b2
      ACoef : TCoefs2; // a0 a1
      W     : TCoefs2; // Delays
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   Initialize;                                                                                    override;
    procedure   SetBiQuadCoefs( aVoice: Integer; const b: TCoefs3; a: TCoefs2);
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
    function    MagnitudeW  ( anOmega   : TSignal): TSignal;
    function    MagnitudeF  ( aFrequency: TSignal): TSignal;
    function    MagnitudedbW( anOmega   : TSignal): TSignal;
    function    MagnitudedBF( aFrequency: TSignal): TSignal;
  end;


  TRBJFilter = class( TBiQuadBase)
  private type
    TState = record
      Omega      : TSignal;
      CosOmega   : TSignal;
      SinOmega   : TSignal;
      Q          : TSignal;
      Alpha      : TSignal;
      AA         : TSignal;
      A          : TCoefs3;
      B          : TCoefs3;
      T          : TFilterType;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  private
    procedure   SetCutOff    ( aVoice: Integer; const aValue: TSignal);
    procedure   SetQValue    ( aVoice: Integer; const aValue: TSignal);
    procedure   SetGain      ( aVoice: Integer; const aValue: TSignal);
    procedure   SetFilterType( aVoice: Integer; aValue: TFilterType);
  protected
    procedure   Initialize;                                                                                    override;
  public
    constructor Create( aPolyphony: Integer; aSampleRate: TSignal; aFilterType: TFilterType);
    procedure   UpdateCoefficients( aVoice: Integer);
  public
    property    CutOff    [ aVoice: Integer] : TSignal     write SetCutOff;
    property    QValue    [ aVoice: Integer] : TSignal     write SetQValue;
    property    Gain      [ aVoice: Integer] : TSignal     write SetGain;
    property    FilterType[ aVoice: Integer] : TFilterType write SetFilterType;
  end;


  TPinkingFilter = class( TSimpleFilter)
  // http://www.musicdsp.org/files/pink.txt
  // +/-0.05dB above 9.2Hz @ 44,100Hz
  private type
    TState = record
      B0 : TSignal;
      B1 : TSignal;
      B2 : TSignal;
      B3 : TSignal;
      B4 : TSignal;
      B5 : TSignal;
      B6 : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   Initialize;                                                                                    override;
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  TBrowningFilter = class( TSimpleFilter)
  private type
    TState = record
      L : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   Initialize;                                                                                    override;
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  TLadderFilterBase = class( TSimpleFilter)
  private type
    TState = record
      CutOff    : TSignal;
      Resonance : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    function    GetCutOff   ( aVoice: Integer): TSignal;                                                        virtual;
    procedure   SetCutOff   ( aVoice: Integer; const aValue: TSignal);                                          virtual;
    function    GetResonance( aVoice: Integer): TSignal;                                                        virtual;
    procedure   SetResonance( aVoice: Integer; const aValue: TSignal);                                          virtual;
  protected
    procedure   Initialize;                                                                                    override;
  public
    property    CutOff   [ aVoice: Integer] : TSignal read GetCutOff    write SetCutOff;
    property    Resonance[ aVoice: Integer] : TSignal read GetResonance write SetResonance;
  end;


  TLadderFilterClass = class of TLadderFilterBase;


  THuovilainenMoog = class( TLadderFilterBase)
  //
  //  Huovilainen developed an improved and physically correct model of the Moog
  //  Ladder filter that builds upon the work done by Smith and Stilson. This model
  //  inserts nonlinearities inside each of the 4 one-pole sections on account of the
  //  smoothly saturating function of analog transistors. The base-emitter voltages of
  //  the transistors are considered with an experimental value of 1.22070313 which
  //  maintains the characteristic sound of the analog Moog. This model also permits
  //  self-oscillation for resonances greater than 1. The model depends on five
  //  hyperbolic tangent functions (tanh) for each sample, and an oversampling factor
  //  of two (preferably higher, if possible). Although a more faithful
  //  representation of the Moog ladder, these dependencies increase the processing
  //  time of the filter significantly. Lastly, a half-sample delay is introduced for
  //  phase compensation at the final stage of the filter.
  //  References: Huovilainen (2004), Huovilainen (2010), DAFX - Zolzer (ed) (2nd ed)
  //  Original implementation: Victor Lazzarini for CSound5
  //  Considerations for oversampling:
  //  http://music.columbia.edu/pipermail/music-dsp/2005-February/062778.html
  //  http://www.synthmaker.co.uk/dokuwiki/doku.php?id=tutorials:oversampling  strict private [BH: dead link]
  //
  const
    Thermal = 0.000025;
    Comp    = 0.5;
  private
    type TState = record
      Stage0     : TSignal;
      Stage1     : TSignal;
      Stage2     : TSignal;
      Stage3     : TSignal;
      StageTanh0 : TSignal;
      StageTanh1 : TSignal;
      StageTanh2 : TSignal;
      Delay0     : TSignal;
      Delay1     : TSignal;
      Delay2     : TSignal;
      Delay3     : TSignal;
      Delay4     : TSignal;
      Delay5     : TSignal;
      Tune       : TSignal;
      Acr        : TSignal;
      ResQuad    : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   SetResonance( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   SetCutOff   ( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   Initialize;                                                                                    override;
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  TImprovedMoog = class( TLadderFilterBase)
  //
  //  This model is based on a reference implementation of an algorithm developed by
  //  Stefano D'Angelo and Vesa Valimaki, presented in a paper published at ICASSP in 2013.
  //  This improved model is based on a circuit analysis and compared against a reference
  //  Ngspice simulation. In the paper, it is noted that this particular model is
  //  more accurate in preserving the self-oscillating nature of the real filter.
  //  References: "An Improved Virtual Analog Model of the Moog Ladder Filter"
  //  Original Implementation: D'Angelo, Valimaki
  //
  strict private
  const
    VT = 0.312;
  private
    type TState = record
      V     : TCoefs4;
      DV    : TCoefs4;
      TV    : TCoefs4;
      X     : TSignal;
      G     : TSignal;
      Drive : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   SetCutOff( aVoice: Integer; const aValue: TSignal);                                            override;
    procedure   Initialize;                                                                                    override;
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  TKrajeskiMoog = class( TLadderFilterBase)
  //
  //  This class implements Tim Stilson's MoogVCF filter
  //  using 'compromise' poles at z = -0.3
  //  Several improments are built in, such as corrections
  //  for cutoff and resonance parameters, removal of the
  //  necessity of the separation table, audio rate update
  //  of cutoff and resonance and a smoothly saturating
  //  tanh() function, clamping output and creating inherent
  //  nonlinearities.
  //  This code is Unlicensed (i.e. public domain); in an email exchange on
  //  4.21.2018 Aaron Krajeski stated: "That work is under no copyright.
  //  You may use it however you might like."
  //  Source: http://song-swap.com/MUMT618/aaron/Presentation/demo.html
  //
  strict private
  private
    type TState = record
      State : TCoefs5;
      Delay : TCoefs5;
      WC    : TSignal;
      G     : TSignal;
      GRes  : TSignal;
      GComp : TSignal;
      Drive : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   SetCutOff   ( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   SetResonance( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   Initialize;                                                                                    override;
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  TMicroTrackerMoog = class( TLadderFilterBase)
  //
  //  Based on an implementation by Magnus Jonsson
  //  https://github.com/magnusjonsson/microtracker (unlicense)
  //
  strict private
  private
    type TState = record
      P0   : TSignal;
      P1   : TSignal;
      P2   : TSignal;
      P3   : TSignal;
      P32  : TSignal;
      P33  : TSignal;
      P34  : TSignal;
      GCut : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   SetCutOff( aVoice: Integer; const aValue: TSignal);                                            override;
    procedure   Initialize;                                                                                    override;
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  TMusicDSPMoog = class( TLadderFilterBase)
  //
  //  This file is unlicensed and uncopyright as found at:
  //  http://www.musicdsp.org/showone.php?id=24
  //  Considering how widely this same code has been used in ~100 projects on GitHub with
  //  various licenses, it might be reasonable to suggest that the license is CC-BY-SA
  //
  strict private
  private
    type TState = record
      GReso   : TSignal;
      GCutOff : TSignal;
      Stage   : TCoefs4;
      Delay   : TCoefs4;
      P       : TSignal;
      K       : TSignal;
      T1      : TSignal;
      T2      : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   SetCutOff   ( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   SetResonance( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   Initialize;                                                                                    override;
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  TVAOnePole = class( TSimpleFilter)
  //
  //  One pole for the OberheimVariationMoog filter
  //
  strict private
  private
    type TState = record
      Alpha    : TSignal;
      Beta     : TSignal;
      Gamma    : TSignal;
      Delta    : TSignal;
      Epsilon  : TSignal;
      A0       : TSignal;
      Feedback : TSignal;
      Z1       : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   SetFeedback( aVoice: Integer; const aValue: TSignal);
    function    GetFeedbackOutput( aVoice: Integer): TSignal;
    procedure   SetAlpha( aVoice: Integer; const aValue: TSignal);
    procedure   SetBeta ( aVoice: Integer; const aValue: TSignal);
    procedure   Initialize;                                                                                    override;
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  TOberheimVariationMoog = class( TLadderFilterBase)
  //
  //  See: http://www.willpirkle.com/forum/licensing-and-book-code/licensing-and-using-book-code/
  //  The license is "You may also use the code from the FX and Synth books without licensing or fees.
  //  The code is for you to develop your own plugins for your own use or for commercial use."
  //
  strict private
  private
    type TState = record
      LPF1          : TVAOnePole;
      LPF2          : TVAOnePole;
      LPF3          : TVAOnePole;
      LPF4          : TVAOnePole;
      K             : TSignal;
      Gamma         : TSignal;
      Alpha0        : TSignal;
      Saturation    : TSignal;
      OberheimCoefs : TCoefs5;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   SetCutOff   ( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   SetResonance( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   Initialize;                                                                                    override;
  public
    destructor  Destroy;                                                                                       override;
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  TRKSimulationMoog = class( TLadderFilterBase)
  //
  //  Copyright (c) 2015, Miller Puckette. All rights reserved.
  //
  //  Redistribution and use in source and binary forms, with or without
  //  modification, are permitted provided that the following conditions are met:
  //
  //  * Redistributions of source code must retain the above copyright notice, this
  //    list of conditions and the following disclaimer.
  //
  //  * Redistributions in binary form must reproduce the above copyright notice,
  //    this list of conditions and the following disclaimer in the documentation
  //    and/or other materials provided with the distribution.
  //
  //  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  //  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  //  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  //  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  //  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  //  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  //  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  //  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  //  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  //  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  //
  // ===================================================================================================================
  //
  //  Imitates a Moog resonant filter by Runge-Kutte numerical integration of
  //  a differential equation approximately describing the dynamics of the circuit.
  //
  //  Useful references:
  //
  //	* Tim Stilson
  //	"Analyzing the Moog VCF with Considerations for Digital Implementation"
  //		Sections 1 and 2 are a reasonably good introduction but the
  //		model they use is highly idealized.
  //
  //	* Timothy E. Stinchcombe
  //	"Analysis of the Moog Transistor Ladder and Derivative Filters"
  //		Long, but a very thorough description of how the filter works including
  //		its nonlinearities
  //
  //	* Antti Huovilainen
  //	"Non-linear digital implementation of the moog ladder filter"
  //		Comes close to giving a differential equation for a reasonably realistic
  //		model of the filter
  //
  //  The differential equations are:
  //
  //	y1' = k * (S(x - r * y4) - S(y1))
  //	y2' = k * (S(y1) - S(y2))
  //	y3' = k * (S(y2) - S(y3))
  //	y4' = k * (S(y3) - S(y4))
  //
  //  where k controls the cutoff frequency, r is feedback (<= 4 for stability), and S(x) is a saturation function.
  //
  strict private
  private
    type TState = record
      State         : TCoefs4;
      Saturation    : TSignal;
      SaturationInv : TSignal;
      Reso          : TSignal;
      F             : TSignal;
    end;
    PState = ^ TState;
  private
    FOverSampleFactor : Integer;
    FStepSize         : TSignal;
    FState            : TArray<TState>;
  private
    procedure   CalculateDerivatives( aVoice: Integer; aSample: TSignal; var DState: TCoefs4; const State: TCoefs4);
    procedure   RungeKuttaSolver    ( aVoice: Integer; aSample: TSignal; var State: TCoefs4);
  protected
    procedure   SetCutOff   ( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   SetResonance( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   Initialize;                                                                                    override;
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  TStilsonMoog = class( TLadderFilterBase)
  //
  //  Copyright (c) 2003, David Lowenfels
  //  Released as the moog~ pd extern.
  //
  //  This code is Unlicensed (i.e. public domain); in an email exchange on
  //  6.14.2019 David Lowenfels stated "You're welcome to use the Moog~ code,
  //  license it free as in beer or whatever :)"
  //
  //  A digital model of the classic Moog filter was presented first by Stilson and
  //  Smith. This model uses a cascade of one-pole IIR filters in series with a global
  //  feedback to produce resonance. A digital realization of this filter introduces a
  //  unit delay, effectively making it a fifth-order filter. Unfortunately, this
  //  delay also has the effect of coupling the cutoff and resonance parameters,
  //  uncharacteristic of the uncoupled control of the original Moog ladder. As a
  //  compromise, a zero can be inserted at z = -0.3 inside each one pole section to
  //  minimize the coupling the parameters (humans are not particularly sensitive to
  //  variations in Q factor). Although fast coefficient updates can be achieved since
  //  the nonlinearities of the Moog are not considered, the filter becomes unstable
  //  with very large resonance values and does not enter self-oscillation.
  //  References: Stilson and Smith (1996), DAFX - Zolzer (ed) (2nd ed)
  //  Original implementation: Tim Stilson, David Lowenfels
  //
  strict private const
    S_STILSON_GAINTABLE : array[ 0 .. 198] of TSignal = (
      0.999969, 0.990082, 0.980347, 0.970764, 0.961304, 0.951996, 0.94281 , 0.933777, 0.924866, 0.916077,
      0.90741 , 0.898865, 0.890442, 0.882141, 0.873962, 0.865906, 0.857941, 0.850067, 0.842346, 0.834686,
      0.827148, 0.819733, 0.812378, 0.805145, 0.798004, 0.790955, 0.783997, 0.77713 , 0.770355, 0.763672,
      0.75708 , 0.75058 , 0.744141, 0.737793, 0.731537, 0.725342, 0.719238, 0.713196, 0.707245, 0.701355,
      0.695557, 0.689819, 0.684174, 0.678558, 0.673035, 0.667572, 0.66217 , 0.65686 , 0.651581, 0.646393,
      0.641235, 0.636169, 0.631134, 0.62619 , 0.621277, 0.616425, 0.611633, 0.606903, 0.602234, 0.597626,
      0.593048, 0.588531, 0.584045, 0.579651, 0.575287, 0.570953, 0.566681, 0.562469, 0.558289, 0.554169,
      0.550079, 0.546051, 0.542053, 0.538116, 0.53421 , 0.530334, 0.52652 , 0.522736, 0.518982, 0.515289,
      0.511627, 0.507996, 0.504425, 0.500885, 0.497375, 0.493896, 0.490448, 0.487061, 0.483704, 0.480377,
      0.477081, 0.473816, 0.470581, 0.467377, 0.464203, 0.46109 , 0.457977, 0.454926, 0.451874, 0.448883,
      0.445892, 0.442932, 0.440033, 0.437134, 0.434265, 0.431427, 0.428619, 0.425842, 0.423096, 0.42038 ,
      0.417664, 0.415009, 0.412354, 0.409729, 0.407135, 0.404572, 0.402008, 0.399506, 0.397003, 0.394501,
      0.392059, 0.389618, 0.387207, 0.384827, 0.382477, 0.380127, 0.377808, 0.375488, 0.37323 , 0.370972,
      0.368713, 0.366516, 0.364319, 0.362122, 0.359985, 0.357849, 0.355713, 0.353607, 0.351532, 0.349457,
      0.347412, 0.345398, 0.343384, 0.34137 , 0.339417, 0.337463, 0.33551 , 0.333588, 0.331665, 0.329773,
      0.327911, 0.32605 , 0.324188, 0.322357, 0.320557, 0.318756, 0.316986, 0.315216, 0.313446, 0.311707,
      0.309998, 0.308289, 0.30658 , 0.304901, 0.303223, 0.301575, 0.299927, 0.298309, 0.296692, 0.295074,
      0.293488, 0.291931, 0.290375, 0.288818, 0.287262, 0.285736, 0.284241, 0.282715, 0.28125 , 0.279755,
      0.27829 , 0.276825, 0.275391, 0.273956, 0.272552, 0.271118, 0.269745, 0.268341, 0.266968, 0.265594,
      0.264252, 0.262909, 0.261566, 0.260223, 0.258911, 0.257599, 0.256317, 0.255035, 0.25375
    );
  strict private
  private
    type TState = record
      State  : TCoefs4;
      Reso   : TSignal;
      Output : TSignal;
      P      : TSignal;
      Q      : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
    procedure   SetCutOff   ( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   SetResonance( aVoice: Integer; const aValue: TSignal);                                         override;
    procedure   Initialize;                                                                                    override;
  public
    function    Tick( aVoice: Integer; aSample: TSignal): TSignal;                                             override;
  end;


  function  LadderTypeToClass( aType: TMoogFilterType): TLadderFilterClass;


implementation


const

  Init2 : TCoefs2 = ( 0.0, 0.0);
  Init3 : TCoefs3 = ( 0.0, 0.0, 0.0);
  Init4 : TCoefs4 = ( 0.0, 0.0, 0.0, 0.0);
  Init5 : TCoefs5 = ( 0.0, 0.0, 0.0, 0.0, 0.0);


  function  FastTanH( X: TSignal): TSignal; inline;
  var
    X2 : TSignal;
  begin
    X2     := Normalize( X * X);
    Result := X * ( 27.0 + X2) / ( 27.0 + 9.0 * X2);
  end;


  function  SoftClip( aValue, aSaturation, aSaturationInverse: TSignal): TSignal;
  //
  // Imitate the (tanh) clipping function of a transistor pair.
  // to 4th order, tanh is x - x*x*x/3; this cubic's
  // plateaus are at +/- 1 so clip to 1 and evaluate the cubic.
  // This is pretty coarse - for instance if you clip a sinusoid this way you
  // can sometimes hear the discontinuity in 4th derivative at the clip point
  //
  var
    v1 : TSignal;
    v2 : TSignal;
  begin
    v1 := Normalize( aValue) * aSaturationinverse;

    if   v1 > 1.0
    then v2 := 1.0
    else if v1 < -1.0
    then v2 := -1.0
    else v2 := v1;

    Result := aSaturation * ( v2 - ( 1.0 / 3.0) * v2 * v2 * v2);
  end;


  function  MoogSaturate( aValue: TSignal): TSignal;
  //
  // Clamp without branching
  // If input - _limit < 0, then it really substracts, and the 0.5 to make it half the 2 inputs.
  // If > 0 then they just cancel, and keeps input normal.
  // The easiest way to understand it is check what happends on both cases.
  //
  var
    X1 : TSignal;
    X2 : TSignal;
  begin
    X1     := Abs( aValue + 0.95);
    X2     := Abs( aValue - 0.95);
    Result := 0.5 * ( X1 - X2);
  end;


  function  MoogLerp( anAmount, a, b: TSignal): TSignal;
  begin
    Result := (1.0 - anAmount) * a + anAmount * b;
  end;


  function  LadderTypeToClass( aType: TMoogFilterType): TLadderFilterClass;
  begin
    case aType of
      MFT_HuovilainenMoog       : Result := THuovilainenMoog      ;
      MFT_ImprovedMoog          : Result := TImprovedMoog         ;
      MFT_KrajeskiMoog          : Result := TKrajeskiMoog         ;
      MFT_MicroTrackerMoog      : Result := TMicroTrackerMoog     ;
      MFT_MusicDSPMoog          : Result := TMusicDSPMoog         ;
      MFT_OberheimVariationMoog : Result := TOberheimVariationMoog;
      MFT_RKSimulationMoog      : Result := TRKSimulationMoog     ;
      MFT_StilsonMoog           : Result := TStilsonMoog          ;
      else                        Result := nil;
    end;
  end;


{ ========
  TSimpleFilter = class
  private
    FPolyphony  : Integer;
    FSampleRate : TSignal;
  public
    property    Polyphony  : Integer read FPolyphony;
    property    SampleRate : TSignal read FSampleRate;
  protected
}

    procedure   TSimpleFilter.Initialize; // virtual;
    begin
      // semi abstract ...
    end;


//  public

    constructor TSimpleFilter.Create( aPolyphony: Integer; aSampleRate: TSignal);
    begin
      inherited Create;
      FPolyphony  := aPolyphony;
      FSampleRate := aSampleRate;
      Initialize;
    end;


    function    TSimpleFilter.Tick( aVoice: Integer; aSample: TSignal): TSignal; // virtual;
    begin
      Result := aSample;
    end;


    procedure   TSimpleFilter.Process( InSample: TSignalArray; var OutSample: TSignalArray); // virtual;
    var
      aVoice : Integer;
    begin
      for aVoice := 0 to Polyphony - 1
      do OutSample[ aVoice] := Tick( aVoice, InSample[ aVoice]);
    end;


{ ========
  TBiQuadBase = class
  private type
    TState = record
      BCoef : TCoefs3; // b0 b1 b2
      ACoef : TCoefs2; // a0 a1
      W     : TCoefs2; // Delays
    end;
  private
    FState : TArray<TState>;
  protected
}

    procedure   TBiQuadBase.Initialize; // override;
    var
      State  : PState;
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        State        := @ FState[ aVoice];
        State^.BCoef := Init3;
        State^.ACoef := Init2;
        State^.W     := Init2;
      end;
    end;


    procedure   TBiQuadBase.SetBiQuadCoefs( aVoice: Integer; const b: TCoefs3; a: TCoefs2);
    var
      State : PState;
    begin
      State        := @ FState[ aVoice];
      State^.BCoef := b;
      State^.ACoef := a;
    end;


//  public

    function    TBiQuadBase.Tick( aVoice: Integer; aSample: TSignal): TSignal;
    var
      State : PState;
    begin
      if   aVoice < Polyphony
      then begin
        State        := @ FState[ aVoice];
        Result       := State^.BCoef[ 0] * aSample + State^.W[ 0];
        State^.W[ 0] := State^.BCoef[ 1] * aSample - State^.ACoef[ 0] * Result + State^.W[ 1];
        State^.W[ 1] := State^.BCoef[ 2] * aSample - State^.ACoef[ 1] * Result;
      end
      else Result := 1.0;
    end;


    function    TBiQuadBase.MagnitudeW( anOmega: TSignal): TSignal;
    var
      State : PState;
      W     : TSignal;
      CosW  : TSignal;
      Cos2W : TSignal;
      B0    : TSignal;
      B1    : TSignal;
      B2    : TSignal;
      A1    : TSignal;
      A2    : TSignal;
    begin
      if   Polyphony > 0
      then begin
        State  := @ FState[ 0];
        W      := anOmega / SampleRate;    // Normalize Omega
        CosW   := Cos( W      );
        Cos2W  := Cos( 2.0 * W);
        B0     := State^.BCoef[ 0];
        B1     := State^.BCoef[ 1];
        B2     := State^.BCoef[ 2];
        A1     := State^.ACoef[ 0];
        A2     := State^.ACoef[ 1];
        Result :=
          Sqrt(
            (
                B0 * B0
              + B1 * B1
              + B2 * B2
              + 2.0 * CosW  * ( B0 * B1 + B1 * B2)
              + 2.0 * Cos2W * B0 * B2
            )
            /
            (
                1.0
              + A1 * A1
              + A2 * A2
              + 2.0 * CosW  * ( A1 + A1 * A2)
              + 2.0 * Cos2W * A2
            )
          );
      end
      else Result := 1.0;
    end;


    function    TBiQuadBase.MagnitudeF( aFrequency: TSignal): TSignal;
    begin
      Result := MagnitudeW( 2.0 * PI * aFrequency);
    end;


    function    TBiQuadBase.MagnitudedBW( anOmega: TSignal): TSignal;
    begin
      Result := UnitsTodB( MagnitudeW( anOmega));
    end;


    function    TBiQuadBase.MagnitudedBF( aFrequency: TSignal): TSignal;
    begin
      Result := UnitsTodB( MagnitudeF( aFrequency));
    end;


{ ========
  TRBJFilter = class( TBiQuadBase)
  private type
    TState = record
      Omega      : TSignal;
      CosOmega   : TSignal;
      SinOmega   : TSignal;
      Q          : TSignal;
      Alpha      : TSignal;
      AA         : TSignal;
      A          : TCoefs3;
      B          : TCoefs3;
      T          : TFilterType;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  public
    property    CutOff    [ aVoice: Integer] : TSignal     write SetCutOff;
    property    QValue    [ aVoice: Integer] : TSignal     write SetQValue;
    property    Gain      [ aVoice: Integer] : TSignal     write SetGain;
    property    FilterType[ aVoice: Integer] : TFilterType write SetFilterType;
  private
}

    procedure   TRBJFilter.SetCutOff( aVoice: Integer; const aValue: TSignal);
    begin
      FState[ aVoice].Omega := TWO_PI * aValue / FSampleRate;
      UpdateCoefficients( aVoice);
    end;


    procedure   TRBJFilter.SetQValue( aVoice: Integer; const aValue: TSignal);
    begin
      FState[ aVoice].Q := aValue;
      UpdateCoefficients( aVoice);
    end;


    procedure   TRBJFilter.SetGain( aVoice: Integer; const aValue: TSignal);
    begin
      FState[ aVoice].AA := aValue;
      UpdateCoefficients( aVoice);
    end;


    procedure   TRBJFilter.SetFilterType( aVoice: Integer; aValue: TFilterType);
    begin
      FState[ aVoice].T := aValue;
      UpdateCoefficients( aVoice);
    end;


//  protected

    procedure   TRBJFilter.Initialize; // override;
    begin
      inherited;
      SetLength( FState, Polyphony);
    end;


//  public

    constructor TRBJFilter.Create( aPolyphony: Integer; aSampleRate: TSignal; aFilterType: TFilterType);
    var
      State  : PState;
      aVoice : Integer;
    begin
      inherited Create( aPolyphony, aSampleRate);

      for aVoice := 0 to Polyphony - 1
      do begin
        State           := @ FState[ aVoice];
        State^.T        := aFilterType;
        State^.Q        := 1.0;
        State^.AA       := 1.0;
        State^.A        := Init3;
        State^.B        := Init3;
        CutOff[ aVoice] := 329.6;
        UpdateCoefficients( aVoice);
      end;
    end;


    procedure   TRBJFilter.UpdateCoefficients( aVoice: Integer);
    var
      Factor : TSignal;
      ANorm  : TCoefs2;
      BNorm  : TCoefs3;
      State  : PState;
    begin
      State           := @ FState[ aVoice];
      State^.CosOmega := Cos( State^.Omega);
      State^.SinOmega := Sin( State^.Omega);

      case State^.T of

        FT_LOWPASS:

          begin
            State^.Alpha := State^.SinOmega / ( 2.0 * State^.Q);
            State^.B[ 0] := ( 1 - State^.CosOmega) / 2.0;
            State^.B[ 1] :=   1 - State^.CosOmega;
            State^.B[ 2] := State^.B[ 0];
            State^.A[ 0] :=  1.0 + State^.Alpha;
            State^.A[ 1] := -2.0 * State^.CosOmega;
            State^.A[ 2] :=  1.0 - State^.Alpha;
          end;

        FT_HIGHPASS:

          begin
            State^.Alpha := State^.SinOmega / ( 2.0 * State^.Q);
            State^.B[ 0] :=  ( 1.0 + State^.CosOmega) / 2;
            State^.B[ 1] := -( 1.0 + State^.CosOmega);
            State^.B[ 2] := State^.B[ 0];
            State^.A[ 0] :=  1.0 + State^.Alpha;
            State^.A[ 1] := -2.0 * State^.CosOmega;
            State^.A[ 2] :=  1.0 - State^.Alpha;
          end;

        FT_BANDPASS:

          begin
            State^.Alpha := State^.SinOmega * Sinh( Ln( 2.0) / 2.0 * State^.Q * State^.Omega / State^.SinOmega);
            State^.B[ 0] := State^.SinOmega / 2.0;
            State^.B[ 1] := 0.0;
            State^.B[ 2] := - State^.B[ 0];
            State^.A[ 0] :=  1.0 + State^.Alpha;
            State^.A[ 1] := -2.0 * State^.CosOmega;
            State^.A[ 2] :=  1.0 - State^.Alpha;
          end;

        FT_ALLPASS:

          begin
            State^.Alpha := State^.SinOmega / ( 2.0 * State^.Q);
            State^.B[ 0] :=   1.0 - State^.Alpha;
            State^.B[ 1] := - 2.0 * State^.CosOmega;
            State^.B[ 2] := 1.0 + State^.Alpha;
            State^.A[ 0] := State^.B[ 2];
            State^.A[ 1] := State^.B[ 1];
            State^.A[ 2] := State^.B[ 0];
          end;

        FT_NOTCH:

          begin
            State^.Alpha := State^.SinOmega * Sinh( Ln( 2.0) / 2.0 * State^.Q * State^.Omega / State^.SinOmega);
            State^.B[ 0] := 1.0;
            State^.B[ 1] := -2.0 * State^.CosOmega;
            State^.B[ 2] := 1.0;
            State^.A[ 0] := 1.0 + State^.Alpha;
            State^.A[ 1] := State^.B[ 1];
            State^.A[ 2] := 1.0 - State^.Alpha;
          end;

        FT_PEAK:

          begin
            State^.Alpha := State^.SinOmega * Sinh( Ln( 2.0) / 2.0 * State^.Q * State^.Omega / State^.SinOmega);
            State^.B[ 0] := 1.0 + ( State^.Alpha * State^.AA);
            State^.B[ 1] := -2.0 * State^.CosOmega;
            State^.B[ 2] := 1.0 - ( State^.Alpha * State^.AA);
            State^.A[ 0] := 1.0 + ( State^.Alpha / State^.AA);
            State^.A[ 1] := State^.B[ 1];
            State^.A[ 2] := 1.0 - ( State^.Alpha / State^.AA);
          end;

        FT_LOW_SHELF:

          begin
            State^.Alpha := State^.SinOmega / 2.0 * Sqrt(( State^.AA + 1.0 / State^.AA) * ( 1.0 / State^.Q - 1.0) + 2.0);
            State^.B[ 0] := State^.AA * (( State^.AA + 1.0) - (( State^.AA - 1.0) * State^.CosOmega) + ( 2.0 * Sqrt( State^.AA) * State^.Alpha));
            State^.B[ 1] := 2.0 * State^.AA * (( State^.AA - 1.0) - (( State^.AA + 1.0) * State^.CosOmega));
            State^.B[ 2] := State^.AA * (( State^.AA + 1.0) - (( State^.AA - 1.0) * State^.CosOmega) - ( 2.0 * Sqrt( State^.AA) * State^.Alpha));
            State^.A[ 0] := (( State^.AA + 1.0) + (( State^.AA - 1.0) * State^.CosOmega) + ( 2.0 * Sqrt( State^.AA) * State^.Alpha));
            State^.A[ 1] := -2.0 * (( State^.AA - 1.0) + (( State^.AA + 1.0) * State^.CosOmega));
            State^.A[ 2] := (( State^.AA + 1.0) + (( State^.AA - 1.0) * State^.CosOmega) - ( 2.0 * Sqrt( State^.AA) * State^.Alpha));
          end;

        FT_HIGH_SHELF:
          begin
            State^.Alpha := State^.SinOmega / 2.0 * Sqrt(( State^.AA + 1.0 / State^.AA) * ( 1.0 / State^.Q - 1.0) + 2.0);
            State^.B[ 0] := State^.AA * (( State^.AA + 1.0) + (( State^.AA - 1.0) * State^.CosOmega) + ( 2.0 * Sqrt( State^.AA) * State^.Alpha));
            State^.B[ 1] := - 2.0 * State^.AA * (( State^.AA - 1.0) + (( State^.AA + 1.0) * State^.CosOmega));
            State^.B[ 2] := State^.AA * (( State^.AA + 1.0) + (( State^.AA - 1.0) * State^.CosOmega) - ( 2.0 * Sqrt( State^.AA) * State^.Alpha));
            State^.A[ 0] := (( State^.AA + 1.0) - (( State^.AA - 1.0) * State^.CosOmega) + ( 2.0 * Sqrt( State^.AA) * State^.Alpha));
            State^.A[ 1] := 2.0 * (( State^.AA - 1.0) - (( State^.AA + 1.0) * State^.CosOmega));
            State^.A[ 2] := (( State^.AA + 1.0) - (( State^.AA - 1.0) * State^.CosOmega) - ( 2.0 * Sqrt( State^.AA) * State^.Alpha));
          end;
      end;

      // Normalize filter coefficients

      Factor := 1.0 / State^.A[ 0];

      ANorm[ 0] := State^.A[ 1] * Factor;
      ANorm[ 1] := State^.A[ 2] * Factor;

      BNorm[ 0] := State^.B[ 0] * Factor;
      BNorm[ 1] := State^.B[ 1] * Factor;
      BNorm[ 2] := State^.B[ 2] * Factor;

      SetBiquadCoefs( aVoice, BNorm, ANorm);
    end;


{ ========
  TPinkingFilter = class( TSimpleFilter)
  // +/-0.05dB above 9.2Hz @ 44,100Hz
  private type
    TState = record
      B0 : TSignal;
      B1 : TSignal;
      B2 : TSignal;
      B3 : TSignal;
      B4 : TSignal;
      B5 : TSignal;
      B6 : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
}

    procedure   TPinkingFilter.Initialize; // override;
    var
      State  : PState;
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        State     := @ FState[ aVoice];
        State^.B0 := 0.0;
        State^.B1 := 0.0;
        State^.B2 := 0.0;
        State^.B3 := 0.0;
        State^.B4 := 0.0;
        State^.B5 := 0.0;
        State^.B6 := 0.0;
      end;
    end;


//  public

    function    TPinkingFilter.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      State : PState;
      Pink  : TSignal;
    begin
      State     := @ FState[ aVoice];
      State^.B0 := 0.99886 * State^.B0 + aSample * 0.0555179;
      State^.B1 := 0.99332 * State^.B1 + aSample * 0.0750759;
      State^.B2 := 0.96900 * State^.B2 + aSample * 0.1538520;
      State^.B3 := 0.86650 * State^.B3 + aSample * 0.3104856;
      State^.B4 := 0.55000 * State^.B4 + aSample * 0.5329522;
      State^.B5 := -0.7616 * State^.B5 - aSample * 0.0168980;
      Pink := (
          State^.B0
        + State^.B1
        + State^.B2
        + State^.B3
        + State^.B4
        + State^.B5
        + State^.B6
        + ( aSample * 0.5362)
      ) * 0.11;
      State^.B6 := aSample * 0.115926;
      Result    := Pink;
    end;


{ ========
  TBrowningFilter = class( TSimpleFilter)
  private type
    TState = record
      L : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
}

    procedure   TBrowningFilter.Initialize; // override;
    var
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do FState[ aVoice].L := 0.0;
    end;


//  public

    function    TBrowningFilter.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      State : PState;
      Brown : TSignal;
    begin
      State    := @ FState[ aVoice];
      Brown    := ( State^.L + ( 0.02 * aSample)) / 1.02;
      State^.L := Brown;
      Result   := Brown * 3.5;    // compensate for gain [loss]
    end;


{ ========
  TLadderFilterBase = class( TSimpleFilter)
  private type
    TState = record
      CutOff    : TSignal;
      Resonance : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  public
    property    CutOff   [ aVoice: Integer] : TSignal read GetCutOff    write SetCutOff;
    property    Resonance[ aVoice: Integer] : TSignal read GetResonance write SetResonance;
  protected
}

    function    TLadderFilterBase.GetCutOff( aVoice: Integer): TSignal; // virtual;
    begin
      Result := FState[ aVoice].CutOff;
    end;


    procedure   TLadderFilterBase.SetCutOff( aVoice: Integer; const aValue: TSignal); // virtual;
    begin
      FState[ aVoice].CutOff := aValue;
    end;


    function    TLadderFilterBase.GetResonance( aVoice: Integer): TSignal; // virtual;
    begin
      Result := FState[ aVoice].Resonance;
    end;


    procedure   TLadderFilterBase.SetResonance( aVoice: Integer; const aValue: TSignal); // virtual;
    begin
      FState[ aVoice].Resonance := aValue;
    end;


//  protected

    procedure   TLadderFilterBase.Initialize; // override;
    var
      State  : PState;
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        State := @ FState[ aVoice];
        State^.CutOff     := 0.0;
        State^.Resonance  := 0.0;
      end;
    end;


{ ========
  THuovilainenMoog = class( TLadderFilterBase)
  strict private
  const
    Thermal = 0.000025;
    Comp    = 0.5;
  private
    type TState = record
      Stage0     : TSignal;
      Stage1     : TSignal;
      Stage2     : TSignal;
      Stage3     : TSignal;
      StageTanh0 : TSignal;
      StageTanh1 : TSignal;
      StageTanh2 : TSignal;
      Delay0     : TSignal;
      Delay1     : TSignal;
      Delay2     : TSignal;
      Delay3     : TSignal;
      Delay4     : TSignal;
      Delay5     : TSignal;
      Tune       : TSignal;
      Acr        : TSignal;
      ResQuad    : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
}

    procedure   THuovilainenMoog.SetResonance( aVoice: Integer; const aValue: TSignal); // override;
    begin
      inherited;
      FState[ aVoice].ResQuad := Resonance[ aVoice] * FState[ aVoice].Acr;
    end;


    procedure   THuovilainenMoog.SetCutOff( aVoice: Integer; const aValue: TSignal); // override;
    var
      State : PState;
      fc    : TSignal;
      f     : TSignal;
      fc2   : TSignal;
      fc3   : TSignal;
      fcr   : TSignal;
    begin
      inherited;
      State          := @ FState[ aVoice];
      fc             := Cutoff[ aVoice] / SampleRate;
      f              := fc * 0.5;                                   // oversampled twice
      fc2            := fc * fc;
      fc3            := fc2 * fc;
      fcr            := 1.8730 * fc3 + 0.4955 * fc2 - 0.6490 * fc + 0.9988;
      State^.Acr     := -3.9364 * fc2 + 1.8409 * fc + 0.9968;
      State^.Tune    := ( 1.0 - exp( - ( TWO_PI * f * fcr))) / Thermal;
      State^.ResQuad := Resonance[ aVoice] * State^.Acr;
    end;


    procedure   THuovilainenMoog.Initialize; // override;
    var
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        CutOff   [ aVoice] := 329.6;
        Resonance[ aVoice] := 0.10;
      end;
    end;


//  public

    function    THuovilainenMoog.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      State  : PState;
      InVal  : TSignal;
      PreVal : TSignal;
    begin
      State  := @ FState[ aVoice];
      PreVal := Normalize( aSample);
      InVal  := Normalize( PreVal - State^.ResQuad * ( State^.Delay5 - Comp * aSample));

      State^.Stage0     := Normalize( State^.Delay0 + State^.Tune * LookupTanh( InVal * Thermal) - State^.StageTanh0);
      State^.Delay0     := State^.Stage0;
      State^.StageTanh0 := Normalize( LookupTanh( State^.Stage0 * Thermal));
      State^.Stage1     := Normalize( State^.Delay1 + State^.Tune * ( State^.StageTanh0 - State^.StageTanh1));
      State^.Delay1     := State^.Stage1;
      State^.StageTanh1 := Normalize( LookupTanh( State^.Stage1 * Thermal));
      State^.Stage2     := Normalize( State^.Delay2 + State^.Tune * ( State^.StageTanh1 - State^.StageTanh2));
      State^.Delay2     := State^.Stage2;
      State^.StageTanh2 := Normalize( LookupTanh( State^.Stage2 * Thermal));
      State^.Stage3     := Normalize( State^.Delay3 + State^.Tune * ( State^.StageTanh2 - State^.Delay3 * Thermal));
      State^.Delay3     := State^.Stage3;
      State^.Delay5     := 0.5 * ( State^.Stage3 + State^.Delay4);
      State^.Delay4     := State^.Stage3;

      InVal := Normalize( PreVal - State^.ResQuad * ( State^.Delay5 - Comp * aSample));

      State^.Stage0     := Normalize( State^.Delay0 + State^.Tune * LookupTanh( InVal * Thermal) - State^.StageTanh0);
      State^.Delay0     := State^.Stage0;
      State^.StageTanh0 := Normalize( LookupTanh( State^.Stage0 * Thermal));
      State^.Stage1     := Normalize( State^.Delay1 + State^.Tune * ( State^.StageTanh0 - State^.StageTanh1));
      State^.Delay1     := State^.Stage1;
      State^.StageTanh1 := Normalize( LookupTanh( State^.Stage1 * Thermal));
      State^.Stage2     := Normalize( State^.Delay2 + State^.Tune * ( State^.StageTanh1 - State^.StageTanh2));
      State^.Delay2     := State^.Stage2;
      State^.StageTanh2 := Normalize( LookupTanh( State^.Stage2 * Thermal));
      State^.Stage3     := Normalize( State^.Delay3 + State^.Tune * ( State^.StageTanh2 - State^.Delay3 * Thermal));
      State^.Delay3     := State^.Stage3;
      State^.Delay5     := 0.5 * ( State^.Stage3 + State^.Delay4);
      State^.Delay4     := State^.Stage3;

      Result := LookupTanh( State^.Delay5);
    end;


{ ========
  TImprovedMoog = class( TLadderFilterBase)
  strict private
  const
    VT = 0.312;
  private
    type TState = record
      V     : TCoefs4;
      DV    : TCoefs4;
      TV    : TCoefs4;
      X     : TSignal;
      G     : TSignal;
      Drive : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
}

    procedure   TImprovedMoog.SetCutOff( aVoice: Integer; const aValue: TSignal); // override;
    var
      State : PState;
    begin
      inherited;
        State    := @ FState[ aVoice];
        State^.X := ( PI * CutOff[ aVoice]) / SampleRate;
	State^.G := 4.0 * PI * VT * CutOff[ aVoice] * ( 1.0 - State^.X) / ( 1.0 + State^.X);
      end;


    procedure   TImprovedMoog.Initialize; // override;
    var
      State  : PState;
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        State              := @ FState[ aVoice];
        State^.V           := Init4;
        State^.DV          := Init4;
        State^.TV          := Init4;
        State^.Drive       := 1.0;
        CutOff   [ aVoice] := 329.6;
        Resonance[ aVoice] := 0.1;      // [0 .. 4]
      end;
    end;


//  public

    function    TImprovedMoog.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      LDV   : TCoefs4;
      State : PState;
    begin
      State   := @ FState[ aVoice];
      aSample := 3.0 * aSample;  // Added some pre-ampification, filter was a  bit lame ..

      LDV      [ 0] := - State^.G * ( LookupTanH(( State^.Drive * 2.0 * aSample + 3.0 * Resonance[ aVoice] * State^.V[ 3]) / ( 2.0 * VT)) + State^.TV[ 0]);
      State^.V [ 0] := State^.V[ 0] + ( LDV[ 0] + State^.DV[ 0]) / ( 2.0 * SampleRate);
      State^.DV[ 0] := LDV[ 0];
      State^.TV[ 0] := LookupTanH( State^.V[ 0] / ( 2.0 * VT));

      LDV      [ 1] := State^.G * ( State^.TV[ 0] - State^.TV[ 1]);
      State^.V [ 1] := State^.V[ 1] + ( LDV[ 1] + State^.DV[ 1]) / ( 2.0 * SampleRate);
      State^.DV[ 1] := LDV[ 1];
      State^.TV[ 1] := LookupTanH( State^.V[ 1] / ( 2.0 * VT));

      LDV      [ 2] := State^.G * ( State^.TV[ 1] - State^.TV[ 2]);
      State^.V [ 2] := State^.V [ 2] + ( LDV[ 2] + State^.DV[ 2]) / ( 2.0 * SampleRate);
      State^.DV[ 2] := LDV[ 2];
      State^.TV[ 2] := LookupTanH( State^.V[ 2] / ( 2.0 * VT));

      LDV      [ 3] := State^.G * ( State^.TV[ 2] - State^.TV[ 3]);
      State^.V [ 3] := State^.V[3] + ( LDV[ 3] + State^.DV[ 3]) / ( 2.0 * SampleRate);
      State^.DV[ 3] := LDV[ 3];
      State^.TV[ 3] := LookupTanH( State^.V[ 3] / ( 2.0 * VT));

      Result := State^.V[ 3];
    end;


{ ========
  TKrajeskiMoog = class( TLadderFilterBase)
  strict private
  private
    type TState = record
      State : TCoefs5;
      Delay : TCoefs5;
      WC    : TSignal;
      G     : TSignal;
      GRes  : TSignal;
      GComp : TSignal;
      Drive : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
}

    procedure   TKrajeskiMoog.SetCutOff( aVoice: Integer; const aValue: TSignal); // override;
    var
      State : PState;
      WC1   : TSignal;
      WC2   : TSignal;
      WC3   : TSignal;
      WC4   : TSignal;
    begin
      inherited;
      State     := @ FState[ aVoice];
      State^.WC := TWO_PI * CutOff[ aVoice] / SampleRate;
      WC1       := State^.WC;
      WC2       := WC1 * WC1;
      WC3       := WC1 * WC2;
      WC4       := WC2 * WC2;
      State^.G  :=
          0.9892 * WC1
        - 0.4342 * WC2
        + 0.1381 * WC3
        - 0.0202 * WC4
        ;
    end;


    procedure   TKrajeskiMoog.SetResonance( aVoice: Integer; const aValue: TSignal); // override;
    var
      State : PState;
      WC1   : TSignal;
      WC2   : TSignal;
      WC3   : TSignal;
    begin
      inherited;
      State       := @ FState[ aVoice];
      WC1         := State^.WC;
      WC2         := WC1 * WC1;
      WC3         := WC2 * WC1;
      State^.GRes := Resonance[ aVoice] * ( 1.0029 + 0.0526 * WC1 - 0.926 * WC2 + 0.0218 * WC3);
    end;


    procedure   TKrajeskiMoog.Initialize; // override;
    var
      State  : PState;
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        State              := @ FState[ aVoice];
        State^.State       := Init5;
        State^.Delay       := Init5;
        State^.GComp       := 1.0;
        State^.Drive       := 1.0;
        CutOff   [ aVoice] := 329.6;
        Resonance[ aVoice] := 0.1;
      end;
    end;


//  public

    function    TKrajeskiMoog.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      State : PState;
      i     : Integer;
    begin
      State            := @ FState[ aVoice];
      State^.State[ 0] := LookupTanH( State^.Drive * ( aSample - 4 * State^.GRes * ( State^.State[ 4] - State^.GComp * aSample)));

      for i := 0 to 3
      do begin
        State^.State[ i + 1] := State^.G * ( 0.3 / 1.3 * State^.State[ i] + 1 / 1.3 * State^.Delay[ i] - State^.State[ i + 1]) + State^.State[ i + 1];
        State^.Delay[ i    ] := State^.State[ i];
      end;

      Result := State^.State[ 4];
    end;


{ ========
  TMicroTrackerMoog = class( TLadderFilterBase)
  strict private
  private
    type TState = record
      P0   : TSignal;
      P1   : TSignal;
      P2   : TSignal;
      P3   : TSignal;
      P32  : TSignal;
      P33  : TSignal;
      P34  : TSignal;
      GCut : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
}
    procedure   TMicroTrackerMoog.SetCutOff( aVoice: Integer; const aValue: TSignal); // override;
    begin
      inherited;
      FState[ aVoice].GCut := Min( CutOff[ aVoice] * TWO_PI / SampleRate, 1);
    end;


    procedure   TMicroTrackerMoog.Initialize; // override;
    var
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        FState   [ aVoice].P0  := 0.0;
        FState   [ aVoice].P1  := 0.0;
        FState   [ aVoice].P2  := 0.0;
        FState   [ aVoice].P3  := 0.0;
        FState   [ aVoice].P32 := 0.0;
        FState   [ aVoice].P33 := 0.0;
        FState   [ aVoice].P34 := 0.0;
        CutOff   [ aVoice]     := 329.6;
        Resonance[ aVoice]     := 0.1;
      end;
    end;


//  public

    function    TMicroTrackerMoog.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      State : PState;
      K     : TSignal;
    begin
      State := @ FState[ aVoice];
      K     := Resonance[ aVoice] * 4.0;

      // Coefficients optimized using differential evolution
      // to make feedback gain 4.0 correspond closely to the
      // border of instability, for all values of omega.

      Result     := State^.P3 * 0.360891 + State^.P32 * 0.417290 + State^.P33 * 0.177896 + State^.P34 * 0.0439725;
      State^.P34 := State^.P33;
      State^.P33 := State^.P32;
      State^.P32 := State^.P3;
      State^.P0  := State^.P0 + ( FastTanH( 2.0 * aSample - K * Result) - FastTanH( State^.P0)) * State^.GCut;
      State^.P1  := State^.P1 + ( FastTanH( State^.P0)   - FastTanH( State^.P1)) * State^.GCut;
      State^.P2  := State^.P2 + ( FastTanH( State^.P1)   - FastTanH( State^.P2)) * State^.GCut;
      State^.P3  := State^.P3 + ( FastTanH( State^.P2)   - FastTanH( State^.P3)) * State^.GCut;
    end;


{ ========
  TMusicDSPMoog = class( TLadderFilterBase)
  strict private
  private
    type TState = record
      GReso   : TSignal;
      GCutOff : TSignal;
      Stage   : TCoefs4;
      Delay   : TCoefs4;
      P       : TSignal;
      K       : TSignal;
      T1      : TSignal;
      T2      : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
}

    procedure   TMusicDSPMoog.SetCutOff( aVoice: Integer; const aValue: TSignal); // override;
    var
      State : PState;
    begin
      inherited;
      State              := @ FState[ aVoice];
      State^.GCutOff     := 2.0 * CutOff[ aVoice] / SampleRate;
      State^.P           := State^.GCutOff * ( 1.8 - 0.8 * State^.GCutOff);
      State^.K           := 2.0 * Sin( State^.GCutOff * PI * 0.5) - 1.0;
      State^.T1          := ( 1.0 - State^.P) * 1.386249;
      State^.T2          := 12.0 + State^.T1 * State^.T1;
      Resonance[ aVoice] := Resonance[ aVoice];
    end;


    procedure   TMusicDSPMoog.SetResonance( aVoice: Integer; const aValue: TSignal); // override;
    var
      State : PState;
    begin
      inherited;
      State        := @ FState[ aVoice];
      State^.GReso := 0.8 * Resonance[ aVoice] * ( State^.T2 + 6.0 * State^.T1) / ( State^.T2 - 6.0 * State^.T1);
    end;


    procedure   TMusicDSPMoog.Initialize; // override;
    var
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        FState   [ aVoice].GReso   := 0.0;
        FState   [ aVoice].GCutOff := 0.0;
        FState   [ aVoice].Stage   := Init4;
        FState   [ aVoice].Delay   := Init4;
        FState   [ aVoice].P       := 0.0;
        FState   [ aVoice].K       := 0.0;
        FState   [ aVoice].T1      := 0.0;
        FState   [ aVoice].T2      := 0.0;
        CutOff   [ aVoice]         := 329.6;
        Resonance[ aVoice]         := 0.1;
      end;
    end;


//  public

    function    TMusicDSPMoog.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      State : PState;
      X     : TSignal;
    begin
      State := @ FState[ aVoice];
      X     := 2.0 * aSample - State^.GReso * State^.Stage[ 3];

      // Four cascaded one-pole filters (bilinear transform)

      State^.Stage[ 0] := X * State^.P + State^.Delay[ 0]  * State^.P - State^.K * State^.Stage[ 0];
      State^.Stage[ 1] := State^.Stage[ 0] * State^.P + State^.Delay[ 1] * State^.P - State^.K * State^.Stage[ 1];
      State^.Stage[ 2] := State^.Stage[ 1] * State^.P + State^.Delay[ 2] * State^.P - State^.K * State^.Stage[ 2];
      State^.Stage[ 3] := State^.Stage[ 2] * State^.P + State^.Delay[ 3] * State^.P - State^.K * State^.Stage[ 3];

      // Clipping band-limited sigmoid

      State^.Stage[ 3] := State^.Stage[3] - ( State^.Stage[ 3] * State^.Stage[ 3] * State^.Stage [3]) / 6.0;

      State^.Delay[ 0] := X;
      State^.Delay[ 1] := State^.Stage[ 0];
      State^.Delay[ 2] := State^.Stage[ 1];
      State^.Delay[ 3] := State^.Stage[ 2];

      Result := State^.Stage[3 ];
    end;


{ ========
  TVAOnePole = class( TSimpleFilter)
  //
  //  One pole for the OberheimVariationMoog filter
  //
  strict private
  private
    type TState = record
      Alpha    : TSignal;
      Beta     : TSignal;
      Gamma    : TSignal;
      Delta    : TSignal;
      Epsilon  : TSignal;
      A0       : TSignal;
      Feedback : TSignal;
      Z1       : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
}

    procedure   TVAOnePole.SetFeedback( aVoice: Integer; const aValue: TSignal);
    begin
      FState[ aVoice].Feedback := aValue;
    end;


    function    TVAOnePole.GetFeedbackOutput( aVoice: Integer): TSignal;
    var
      State : PState;
    begin
      State  := @ FState[ aVoice];
      Result := State^.Beta * ( State^.Z1 + State^.Feedback * State^.Delta);
    end;


    procedure   TVAOnePole.SetAlpha( aVoice: Integer; const aValue: TSignal);
    begin
      FState[ aVoice].Alpha := aValue;
    end;


    procedure   TVAOnePole.SetBeta( aVoice: Integer; const aValue: TSignal);
    begin
      FState[ aVoice].Beta := aValue;
    end;


    procedure   TVAOnePole.Initialize; // override;
    var
      aVoice : Integer;
      State  : PState;
    begin
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        State           := @ FState[ aVoice];
        State^.Alpha    := 1.0;
        State^.Beta     := 0.0;
        State^.Gamma    := 1.0;
        State^.Delta    := 0.0;
        State^.Epsilon  := 0.0;
        State^.A0       := 1.0;
        State^.Feedback := 0.0;
        State^.Z1       := 0.0;
      end;
    end;


//  public

    function    TVAOnePole.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      State : PState;
      VN    : TSignal;
    begin
      State     := @ FState[ aVoice];
      aSample   := aSample * State^.Gamma + State^.Feedback + State^.Epsilon * GetFeedbackOutput( aVoice);
      VN        := ( State^.A0 * aSample - State^.Z1) * State^.Alpha;
      Result    := vn + State^.Z1;
      State^.z1 := VN + Result;
    end;


{ ========
  TOberheimVariationMoog = class( TLadderFilterBase)
  strict private
  private
    type TState = record
      LPF1          : TVAOnePole;
      LPF2          : TVAOnePole;
      LPF3          : TVAOnePole;
      LPF4          : TVAOnePole;
      K             : TSignal;
      Gamma         : TSignal;
      Alpha0        : TSignal;
      Saturation    : TSignal;
      OberheimCoefs : TCoefs5;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
}

    procedure   TOberheimVariationMoog.SetCutOff( aVoice: Integer; const aValue: TSignal); // override;
    var
      State : PState;
      WD    : TSignal;
      T     : TSignal;
      WA    : TSignal;
      G     : TSignal;
      BigG  : TSignal;
      G1    : TSignal;
    begin
      inherited;
      State := @ FState[ aVoice];

      // prewarp for BZT

      WD   := TWO_PI * CutOff[ aVoice];
      T    := 1.0 / SampleRate;
      WA   := (2.0 / T) * Tan( WD * T * 0.5);
      G    := WA * T * 0.5;
      G1   := 1.0 / ( 1.0 + G);
      BigG := G * G1;               // Feedforward coeff

      State^.LPF1.SetAlpha( aVoice, BigG);
      State^.LPF2.SetAlpha( aVoice, BigG);
      State^.LPF3.SetAlpha( aVoice, BigG);
      State^.LPF4.SetAlpha( aVoice, BigG);
      State^.LPF1.SetBeta ( aVoice, BigG * BigG * BigG * G1);
      State^.LPF2.SetBeta ( aVoice, BigG * BigG        * G1);
      State^.LPF3.SetBeta ( aVoice, BigG               * G1);
      State^.LPF4.SetBeta ( aVoice, 1.0                * G1);

      State^.Gamma  := BigG * BigG * BigG * BigG;
      State^.Alpha0 := 1.0 / ( 1.0 + State^.K * State^.Gamma);

      // Oberheim variations / LPF4

      State^.OberheimCoefs[ 0] := 0.0;
      State^.OberheimCoefs[ 1] := 0.0;
      State^.OberheimCoefs[ 2] := 0.0;
      State^.OberheimCoefs[ 3] := 0.0;
      State^.OberheimCoefs[ 4] := 1.0;
    end;


    procedure   TOberheimVariationMoog.SetResonance( aVoice: Integer; const aValue: TSignal); // override;
    begin
      inherited;

      // this maps resonance = 1->10 to K = 0 -> 4 ... but that seems to give not enough resonance ...
   // FState[ aVoice].K := ( 4.0) * ( aValue - 1.0)/( 10.0 - 1.0);

      // still not self oscillating .. but better reso range ..

      FState[ aVoice].K := 4.0 * aValue;
    end;


    procedure   TOberheimVariationMoog.Initialize; // override;
    var
      State  : PState;
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        State := @ FState[ aVoice];
        State^.LPF1        := TVAOnePole.Create( Polyphony, SampleRate);
        State^.LPF2        := TVAOnePole.Create( Polyphony, SampleRate);
        State^.LPF3        := TVAOnePole.Create( Polyphony, SampleRate);
        State^.LPF4        := TVAOnePole.Create( Polyphony, SampleRate);
        State^.Saturation  := 1.0;
        CutOff   [ aVoice] := 329.6;
        Resonance[ aVoice] := 0.1;
      end;
    end;


//  public

    destructor  TOberheimVariationMoog.Destroy; // override;
    var
      State  : PState;
      aVoice : Integer;
    begin
      for aVoice := 0 to Polyphony - 1
      do begin
        State := @ FState[ aVoice];
        FreeAndNil( State^.LPF1);
        FreeAndNil( State^.LPF2);
        FreeAndNil( State^.LPF3);
        FreeAndNil( State^.LPF4);
      end;
      inherited;
    end;


    function    TOberheimVariationMoog.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      State  : PState;
      Input  : TSignal;
      Sigma  : TSignal;
      U      : TSignal;
      Stage1 : TSignal;
      Stage2 : TSignal;
      Stage3 : TSignal;
      Stage4 : TSignal;
    begin
      State := @ FState[ aVoice];
      Input := 2.0 * aSample;

      Sigma :=
          State^.LPF1.GetFeedbackOutput( aVoice)
        + State^.LPF2.GetFeedbackOutput( aVoice)
        + State^.LPF3.GetFeedbackOutput( aVoice)
        + State^.LPF4.GetFeedbackOutput( aVoice)
        ;

      Input := Input * 1.0 + State^.K;

      // calculate input to first filter

      U      := ( Input - State^.K * Sigma) * State^.Alpha0;
      U      := LookupTanH( State^.Saturation * U);
      stage1 := State^.LPF1.Tick( aVoice, U     );
      stage2 := State^.LPF2.Tick( aVoice, Stage1);
      stage3 := State^.LPF3.Tick( aVoice, Stage2);
      stage4 := State^.LPF4.Tick( aVoice, Stage3);

      // Oberheim variations

      Result :=
          State^.OberheimCoefs[ 0] * U
        + State^.OberheimCoefs[ 1] * Stage1
        + State^.OberheimCoefs[ 2] * Stage2
        + State^.OberheimCoefs[ 3] * Stage3
        + State^.OberheimCoefs[ 4] * Stage4
        ;
    end;


{ ========
  TRKSimulationMoog = class( TLadderFilterBase)
  strict private
  private
    type TState = record
      State         : TCoefs4;
      Saturation    : TSignal;
      SaturationInv : TSignal;
      Reso          : TSignal;
      F             : TSignal;
    end;
    PState = ^ TState;
  private
    FOverSampleFactor : Integer;
    FStepSize         : TSignal;
    FState            : TArray<TState>;
  private
}

    procedure   TRKSimulationMoog.CalculateDerivatives( aVoice: Integer; aSample: TSignal; var DState: TCoefs4; const State: TCoefs4);
    var
      GState    : PState;
      SatStateN : TSignal;
      SatState0 : TSignal;
      SatState1 : TSignal;
      SatState2 : TSignal;
      SatState3 : TSignal;
      Sat       : TSignal;
      SatInv    : TSignal;
      Reso      : TSignal;
      F         : TSignal;
    begin
      GState     := @ FState[ aVoice];
      Sat        := GState^.Saturation;
      SatInv     := GState^.SaturationInv;
      Reso       := GState^.Reso;
      F          := GState^.F;
      SatStateN  := SoftClip( aSample - Reso * State[ 3], Sat, SatInv);
      SatState0  := SoftClip( State[ 0], Sat, SatInv);
      SatState1  := SoftClip( State[ 1], Sat, SatInv);
      SatState2  := SoftClip( State[ 2], Sat, SatInv);
      SatState3  := SoftClip( State[ 3], Sat, SatInv);
      DState[ 0] := F * ( SatStateN - SatState0);
      DState[ 1] := F * ( SatState0 - SatState1);
      DState[ 2] := F * ( SatState1 - SatState2);
      DState[ 3] := F * ( SatState2 - SatState3);
    end;


    procedure   TRKSimulationMoog.RungeKuttaSolver( aVoice: Integer; aSample: TSignal; var State: TCoefs4);
    var
      i         : Integer;
      Deriv1    : TCoefs4;
      Deriv2    : TCoefs4;
      Deriv3    : TCoefs4;
      Deriv4    : TCoefs4;
      TempState : TCoefs4;
    begin
      CalculateDerivatives( aVoice, aSample, Deriv1, State);

      for i := 0 to 3
      do TempState[ i] := State[ i] + 0.5 * FStepSize * Deriv1[ i];

      CalculateDerivatives( aVoice, aSample, Deriv2, TempState);

      for i := 0 to 3
      do tempState[ i] := State[ i] + 0.5 * FStepSize * Deriv2[ i];

      CalculateDerivatives( aVoice, aSample, Deriv3, TempState);

      for i := 0 to 3
      do TempState[ i] := State[ i] + FStepSize * Deriv3[ i];

      CalculateDerivatives( aVoice, aSample, Deriv4, TempState);

      for i := 0 to 3
      do State[ i] := State[ i] + ( 1.0 / 6.0) * FStepSize * ( Deriv1[ i] + 2.0 * Deriv2[ i] + 2.0 * Deriv3[ i] + Deriv4[ i]);
    end;


//  protected

    procedure   TRKSimulationMoog.SetCutOff( aVoice: Integer; const aValue: TSignal); // override;
    begin
      inherited;
      FState[ aVoice].F := TWO_PI * CutOff[ aVoice];
    end;


    procedure   TRKSimulationMoog.SetResonance( aVoice: Integer; const aValue: TSignal); // override;
    begin
      inherited;
      FState[ aVoice].Reso := 5.0 * Resonance[ aVoice];
    end;


    procedure   TRKSimulationMoog.Initialize; // override;
    var
      State  : PState;
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        State                := @ FState[ aVoice];
        State^.State         := Init4;
        State^.Saturation    := 3.0;
        State^.SaturationInv := 1.0 / State^.Saturation;
        FOverSampleFactor    := 1;
        FStepSize            := 1.0 / ( FOverSampleFactor * SampleRate);
        CutOff   [ aVoice]   := 329.6;
        Resonance[ aVoice]   := 0.1;
      end;
    end;


//  public

    function    TRKSimulationMoog.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      State : PState;
      j     : Integer;
    begin
      State := @ FState[ aVoice];

      for j := 0 to FOversampleFactor - 1
      do RungekuttaSolver( aVoice, 3.0 * aSample, State^.State);

      Result := State^.State[ 3];
    end;


{ ========
  TStilsonMoog = class( TLadderFilterBase)
  //
  //  Copyright (c) 2003, David Lowenfels
  //  Released as the moog~ pd extern.
  //
  //  This code is Unlicensed (i.e. public domain); in an email exchange on
  //  6.14.2019 David Lowenfels stated "You're welcome to use the Moog~ code,
  //  license it free as in beer or whatever :)"
  //
  //  A digital model of the classic Moog filter was presented first by Stilson and
  //  Smith. This model uses a cascade of one-pole IIR filters in series with a global
  //  feedback to produce resonance. A digital realization of this filter introduces a
  //  unit delay, effectively making it a fifth-order filter. Unfortunately, this
  //  delay also has the effect of coupling the cutoff and resonance parameters,
  //  uncharacteristic of the uncoupled control of the original Moog ladder. As a
  //  compromise, a zero can be inserted at z = -0.3 inside each one pole section to
  //  minimize the coupling the parameters (humans are not particularly sensitive to
  //  variations in Q factor). Although fast coefficient updates can be achieved since
  //  the nonlinearities of the Moog are not considered, the filter becomes unstable
  //  with very large resonance values and does not enter self-oscillation.
  //  References: Stilson and Smith (1996), DAFX - Zolzer (ed) (2nd ed)
  //  Original implementation: Tim Stilson, David Lowenfels
  //
  strict private
  private
    type TState = record
      State  : TCoefs4;
      Reso   : TSignal;
      Output : TSignal;
      P      : TSignal;
      Q      : TSignal;
    end;
    PState = ^ TState;
  private
    FState : TArray<TState>;
  protected
}

    procedure   TStilsonMoog.SetCutOff( aVoice: Integer; const aValue: TSignal); // override;
    var
      FC : TSignal;
      X2 : TSignal;
      X3 : TSignal;
    begin
      Inherited;

      // Normalized cutoff between [0, 1]

      FC := CutOff[ aVoice] / SampleRate;
      X2 := FC * FC;
      X3 := X2 * FC;

      // Frequency & amplitude correction (Cubic Fit)

      FState[ aVoice].P := -0.69346 * X3 - 0.59515 * X2 + 3.2937 * FC - 1.0072;
      Resonance[ aVoice] := Resonance[ aVoice];
    end;


    procedure   TStilsonMoog.SetResonance( aVoice: Integer; const aValue: TSignal); // override;
    var
      State  : PState;
      IX     : TSignal;
      IXFrac : TSignal;
      IXInt  : Integer;
    begin
      inherited;
      State       := @ FState[ aVoice];
      State^.Reso := Min( Resonance[ aVoice], 1.0);
      IX          := State^.P * 99;
      IXInt       := Floor( IX);
      IXFrac      := IX - IXInt;
      State^.Q    := State^.Reso * MoogLerp( IXFrac, S_STILSON_GAINTABLE[ IXInt + 99], S_STILSON_GAINTABLE[ IXInt + 100]);
    end;


    procedure   TStilsonMoog.Initialize; // override;
    var
      aVoice : Integer;
    begin
      inherited;
      SetLength( FState, Polyphony);

      for aVoice := 0 to Polyphony - 1
      do begin
        FState[ aVoice].State := Init4;
        CutOff   [ aVoice]    := 329.6;
        Resonance[ aVoice]    := 0.1;
      end;
    end;


//  public

    function    TStilsonMoog.Tick( aVoice: Integer; aSample: TSignal): TSignal; // override;
    var
      State      : PState;
      LocalState : TSignal;
      Input      : TSignal;
      Pole       : Integer;
    begin
      State := @ FState[ aVoice];

      // Scale by arbitrary value on account of our saturation function

      Input := aSample * 0.65;

      // Negative Feedback

      State^.Output := 0.25 * ( Input - State^.Output);

      for Pole := 0 to 3
      do begin
        LocalState          := State^.State[ Pole];
        State^.Output       := MoogSaturate( State^.Output + State^.P * ( State^.Output - LocalState));
        State^.State[ pole] := State^.Output;
        State^.Output       := Normalize( MoogSaturate( State^.Output + LocalState));
      end;

      Result        := State^.Output;
      State^.Output := Result * State^.Q;    // Scale stateful output by Q
    end;


end.

