unit PadSynth;


// source : http://zynaddsubfx.sourceforge.net/doc/PadSynth/PadSynth.htm (PadSynth_CPP)
//  COPYRIGHT 2002 .. 2005 Nasca Octavian Paul
// Released under Public Domain
//
// translated to Delphi for use in Wren by Blue Hell
// For the modifications only:

{

  (C) COPYRIGHT 2019 .. 2022 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl
}


interface

uses

  Math,

  KnobsUtils, KnobsConversions,

  fftwgen, fftw_;


type

  TCalcRelativeFrequencies = reference to function( N: Integer): TSignal;

  {
    examples for RelFCalculator functions :

    function    TPadSynth.RelF1( N: Integer): TSignal;
    begin
      Result := N;
    end;


    function    TPadSynth.RelF2( N: Integer): TSignal;
    begin
      Result := N * ( 1.0 + 0.1 * N);
    end;
  }


  TPadSynth = class
  private
    FSize              : Integer;                    // Number of FFT points
    FSizeDiv2          : Integer;                    // Half the number of FFT points
    FSampleRate        : Integer;                    // Intended play rate for generated samples
    FRelFCalculator    : TCalcRelativeFrequencies;   // Reference to calculator for relative Partials frequencies
    FNumberPartials    : Integer;                    // The number of partials used
    FPartialAmplitudes : TSignalArray;               // Amplitudes for the partials
    FFreqAmp           : TSignalArray;               // Amplitude spectrum
    FFreqReal          : TSignalArray;               // Real      part for the pre-IFFT generated data
    FFreqImag          : TSignalArray;               // Imaginary part for the pre-IFFT generated data
    FFFTWPlanInv       : Pointer;                    // The IFFT plan for FFTW
    FPfftwData         : Pfftw_real;                 // The I/O  data for FFTW
  private
    function    GetPartialAmplitude( anIndex: Integer): TSignal;
    procedure   SetPartialAmplitude( anIndex: Integer; aValue: TSignal);
    procedure   IFFT( const aSamples: TSignalArray);
    function    Profile( fi, bwi: TSignal): TSignal;
  public
    constructor Create( aSize, aSampleRate, aNumberPartials: Integer; aRelFCalculator: TCalcRelativeFrequencies = nil);
    destructor  Destroy;                                                                                       override;
    procedure   Synthesize( aFreq, aBandwidth, aBandwidthScale: TSignal; const aSamples: TSignalArray);
  public
    property    PartialAmplitude[ anIndex: Integer]: TSignal                  read GetPartialAmplitude write SetPartialAmplitude;
    property    RelFCalculator                     : TCalcRelativeFrequencies read FRelFCalculator     write FRelFCalculator;
  end;


implementation


{ ========
  TPadSynth = class
  private
    FSize              : Integer;                    // Number of FFT points
    FSizeDiv2          : Integer;                    // Half the number of FFT points
    FSampleRate        : Integer;                    // Intended play rate for generated samples
    FRelFCalculator    : TCalcRelativeFrequencies;   // Reference to calculator for relative Partials frequencies
    FNumberPartials    : Integer;                    // The number of partials used
    FPartialAmplitudes : TSignalArray;               // Amplitudes for the partials
    FFreqAmp           : TSignalArray;               // Amplitude spectrum
    FFreqReal          : TSignalArray;               // Real      part for the pre-IFFT generated data
    FFreqImag          : TSignalArray;               // Imaginary part for the pre-IFFT generated data
    FFFTWPlanInv       : Pointer;                    // The IFFT plan for FFTW
    FPfftwData         : Pfftw_real;                 // The I/O  data for FFTW
  public
    property    PartialAmplitude[ anIndex: Integer]: TSignal                  read GetPartialAmplitude write SetPartialAmplitude;
    property    RelFCalculator                     : TCalcRelativeFrequencies read FRelFCalculator     write FRelFCalculator;
  private
}

    function    TPadSynth.GetPartialAmplitude( anIndex: Integer): TSignal;
    begin
      if   ( anIndex >  Low ( FPartialAmplitudes))      // First index is not used, 2nd index is the first partial
      and  ( anIndex <= High( FPartialAmplitudes))
      then Result := FPartialAmplitudes[ anIndex]
      else Result := 0.0;
    end;


    procedure   TPadSynth.SetPartialAmplitude( anIndex: Integer; aValue: TSignal);
    begin
      if   ( anIndex >  Low ( FPartialAmplitudes))      // First index is not used, 2nd index is the first partial
      and  ( anIndex <= High( FPartialAmplitudes))
      then FPartialAmplitudes[ anIndex] := aValue;
    end;


    procedure   TPadSynth.IFFT( const aSamples: TSignalArray);
    var
      i       : Integer;
      Maximum : TSignal;
    begin
      Pfftw_real_array( FPfftwData )^[ FSizeDiv2].v := 0.0;

      for i := 0 to FSizeDiv2 - 1
      do begin
        Pfftw_real_array( FPfftwData )^[ i].v := FFreqReal[ i];

        if   i <> 0
        then Pfftw_real_array( FPfftwData )^[ FSize - i].v := FFreqImag[ i];
      end;

      fftw_execute( FFFTWPlanInv);
      Maximum := 0.0;

      for i := 0 to FSize - 1
      do begin
        aSamples[ i] := Pfftw_real_array( FPfftwData)^[ i].v;

        if   Abs( aSamples[ i]) > Maximum
        then Maximum := Abs( aSamples[ i]);
      end;

      if   Maximum < 1e-5
      then Maximum := 1e-5;

      Maximum := 1 / ( Maximum * SQRT2);

      for i := Low( aSamples) to High( aSamples)
      do aSamples[ i] := aSamples[ i] * Maximum;
    end;




    function    TPadSynth.Profile( fi, bwi: TSignal): TSignal;
    var
      x : TSignal;
    begin
      if   bwi <  1e-5
      then bwi := 1e-5;

      x := fi / bwi;
      x := x * x;

      if   x > 14.71280603
      then Result := 0.0
      else Result := Exp( -x) / bwi;
    end;


  // public

    constructor TPadSynth.Create( aSize, aSampleRate, aNumberPartials: Integer; aRelFCalculator: TCalcRelativeFrequencies = nil);
    begin
      Assert(( aSize and 1) = 0); // Size should be a multiple of 2, preferrably a power even (fftw seems to not care too much there)

      FSize           := aSize;
      FSizeDiv2       := aSize div 2;
      FSampleRate     := aSampleRate;
      FNumberPartials := aNumberPartials;

      if   Assigned( aRelFCalculator)
      then FRelFCalculator := aRelFCalculator
      else FRelFCalculator := function( N: Integer): TSignal begin Result := 1.0 * N; end; // By default 1, 2, 3, 4 etc.

      SetLength( FPartialAmplitudes, FNumberPartials);       // SetLength initializes data to 0.0
      SetLength( FFreqAmp          , FSizeDiv2      );
      SetLength( FFreqReal         , FSizeDiv2      );
      SetLength( FFreqImag         , FSizeDiv2      );
      FPartialAmplitudes[ 1] := 1.0;                         // All other elements are zero, first index (0) not used
      FPfftwData             := fftw_malloc( FSize * SizeOf( Tfftw_real));
      FFFTWPlanInv           := fftw_plan_r2r_1d( FSize, FPfftwData, FPfftwData, FFTW_HC2R, FFTW_MEASURE);
    end;


    destructor  TPadSynth.Destroy; // override;
    begin
      if   Assigned( FPfftwData)
      then fftw_free( FPfftwData);

      if   Assigned( FFFTWPlanInv)
      then fftw_destroy_plan( FFFTWPlanInv);
    end;


    procedure   TPadSynth.Synthesize( aFreq, aBandwidth, aBandwidthScale: TSignal; const aSamples: TSignalArray);
    var
      i          : Integer;
      nh         : Integer;
      bw_HZ      : TSignal;
      bwi        : TSignal;
      fi         : TSignal;
      hProfile   : TSignal;
      RF         : TSignal;
      Phase      : TSignal;
      NumParts   : Integer;
    begin
      for i := Low( FFreqAmp) to High( FFreqAmp)                                                  // FFreqAmp not auto cleared, so do it
      do FFreqAmp[ i] := 0.0;

      // Build FFT image from partial specs

      if   FNumberPartials >= Floor( FSampleRate / aFreq)
      then NumParts := Floor( FSampleRate / aFreq) - 1
      else NumParts := FNumberPartials;

      for nh := 1 to NumParts - 1                                                                 // for each Partial
      do begin
        RF    := FRelFCalculator( nh);                                                            // Relative frequency for current partial
        bw_HZ := ( Power( 2.0, aBandWidth / 1200.0) - 1.0) * aFreq * Power( RF, aBandWidthScale); // bandwidth of current Partial in Hz
        bwi   := bw_HZ / ( 2.0 * FSamplerate);
        fi    := aFreq * RF / FSamplerate;

        for i := Low( FFreqAmp) to High( FFreqAmp)
        do begin
          hProfile     := Profile( i / FSize - fi, bwi);
          FFreqAmp[ i] := FFreqAmp[ i] + hProfile * PartialAmplitude[ nh];
        end;
      end;

      // Add random phases and 'complexify'.

      for i := Low( FFreqAmp) to High( FFreqAmp)
      do begin
        Phase         := Random * TWO_PI;
        FFreqReal[ i] := FFreqAmp[ i] * Cos( Phase);
        FFreqImag[ i] := FFreqAmp[ i] * Sin( Phase);
      end;

      // IFFT it & normalize & return samples to caller - from FFreqReal & FFreqImag, these are 'class global'.

      IFFT( aSamples);
    end;

end.

