unit anti_derivatives;

{
  Code modelled after :

    Proceedings of the 19th International Conference on Digital Audio Effects (DAFx-16),
    Brn, Czech Republic, Sept. 5-9 2016.

    Reducing the Aliasing of Nonlinear Waveshaping Using Continuous-Time Convolution.

      Julian D. Parker,
      Vadim Zavalishin
      Efflam Le Bivic

      Native Intstrumnets GmbH, Berlin, Germany

    Code for Lockhart model found at : https://github.com/danmohd/LockhartWavefolderJUCE - no coypyright info
    available.

      This is an implementation of 'Virtual Analog Model of the Lockhart Wavefolder' Conference Paper 2017 by
      Fabian Esqueda, Julian D. Parker, Henry Pintynen and Stefan Bilbao.

    Code made for use in the Wren Synth

  ======================================================================================================================

   COPYRIGHT 2021 .. 2022 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl
}

interface

uses

  System.SysUtils, System.Math,

  KnobsUtils;


type

  TIntegralBase = class
  strict private
  const
    EPSILON = 1e-6;
  private
    FValue1    : TSignal;
    FIntegral1 : TSignal;
  protected
    function    Value   ( aValue: TSignal): TSignal;                                                            virtual;
    function    Integral( aValue: TSignal): TSignal;                                                            virtual;
  public
    function    Process( aValue: TSignal): TSignal;
  end;


  TIntegralPoly< T: TIntegralBase, constructor> = class( TIntegralBase)
  private
    FLocked     : Boolean;
    FPolyphony  : Integer;
    FProcessors : TArray< T>;
  public
    constructor Create( aPolyphony: Integer);
    destructor  Destroy;                                                                                       override;
    function    Process( aVoice: Integer; aValue: TSignal): TSignal;
  end;


  TTanh = class( TIntegralBase)
  protected
    function    Value   ( aValue: TSignal): TSignal;                                                           override;
    function    Integral( aValue: TSignal): TSignal;                                                           override;
  end;


  THardClip = class( TIntegralBase)
  protected
    function    Value   ( aValue: TSignal): TSignal;                                                           override;
    function    Integral( aValue: TSignal): TSignal;                                                           override;
  end;


  TLockhart = class( TIntegralBase)
  strict private
  const
    R     = 15.0e3;
    VIs   =  1.0e-17;
    Vt    =  0.025864;
    Af    =  1.0;
    Nu    =  1.0;
    B1Rec = Nu * Vt;
    B1    =  1.0 / B1Rec;
    B2    =  2.0 * Af * B1 / R;
    D1    = VIs * B1;
    A1    =  2.0 * Af / R;
  private
    FRL : TSignal;
  protected
    function    Value   ( aValue: TSignal): TSignal;                                                           override;
    function    Integral( aValue: TSignal): TSignal;                                                           override;
  public
    property    RL: TSignal read FRL write FRL;
  end;


  function  LambertW( x: TSignal): TSignal;
  function  PolyLog2( x: TSignal): TSignal;


implementation



  function LambertW( x: TSignal): TSignal;
  const
    THRESH = 10.0e-7;
  var
    num : TSignal;
    den : TSignal;
    a   : TSignal;
    b   : TSignal;
    ia  : TSignal;
    w1  : TSignal;
    z   : TSignal;
    q   : TSignal;
    e   : TSignal;
    i   : Integer;
  begin
    // Upper branch of lambert-W function for non negative inputs.
    // Function adapted from F.Esqueda 2 / 10 / 17 based on implementation presented
    // by Darko Veberic - "Lambert W Function for Applications in Physics"

    Assert( x >= 0, 'only upper branch of Lambert-W function was implemented');

    if x = 0.0
    then Result := 0.0    // handle 0 separately
    else begin            // initial guesses
      if x < 0.14546954290661823
      then begin
        num    := 1.0 + ( 5.931375839364438 + ( 11.39220550532913 + (  7.33888339911111 + 0.653449016991959 * x) * x) * x) * x;
        den    := 1.0 + ( 6.931373689597704 + ( 16.82349461388016 + ( 16.43072324143226 + 5.115235195211697 * x) * x) * x) * x;
        Result := x * num / den;
      end
      else if x < 8.706658967856612
      then begin
        num    := 1.0 + ( 2.4450530707265568 + ( 1.3436642259582265 + ( 0.14844005539759195 + 0.0008047501729130  * x) * x) * x) * x;
        den    := 1.0 + ( 3.4447089864860025 + ( 3.2924898573719523 + ( 0.9164600188031222  + 0.05306864044833221 * x) * x) * x) * x;
        Result := x * num / den;
      end
      else begin
        if   x > 1e40
        then x := 1e40;

        a      := Ln( x);
        b      := Ln( a);
        ia     := 1.0 / a;
        Result := Normalize( a - b + ( b * ia) + 0.5 * b * ( b - 2.0) * ( ia * ia) + ( 1.0 / 6.0) * ( 2.0 * b * b - 9.0 * b + 6.0) * ( ia * ia * ia));

        // Possible typo in Fabian's MATLAB code. His has ... (b * ia) * 0.5 * ... instead of ... (b * ia) + 0.5 * ... in the above line.
      end;

      // Fritsch's iterations ( limited to 3 iterations or when error e is below the threshold, whichever comes first)

      for i := 1 to 3    // Originally 20 ... tested some with 2 ... it 'always' breaks out then .. soo .. choose 3 ..
      do begin           // maybe. Also raised TRESH from 10e-12 to 10e-7      //       |
        w1     := Result + 1.0;                                                //       |
        z      := Normalize( Ln( x) - Ln( Result) - Result);                   //       |
        q      := Normalize( 2.0 * w1 * ( w1 + ( 2.0 / 3.0) * z));             //       |
        e      := Normalize(( z / w1) * (( q - z) / ( q - 2.0 * z)));          //       |
        Result := Result * ( 1.0 + e);                                         //       |
                                                                               //       |
        if   Abs( e) < THRESH                                                  //       |
        then Break;                                                            //   <---+
      end;
    end;
  end;


  function  PolyLog2( x: TSignal): TSignal;
  //
  //  Source :: https://github.com/Expander/polylogarithm/
  //
  //  Part of Polylogarithm, MIT licensed
  //
  //  @brief Real dilogarithm \f$\mathrm{Li}_2(x)\f$
  //  @param x real argument
  //  @return \f$\mathrm{Li}_2(x)\f$
  //  @author Alexander Voigt
  //
  //  Implemented as an economized Pade approximation with a maximum error of 4.16e-18.
  //
  //  original function :: double li2(double x)
  //
  const
    Pa : array[ 0 .. 7] of TSignal =
    (
       1.0706105563309304277e+0,
      -4.5353562730201404017e+0,
       7.4819657596286408905e+0,
      -6.0516124315132409155e+0,
       2.4733515209909815443e+0,
      -4.6937565143754629578e-1,
       3.1608910440687221695e-2,
      -2.4630612614645039828e-4
    );
    Qa : array[ 0 .. 7] of TSignal =
    (
       1.0000000000000000000e+0,
      -4.5355682121856044935e+0,
       8.1790029773247428573e+0,
      -7.4634190853767468810e+0,
       3.6245392503925290187e+0,
      -8.9936784740041174897e-1,
       9.8554565816757007266e-2,
      -3.2116618742475189569e-3
    );
  var
    y  : TSignal;
    r  : TSignal;
    s  : TSignal;
    l  : TSignal;
    z  : TSignal;
    z2 : TSignal;
    z4 : TSignal;
    p  : TSignal;
    q  : TSignal;
  begin
     /// transform to [0, 1/2)

     if   x < -1
     then begin
        l := Ln( 1.0 - x);
        y := 1.0 / ( 1.0 - x);
        r := -PI * PI / 6.0 + l * ( 0.5 * l - Ln( -x));
        s := 1;
     end
     else if x = -1.0
     then Exit( -PI * PI / 12.0)
     else if x < 0.0
     then begin
        l := LnXP1( -x);
        y := x / ( x - 1.0);
        r := -0.5 * l * l;
        s := -1.0;
     end
     else if x = 0.0
     then Exit( 0.0)
     else if x < 0.5
     then begin
        y := x;
        r := 0.0;
        s := 1.0;
     end
     else if x < 1
     then begin
        y := 1.0 - x;
        r := PI * PI / 6.0 - Ln( x) * Ln( 1 - x);
        s := -1.0;
     end
     else if x = 1.0
     then Exit( PI * PI / 6.0)
     else if x < 2.0
     then begin
        l := Ln( x);
        y := 1.0 - 1.0 / x;
        r := PI * PI / 6.0 - l * ( Ln( 1.0 - 1.0 / x) + 0.5 * l);
        s := 1.0;
     end
     else begin
        l := Ln( x);
        y := 1.0 / x;
        r := PI * PI / 3.0 - 0.5 * l * l;
        s := -1.0;
     end;

     z  :=  y - 0.25;
     z2 :=  z *  z;
     z4 := z2 * z2;

     p := Pa[ 0] + z * Pa[ 1] + z2 * ( Pa[ 2] + z * Pa[ 3]) + z4 * ( Pa[ 4] + z * Pa[ 5] + z2 * ( Pa[ 6] + z * Pa[ 7]));
     q := Qa[ 0] + z * Qa[ 1] + z2 * ( Qa[ 2] + z * Qa[ 3]) + z4 * ( Qa[ 4] + z * Qa[ 5] + z2 * ( Qa[ 6] + z * Qa[ 7]));

     Result := r + s * y * p / q;
  end;




{ ========
  TIntegralBase = class
  private
    FValue1    : TSignal;
    FIntegral1 : TSignal;
  protected
}

    function    TIntegralBase.Value( aValue: TSignal): TSignal; // virtual;
    begin
      // Should be abstract .. but compiler complains on that ...
      Result := aValue;
    end;


    function    TIntegralBase.Integral( aValue: TSignal): TSignal; // virtual;
    begin
      // Should be abstract .. but compiler complains on that ...
      Result := aValue;
    end;


//  public

    function    TIntegralBase.Process( aValue: TSignal): TSignal;
    var
      NewIntegral : TSignal;
      Delta       : TSignal;
    begin
      NewIntegral := Integral( aValue);
      Delta       := Normalize( aValue - FValue1);

      if   Abs( Delta) < EPSILON
      then Result := Value(( aValue + FValue1) / 2.0)                      // + O( Delta * Delta) :: the order of the error
      else Result := Normalize(( NewIntegral - FIntegral1) / Delta);

      FValue1    := aValue;
      FIntegral1 := NewIntegral;
    end;


{ ========
  TIntegralPoly< T: TIntegralBase, constructor> = class
  private
    FPolyphony  : Integer;
    FProcessors : TArray< T>;
  public
}

    constructor TIntegralPoly< T>.Create( aPolyphony: Integer);
    var
      i : Integer;
      O : T;
    begin
      Assert( aPolyphony > 0);
      FLocked := True;
      inherited Create;
      SetLength( FPRocessors, aPolyphony);

      for i := 0 to aPolyphony - 1
      do FProcessors[ i] := T.Create;

      FPolyphony := aPolyphony;
      FLocked    := False;
    end;


    destructor  TIntegralPoly< T>.Destroy; // override;
    var
      i : Integer;
    begin
      FLocked := True;

      for i := 0 to FPolyphony - 1
      do FreeAndNil( FProcessors[ i]);

      inherited;
    end;


    function    TIntegralPoly< T>.Process( aVoice: Integer; aValue: TSignal): TSignal;
    begin
      Assert(( aVoice >= 0) and ( aVoice < FPolyphony));
      Result := FProcessors[ aVoice].Process( aValue);
    end;


{ ========
  TTanh = class( TIntegralBase)
  protected
}

    function    TTanh.Value( aValue: TSignal): TSignal; // override;
    begin
      Result := Tanh( aValue);
    end;


    function    TTanh.Integral( aValue: TSignal): TSignal; // override;
    begin
      if   Abs( aValue) > 100.0
      then Result := aValue - Ln( 2.0)
      else Result := Ln( Cosh( aValue));
    end;


{ ========
  THardClip = class( TIntegralBase)
  protected
}

    function    THardClip.Value( aValue: TSignal): TSignal; // override;
    begin
      if   ( aValue <= -1)
      or   ( aValue >= +1)
      then Result := Sign( aValue)
      else Result := aValue;
    end;


    function    THardClip.Integral( aValue: TSignal): TSignal; // override;
    begin
      if   ( aValue <= -1)
      or   ( aValue >= +1)
      then Result := aValue * Sign( aValue) - 0.5
      else Result := 0.5 * aValue * aValue;
    end;


{ ========
  TLockhart = class( TIntegralBase)
  strict private
  const
    R     = 15.0e3;
    VIs   =  1.0e-17;
    Vt    =  0.025864;
    Af    =  1.0;
    Nu    =  1.0;
    B1Rec = Nu * Vt;
    B1    =  1.0 / B1Rec;
    B2    =  2.0 * Af * B1 / R;
    D1    = VIs * B1;
    A1    =  2.0 * Af / R;
  private
    FRL : TSignal;
  public
    property    RL: TSignal read FRL write FRL;
  protected
}

    function    TLockhart.Value( aValue: TSignal): TSignal; // override;
    var
      L : TSignal;
    begin
      L      := Sign( aValue);
      Result := Normalize( A1 * RL * aValue - L * B1Rec * LambertW( D1 * RL * Exp( L * ( B1 + B2 * RL) * aValue)));
    end;


    function    TLockhart.Integral( aValue: TSignal): TSignal; // override;
     var
       psi : TSignal;
    begin
      psi    := LambertW( D1 * RL * Exp( Sign( aValue) * ( B1 + B2 * RL) * aValue));
      Result := Normalize( 0.5 * A1 * RL * aValue * aValue - ( B1Rec / ( 2.0 * ( B1 + B2 * RL))) * psi * ( psi + 2.0));
    end;





end.
