unit boids;

interface

uses

  System.Types, System.UITypes, System.SysUtils,

  Vcl.Graphics,

  KnobsUtils;

type

  TBoidsPoint3D = record
  public
    X : TSignal;
    Y : TSignal;
    Z : TSignal;
  public
    procedure  Clear;
    procedure  AddIn      ( const aPoint  : TBoidsPoint3D);
    procedure  SubtractOut( const aPoint  : TBoidsPoint3D);
    function   DistanceTo ( const aPoint  : TBoidsPoint3D): TSignal;
    procedure  DivideBy   ( const aScalar : TSignal );
    procedure  MultiplyBy ( const aScalar : TSignal );
    function   Norm : TSignal;
    procedure  LimitTo( aValue : TSignal);
    procedure  AddToX ( aValue : TSignal);
    procedure  AddToY ( aValue : TSignal);
    procedure  AddToZ ( aValue : TSignal);
  end;


  TBoidsPoints3D = array of TBoidsPoint3D;


  TBoidsBoid = record
    Position    : TBoidsPoint3D;
    OldPosition : TBoidsPoint3D;
    Velocity    : TBoidsPoint3D;
  end;

  TBoidsBoids         = array of TBoidsBoid;
  TBoidsFlocks        = array of TBoidsBoids;
  TBoidsBoidFollowMap = set of Byte;


  TBoidsSystem = class
  private
    FSystemSize    : Integer;
    FFlockSize     : Integer;
    FFlocks        : TBoidsFlocks;
    FBounds        : TBoidsPoint3D;
    FVLimit        : TSignal;
    FTargetSet     : Boolean;
    FAntiTargetSet : Boolean;
    FTarget        : TBoidsPoint3D;
    FFollows       : TBoidsPoints3D;
    FFollowMap     : TBoidsBoidFollowMap;
  private
    function    GetBoids( anIndex: Integer): TBoidsBoids;
    function    GetData: TBoidsFlocks;
    procedure   SetData ( const aData: TBoidsFlocks);
  private
    function    FlyToCenter1  ( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    function    KeepDistance  ( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    function    MatchSpeed    ( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    function    RandomizeSpeed( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    function    TrackTarget   ( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    function    FollowTarget  ( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
  public
    constructor Create( aSystemSize, aFlockSize: Integer; aBounds: TBoidsPoint3D);
    destructor  Destroy;                                                                                       override;
    procedure   Proceed;
    procedure   Follow( aPoint: TBoidsPoint3D; aFlockIndex: Integer; DoFollow: Boolean);
  public
    property    SystemSize                : Integer         read FSystemSize;
    property    FlockSize                 : Integer         read FFlockSize;
    property    Flocks[ anIndex: Integer] : TBoidsBoids     read GetBoids;
    property    VLimit                    : TSignal         read FVLimit        write FVLimit       ;
    property    Bounds                    : TBoidsPoint3D   read FBounds        write FBounds       ;
    property    TargetSet                 : Boolean         read FTargetSet     write FTargetSet    ;
    property    AntiTargetSet             : Boolean         read FAntiTargetSet write FAntiTargetSet;
    property    Target                    : TBoidsPoint3D   read FTarget        write FTarget       ;
    property    Data                      : TBoidsFlocks    read GetData        write SetData       ;
  end;


  TBoidsPainter = class
  private
    FBitmap : TBitmap;
  public
    constructor Create( aWidth, aHeight: Integer);
    destructor  Destroy;                                                                                       override;
    procedure   ShowBoidSystem( const aSystem: TBoidsSystem; const aCanvas: TCanvas);
  end;


  function  BoidsPoint( X, Y, Z: Integer): TBoidsPoint3D;



implementation



  function  BoidsPoint( X, Y, Z: Integer): TBoidsPoint3D;
  begin
    Result.X := X;
    Result.Y := Y;
    Result.Z := Z;
  end;



{ ========
  TBoidPoint = record
  public
    X : TSignal;
    Y : TSignal;
  public
}

    procedure  TBoidsPoint3D.Clear;
    begin
      X := 0.0;
      Y := 0.0;
      Z := 0.0;
    end;


    procedure  TBoidsPoint3D.AddIn( const aPoint: TBoidsPoint3D);
    begin
      X := X + aPoint.X;
      Y := Y + aPoint.Y;
      Z := Z + aPoint.Z;
    end;


    procedure  TBoidsPoint3D.SubtractOut( const aPoint: TBoidsPoint3D);
    begin
      X := X - aPoint.X;
      Y := Y - aPoint.Y;
      Z := Z - aPoint.Z;
    end;


    function   TBoidsPoint3D.DistanceTo( const aPoint: TBoidsPoint3D): TSignal;
    var
      dx : TSignal;
      dy : TSignal;
      dz : TSignal;
    begin
      dx     := X - aPoint.X;
      dy     := Y - aPoint.Y;
      dz     := Z - aPoint.Z;
      Result := Sqrt( dx * dx + dy * dy + dz * dz);
    end;


    procedure  TBoidsPoint3D.DivideBy( const aScalar: TSignal);
    begin
      X := X / aScalar;
      Y := Y / aScalar;
      Z := Z / aScalar;
    end;


    procedure  TBoidsPoint3D.MultiplyBy( const aScalar: TSignal);
    begin
      X := X * aScalar;
      Y := Y * aScalar;
      Z := Z * aScalar;
    end;


    function   TBoidsPoint3D.Norm: TSignal;
    begin
      Result := Sqrt( X * X + Y * Y + Z * Z);
    end;


    procedure  TBoidsPoint3D.LimitTo( aValue: TSignal);
    var
      N : TSignal;
    begin
      if  aValue > 0
      then begin
        N := Norm;

        if   N > aValue
        then MultiplyBy( aValue / N);
      end;
    end;


    procedure  TBoidsPoint3D.AddToX( aValue: TSignal);
    begin
      X := X + aValue;
    end;


    procedure  TBoidsPoint3D.AddToY( aValue: TSignal);
    begin
      Y := Y + aValue;
    end;


    procedure  TBoidsPoint3D.AddToZ( aValue: TSignal);
    begin
      Z := Z + aValue;
    end;


{ ========
  TBoidSystem = class
  private
    FBitmap        : TBitmap;
    FSystemSize    : Integer;
    FFlockSize     : Integer;
    FFlocks        : TFlocks;
    FBounds        : TBoidPoint;
    FVLimit        : TSignal;
    FTargetSet     : Boolean;
    FAntiTargetSet : Boolean;
    FTarget        : TBoidPoint;
    FFollows       : TBoidPoints;
    FFollowMap     : TBoidFollowMap;
    FUnFollowMap   : TBoidFollowMap;
  public
    property    VLimit          : TSignal    read FVLimit          write FVLimit         ;
    property    Bounds          : TBoidPoint read FBounds          write FBounds         ;
    property    BoidBorderColor : TColor     read FBoidBorderColor write FBoidBorderColor;
    property    BoidColor       : TColor     read FBoidColor       write FBoidColor      ;
    property    TargetSet       : Boolean    read FTargetSet       write FTargetSet      ;
    property    AntiTargetSet   : Boolean    read FAntiTargetSet   write FAntiTargetSet  ;
    property    Target          : TBoidPoint read FTarget          write FTarget         ;
  private
}

    function    TBoidsSystem.GetBoids( anIndex: Integer): TBoidsBoids;
    begin
      if   ( anIndex >= 0)
      and  ( anIndex <  FSystemSize)
      then Result := FFlocks[ anIndex]
      else Result := nil
    end;


    function    TBoidsSystem.GetData: TBoidsFlocks;
    begin
    end;


    procedure   TBoidsSystem.SetData ( const aData: TBoidsFlocks);
    begin
    end;


//  private

    function    TBoidsSystem.FlyToCenter1( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    // fly to center of mass
    var
      i : Integer;
      j : Integer;
      P : TBoidsPoint3D;
    begin
      Result.Clear;
      P     .Clear;

      if   FFlockSize > 1
      then begin
        for i := 0 to FSystemSize - 1
        do begin
          if i = aSysIndex                      // For 'self' group
          then begin
            for j := 0 to FFlockSize - 1
            do begin
              if   j <> aBoidIndex
              then Result.AddIn( FFlocks[ aSysIndex][ j].Position);
            end;
          end
          else begin                            // For other groups
            for j := 0 to FFlockSize - 1
            do P.AddIn( FFlocks[ i][ j].Position);
          end;
        end;

        Result.DivideBy( FFlockSize - 1);

        if   FSystemSize > 0
        then P.DivideBy(( FSystemSize - 1) * FFlockSize);

        P.SubtractOut( FFlocks[ aSysIndex][ aBoidIndex].Position);
        P.DivideBy( 50.0);
        Result.SubtractOut( FFlocks[ aSysIndex][ aBoidIndex].Position);
        Result.AddIn( P);
        Result.DivideBy   ( 950);
      end;
    end;


    function    TBoidsSystem.KeepDistance( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    // Keep away from other objects and/or boids
    var
      i : Integer;
      j : Integer;
    begin
      Result.Clear;

      for i := 0 to FSystemSize - 1
      do begin
        if   i = aSysIndex
        then begin
          for j := 0 to FFlockSize - 1
          do begin
            if   j <> aBoidIndex
            then begin
              if   FFlocks[ aSysIndex][ j].Position.DistanceTo( FFlocks[ aSysIndex][ aBoidIndex].Position) < 5
              then begin
                Result.SubtractOut( FFlocks[ aSysIndex][ j         ].Position);
                Result.AddIn      ( FFlocks[ aSysIndex][ aBoidIndex].Position);
              end;
            end;
          end;
        end
        else begin
          for j := 0 to FFlockSize - 1
          do begin
            if   FFlocks[ i][ j].Position.DistanceTo( FFlocks[ aSysIndex][ aBoidIndex].Position) < 5
            then begin
              Result.SubtractOut( FFlocks[ aSysIndex][ j         ].Position);
              Result.AddIn      ( FFlocks[ aSysIndex][ aBoidIndex].Position);
            end;
          end;
        end;
      end;

      Result.DivideBy( 2.0);
    end;


    function    TBoidsSystem.MatchSpeed( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    // Velocity matching with other boids
    var
      i : Integer;
    begin
      Result.Clear;

      if   FFlockSize > 1
      then begin
        for i := 0 to FFlockSize - 1
        do begin
          if   i <> aBoidIndex
          then Result.AddIn( FFlocks[ aSysIndex][ i].Velocity);
        end;

        Result.DivideBy   ( FFlockSize - 1);
        Result.Subtractout( FFlocks[ aSysIndex][ aBoidIndex].Velocity);
        Result.DivideBy( 12);
      end;
    end;


    function    TBoidsSystem.RandomizeSpeed( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    // Randomize speed somewhat
    begin
      Result.Clear;

      if   FFlockSize > 1
      then begin
        Result.X := 0.1 - 0.2 * Random;
        Result.Y := 0.1 - 0.2 * Random;
      end;
    end;


    function    TBoidsSystem.TrackTarget( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    // Make all boid flocks follow an external target
    begin
      Result.Clear;

      if   TargetSet
      or   AntiTargetSet
      then begin
        Result := Target;
        Result.SubtractOut( aBoid.Position);
        Result.DivideBy( 100);
      end;

      if   AntiTargetSet
      then Result.DivideBy( -1);
    end;


    function    TBoidsSystem.FollowTarget( const aBoid: TBoidsBoid; aSysIndex, aBoidIndex: Integer): TBoidsPoint3D;
    // Make selected boid flocks follow an external target
    begin
      Result.Clear;

      if   aSysIndex in FFollowMap
      then begin
        Result := FFollows[ aSysIndex];
        Result.SubtractOut( aBoid.Position);
        Result.DivideBy( 100);
      end;
    end;


//  public

    constructor TBoidsSystem.Create( aSystemSize, aFlockSize: Integer; aBounds: TBoidsPoint3D);
    var
      i : Integer;
      j : Integer;
      R : TSignal;
    begin
      SetLength( FFlocks , aSystemSize);
      SetLength( FFollows, aSystemSize);

      for i := 0 to aSystemSize - 1
      do SetLength( FFlocks[ i], aFlockSize);

      FBounds     := aBounds;
      FSystemSize := aSystemSize;
      FFlockSize  := aFlockSize;
      FVLimit     := 0.0;   // Do not limit velocities by default

      for i := 0 to FSystemSize - 1
      do begin
        for j := 0 to FFlockSize - 1
        do begin
          FFlocks[ i][ j].Velocity.X := 100.0 - 200.0 * Random;
          FFlocks[ i][ j].Velocity.Y := 100.0 - 200.0 * Random;
          R := Random;

          if   R <= 0.33
          then begin
            FFlocks[ i][ j].Position.Y := Random( Round( FBounds.Y));
            FFlocks[ i][ j].Position.Z := Random( Round( FBounds.Z));

            if   Random > 0.5
            then FFlocks[ i][ j].Position.X := - Random( 100)
            else FFlocks[ i][ j].Position.X := FBounds.X + Random( 100);
          end
          else if R < 0.67
          then begin
            FFlocks[ i][ j].Position.X := Random( Round( FBounds.X));
            FFlocks[ i][ j].Position.Z := Random( Round( FBounds.Z));

            if   Random > 0.5
            then FFlocks[ i][ j].Position.Y := - Random( 100)
            else FFlocks[ i][ j].Position.Y := FBounds.Y + Random( 100);
          end
          else begin
            FFlocks[ i][ j].Position.X := Random( Round( FBounds.X));
            FFlocks[ i][ j].Position.Y := Random( Round( FBounds.Y));

            if   Random > 0.5
            then FFlocks[ i][ j].Position.Z := - Random( 100)
            else FFlocks[ i][ j].Position.Z := FBounds.Z + Random( 100);
          end;
        end;
      end;
    end;


    destructor  TBoidsSystem.Destroy; // override;
    var
      aSystemSize : Integer;
      i           : Integer;
    begin
      aSystemSize := FSystemSize;
      FSystemSize := 0;
      FFlockSize  := 0;

      for i := 0 to aSystemSize - 1
      do SetLength( FFlocks[ i], 0);

      SetLength( FFlocks , 0);
      SetLength( FFollows, 0);
      inherited;
    end;


    procedure   TBoidsSystem.Proceed;
    var
      i      : Integer;
      j      : Integer;
      vel    : TBoidsPoint3D;
      boid   : TBoidsBoid;
      bounce : TSignal;
    begin
      for i := 0 to FSystemSize - 1
      do begin
        for j := 0 to FFlockSize - 1
        do begin
          boid := FFlocks[ i][ j];
          vel  := boid.Velocity;
          boid.OldPosition := boid.Position;

          vel.AddIn( FlyToCenter1  ( boid, i, j));
          vel.AddIn( KeepDistance  ( boid, i, j));
          vel.AddIn( MatchSpeed    ( boid, i, j));
          vel.AddIn( RandomizeSpeed( boid, i, j));
          vel.AddIn( TrackTarget   ( boid, i, j));
          vel.AddIn( FollowTarget  ( boid, i, j));

          vel.LimitTo( FVLimit);
          boid.Velocity := vel;
          boid.Position.AddIn( vel);
          bounce := 1.0 + 2.0 * Random;

          if   boid.Position.X < 0
          then boid.Velocity.AddToX( bounce)
          else if boid.Position.X > FBounds.X
          then boid.Velocity.AddToX( - bounce);

          if   boid.Position.Y < 0
          then boid.Velocity.AddToY( bounce)
          else if boid.Position.Y > FBounds.Y
          then boid.Velocity.AddToY( - bounce);

          if   boid.Position.Z < 0
          then boid.Velocity.AddToZ( bounce)
          else if boid.Position.Z > FBounds.Z
          then boid.Velocity.AddToZ( - bounce);

          FFlocks[ i][ j] := boid;
        end;
       end;
    end;


    procedure   TBoidsSystem.Follow( aPoint: TBoidsPoint3D; aFlockIndex: Integer; DoFollow: Boolean);
    // Select or unselect a particular flock to follow an external target
    begin
      if   DoFollow
      and  ( aFlockIndex >= 0          )
      and  ( aFlockIndex <  FSystemSize)
      then begin
        FFollows[ aFlockIndex] := aPoint;
        FFollowMap := FFollowMap + [ aFlockIndex];
      end
      else FFollowMap := FFollowMap - [ aFlockIndex];
    end;



{ ========
  TBoidViewer = class
  private
    FBitmap : TBitmap;
  public
}

    constructor TBoidsPainter.Create( aWidth, aHeight: Integer);
    begin
      FBitmap := TBitmap.Create;
      FBitmap.Width  := aWidth;
      FBitmap.Height := aHeight;
    end;


    destructor  TBoidsPainter.Destroy; // override;
    begin
      FreeAndNil( FBitmap)
    end;


    procedure   TBoidsPainter.ShowBoidSystem( const aSystem: TBoidsSystem; const aCanvas: TCanvas);
    const
      Colors : array[ 0 .. 3] of TColor = (
        clBlue  ,
        clRed   ,
        clGreen ,
        clPurple
      );
    var
      i : Integer;
      j : Integer;
      b : TBoidsPoint3D;
      S : TSignal;
    begin
      if  Assigned( aSystem)
      and Assigned( aCanvas)
      and Assigned( FBitmap)
      then begin
        FBitmap.Canvas.Pen  .Color := clWhite;
        FBitmap.Canvas.Brush.Color := clWhite;
        FBitmap.Canvas.Rectangle( 0, 0, FBitmap.Width, FBitmap.Height);

        for i := 0 to aSystem.SystemSize - 1
        do begin
          FBitmap.Canvas.Pen  .Color := Colors[ i mod 4];
          FBitmap.Canvas.Brush.Color := Colors[ i mod 4];

          for j := 0 to aSystem.FlockSize - 1
          do begin
            b := aSystem.Flocks[ i][ j].Position;

            if   aSystem.Bounds.Z >= 1
            then begin
              S := 6 * ( aSystem.Bounds.Z - b.Z) / aSystem.Bounds.Z;

              if   S < 2
              then S := 2;
            end
            else S := 6;

            FBitmap.Canvas.Rectangle( Round( b.X), Round( b.Y), Round( b.X + S), Round( b.Y + S));
          end;
        end;

        aCanvas.Draw( 0, 0, FBitmap);
      end;
    end;



end.
