unit gen_reader;

{
  A simple genome reader/player for A, C, G, T specified genomes. Any text file will do
  and all characters not in [ A, C, G, T, a, c, g, t] or just being skipped.


   COPYRIGHT 2021 .. 2022 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl
}

interface

uses

  System.SysUtils,

  IntSequences;


type

  TGenome = class( TBitArray)
  private
    FPosition : Integer;
  private
    function    GetBaseCount    : Integer;
    function    GetBasePosition : Integer;
  public
    procedure   ReadFromFile( const aFileName: string);
    procedure   Reset;
    function    NextBase: Byte;
    function    NextBaseSet( aSize: Integer): Integer;
  public
    property    BaseCount    : Integer read GetBaseCount;
    property    BasePosition : Integer read GetBasePosition;
  end;


implementation


{ ========
  TGenome = class( TBitArray)
  private
    FPosition : Integer;
  public
    property    BaseCount    : Integer read GetBaseCount;
    property    BasePosition : Integer read GetBasePosition;
  private
}

    function    TGenome.GetBaseCount: Integer;
    begin
      Result := BitCount div 2;
    end;


    function    TGenome.GetBasePosition: Integer;
    begin
      Result := FPosition;
    end;


//  public

    procedure   TGenome.ReadFromFile( const aFileName: string);

      procedure AddBit( aValue: Byte);
      begin
        Bit[ BitCount] := aValue;
      end;

    var
      aFile : TextFile;
      C     : Char;
    begin
      if   FileExists( aFileName)
      then begin
        AssignFile( aFile, aFileName);

        try
          System.Reset( aFile);

          while not Eof( aFile)
          do begin
            Read( aFile, C);

            if      CharInSet( C, [ 'w', 'W'])   // Weak -> A or T
            then begin
              if   Random >= 0.5
              then C := 'a'
              else C := 't'
            end
            else if CharInSet( C, [ 's', 'S'])   // Strong -> C or G
            then begin
              if   Random >= 0.5
              then C := 'c'
              else C := 'g'
            end
            else if CharInSet( C, [ 'm', 'M'])   // aMino -> A or C
            then begin
              if   Random >= 0.5
              then C := 'a'
              else C := 'c'
            end
            else if CharInSet( C, [ 'k', 'K'])   // Keto -> G or T
            then begin
              if   Random >= 0.5
              then C := 'g'
              else C := 't'
            end
            else if CharInSet( C, [ 'r', 'R'])   // puRine -> A or G
            then begin
              if   Random >= 0.5
              then C := 'a'
              else C := 'g'
            end
            else if CharInSet( C, [ 'y', 'Y'])   // pYrimidine -> C or T
            then begin
              if   Random >= 0.5
              then C := 'c'
              else C := 't'
            end
            else if CharInSet( C, [ 'b', 'B'])   // not A -> C, G or T
            then begin
              case Random( 3) of
                0 :  C := 'c';
                1 :  C := 'g';
                else C := 't';
              end;
            end
            else if CharInSet( C, [ 'd', 'D'])   // not C -> A, G or T
            then begin
              case Random( 3) of
                0 :  C := 'a';
                1 :  C := 'g';
                else C := 't';
              end;
            end
            else if CharInSet( C, [ 'h', 'H'])   // not G -> A, C or T
            then begin
              case Random( 3) of
                0 :  C := 'a';
                1 :  C := 'c';
                else C := 't';
              end;
            end
            else if CharInSet( C, [ 'v', 'V'])   // not ( U or T) -> A, C or G
            then begin
              case Random( 3) of
                0 :  C := 'a';
                1 :  C := 'c';
                else C := 'g';
              end;
            end
            else if CharInSet( C, [ 'n', 'N'])   // aNy -> A, C, G or T
            then begin
              case Random( 4) of
                0 :  C := 'a';
                1 :  C := 'c';
                2 :  C := 'g';
                else C := 't';
              end;
            end;

            case C of
              'a', 'A' : begin AddBit( 0); AddBit( 0); end;
              'c', 'C' : begin AddBit( 1); AddBit( 0); end;
              'g', 'G' : begin AddBit( 0); AddBit( 1); end;
              't', 'T' : begin AddBit( 1); AddBit( 1); end;
              'u', 'U' : begin AddBit( 1); AddBit( 1); end;
            end;
          end;
        finally
          CloseFile( aFile);
        end;
      end;

      Reset;
    end;


    procedure   TGenome.Reset;
    begin
      FPosition := 0;
    end;


    function    TGenome.NextBase: Byte;
    begin
      Result := Bit[ FPosition] + 2 * Bit[ FPosition + 1];
      Inc( FPosition, 2);

      if   FPosition > BitCount
      then Reset;
    end;


    function    TGenome.NextBaseSet( aSize: Integer): Integer;
    var
      i : Integer;
    begin
      Result   := 0;

      for i := 0 to aSize - 1
      do  Result := ( Result shl 2) + NextBase;
    end;


end.
