unit solresols;

{
   COPYRIGHT 2022 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl

  ===============================================================================

  The wiki blurp for the language solresol starts with:

  Solresol is a constructed language devised by Francois Sudre, beginning in
  1827.

  And then some more, see Wikipedia for the solresol lemma.

  What is implemented here ais a primitive english word to solresol translator,
  it takes no grammar into account and it has a pretty limited word set.

  The idea is to have a simple text to notes thingy, it would be cool to have the
  five symbol solresol words too ...
}

{$ifdef DEBUG}
  { $define TEST_SOLRESOLS}
{$endif}

interface

uses

  System.SysUtils, System.Classes, System.Generics.Collections,

  KnobsUtils;


type

  TSolReSolSymbols = TArray< Byte>;
  ESolResError     = class( Exception);

const

  SymDo = 0;      // Noraml values returned as part of a found word
  SymRe = 1;
  SymMi = 2;
  SymFa = 3;
  SymSo = 4;
  SymLa = 5;
  SymTi = 7;
  SymBk = 8;      // Special value which will end all sequences, also after SymErr
  SymEr = 9;      // Special value returned when a word was not found


  procedure InitSolReSols;
  procedure DoneSolReSols;
  function  ParseEnglish ( const aValue: string): TSolReSolSymbols;


implementation


type

  TWordList = TDictionary< string, TSolReSolSymbols>;


var

  WordList  : TWordList = nil;
  InitCount : Integer   = 0;


  procedure InitDictionary; forward;


  procedure InitSolReSols;
  begin
    if   InitCount = 0
    then InitDictionary;

    Inc( InitCount);
  end;


  procedure DoneSolReSols;
  begin
    if   InitCount > 0
    then Dec( InitCount);

    if   InitCount <= 0
    then begin
      FreeAndNil( WordList);
      InitCount := 0;
    end;
  end;


  function  ParseSpecs( const aSpecs: string): TSolReSolSymbols;
  var
    T : string;
    p : Integer;
    i : Integer;
  begin
    Result := [];
    p      := Length( aSpecs);
    Assert( not Odd( p), 'need an even amount of things here ...');
    p := p div 2;

    for i := 0 to p - 1
    do begin
      T := LowerCase( Copy( aSpecs, 2 * i + 1, 2));

      if      T = 'do' then Result := Result + [ SymDo]
      else if T = 're' then Result := Result + [ SymRe]
      else if T = 'mi' then Result := Result + [ SymMi]
      else if T = 'fa' then Result := Result + [ SymFa]
      else if T = 'so' then Result := Result + [ SymSo]
      else if T = 'la' then Result := Result + [ SymLa]
      else if T = 'ti' then Result := Result + [ SymTi]
      else raise ESolResError.Create( 'garbage ...');
    end;
  end;


  procedure AddItem( const aSpecs, aWords: string);
  var
    Words   : TStringList;
    Symbols : TSolReSolSymbols;
    i       : Integer;
  begin
    Symbols := ParseSpecs( aSpecs);
    Words   := Explode( aWords, ',');

    try
      for i := 0 to Words.Count - 1
      do WordList.AddOrSetValue( Trim( Words[ i]), Symbols);
    finally
      FreeAndNil( Words);
    end;
  end;


  procedure InitDictionary;
  begin
    WordList := TWordList.Create;

    // Ooooh .. oeh .. the debugger does not like this!
    //
    // funy .. when a late breakpoint is set it stops at the
    //
    //   'tisotifa', 'destroy, kill, annihilate, slaughter, massacre, murder, destruction, killing, extermination, carnage, destroyer, killer'
    //
    // item :) It then makes garbage in T in ParseSpecs on T := LowerCase( Copy( aSpecs, 2 * i + 1, 2));
    //
    // The debugger seems to run out of memory.
    //
    // But it seems ok as long as you do not try to debug past there, and it seems to run ok.

    AddItem( 'do'        , 'no, not, nor');
    AddItem( 're'        , 'and');
    AddItem( 'mi'        , 'or');
    AddItem( 'fa'        , 'at, to');
    AddItem( 'so'        , 'if');
    AddItem( 'la'        , 'the');
    AddItem( 'ti'        , 'yes, willingly');
    AddItem( 'dore'      , 'i, me, myself, personally, we, ourselves');
    AddItem( 'domi'      , 'you, yourself');
    AddItem( 'dofa'      , 'he, she, it, him, her, them, they');
    AddItem( 'doso'      , 'self, oneself');
    AddItem( 'dola'      , 'one, someone, another    person');
    AddItem( 'doti'      , 'other, another, different, alternative');
    AddItem( 'redo'      , 'my, mine');
    AddItem( 'remi'      , 'your, yours');
    AddItem( 'refa'      , 'his, her, its');
    AddItem( 'reso'      , 'our, ours');
    AddItem( 'rela'      , 'your, yours');
    AddItem( 'reti'      , 'their, theirs');
    AddItem( 'mido'      , 'for');
    AddItem( 'mire'      , 'that, which, who');
    AddItem( 'mifa'      , 'whose');
    AddItem( 'miso'      , 'well');
    AddItem( 'mila'      , 'voila, behold');
    AddItem( 'fado'      , 'what?');
    AddItem( 'fare'      , 'thatthose');
    AddItem( 'fami'      , 'this, these');
    AddItem( 'faso'      , 'here');
    AddItem( 'fala'      , 'good, tasty, delectable, exquisite, delicious');
    AddItem( 'fati'      , 'much, very, lots, many');
    AddItem( 'sodo'      , 'nothing, nil, naught, null');
    AddItem( 'sore'      , 'why, why?');
    AddItem( 'somi'      , 'wrong, evil');
    AddItem( 'sofa'      , 'because, for');
    AddItem( 'sola'      , 'always, perpetuate, perpetuately, constantly');
    AddItem( 'soti'      , 'thanks, thank');
    AddItem( 'lado'      , 'yesterday');
    AddItem( 'lare'      , 'today');
    AddItem( 'lami'      , 'tomorrow');
    AddItem( 'lafa'      , 'bad');
    AddItem( 'laso'      , 'never');
    AddItem( 'lati'      , 'of');
    AddItem( 'tido'      , 'how');
    AddItem( 'tire'      , 'each, every');
    AddItem( 'timi'      , 'hello');
    AddItem( 'tifa'      , 'little, scarcely');
    AddItem( 'tiso'      , 'mr, mister, sir');
    AddItem( 'tila'      , 'madame, ms, miss');
    AddItem( 'doredo'    , 'time');
    AddItem( 'doremi'    , 'day');
    AddItem( 'dorefa'    , 'week');
    AddItem( 'doreso'    , 'month');
    AddItem( 'dorela'    , 'year');
    AddItem( 'doreti'    , 'century');
    AddItem( 'domido'    , 'universe, creation');
    AddItem( 'domire'    , 'infinity, unlimited, endless, undefined, immensity, unbounded, immeasurable');
    AddItem( 'domifa'    , 'providence, providential, welfare');
    AddItem( 'domiso'    , 'god');
    AddItem( 'domila'    , 'eternize, perpetuate, eternal, eternity');
    AddItem( 'domiti'    , 'immortalize, immortal, indestructible, imperishable');
    AddItem( 'dofado'    , 'majesty, grandeur, magnitude, imposing, majestic');
    AddItem( 'dofare'    , 'supreme, sovreign');
    AddItem( 'dofami'    , 'magnificence, splendor, superb');
    AddItem( 'dofaso'    , 'marvel, wonder');
    AddItem( 'dofala'    , 'miracle, supernatural, miraculous');
    AddItem( 'dofati'    , 'adore, worship');
    AddItem( 'dosodo'    , 'pray');
    AddItem( 'dosore'    , 'believe');
    AddItem( 'dosomi'    , 'jesus');
    AddItem( 'dosofa'    , 'mary');
    AddItem( 'dosola'    , 'religion');
    AddItem( 'dosoti'    , 'sanctify');
    AddItem( 'dolado'    , 'charity, charitable, charitably');
    AddItem( 'dolare'    , 'philanthropy, philanthropist');
    AddItem( 'dolami'    , 'despise, contempt, contemptuous');
    AddItem( 'dolafa'    , 'sympathize, pity, compassion');
    AddItem( 'dolaso'    , 'inexcusable, unforgivable');
    AddItem( 'dolati'    , 'incomparable, invaluable, unparalleled');
    AddItem( 'dotido'    , 'help, aid, assist, rescue');
    AddItem( 'dotire'    , 'support, base');
    AddItem( 'dotimi'    , 'unfaithfulness, infidelity');
    AddItem( 'dotifa'    , 'accomplish, fulfill');
    AddItem( 'dotiso'    , 'duty, obligation');
    AddItem( 'dotila'    , 'noble, ennobled, dignity, worthy, just, pride');
    AddItem( 'redore'    , 'philosophy, philosopher, philosophical');
    AddItem( 'redomi'    , 'loathe, repugnance, repulsive');
    AddItem( 'redofa'    , 'moral, morally, moralize');
    AddItem( 'redoso'    , 'customs');
    AddItem( 'redola'    , 'proverb, maxim');
    AddItem( 'redoti'    , 'honor, honorable, venerable, respectable');
    AddItem( 'remido'    , 'limit, confine, border, circumscribe');
    AddItem( 'remire'    , 'merit, laudable, praised');
    AddItem( 'remifa'    , 'benevolence, well-doing');
    AddItem( 'remiso'    , 'alms, charity');
    AddItem( 'remila'    , 'give, gift, present');
    AddItem( 'remiti'    , 'inexplicable, indefinable, incomprehensible, inconceiveable, enigmatic');
    AddItem( 'refado'    , 'notability, significant, fame');
    AddItem( 'refare'    , 'dignities, honors, honorary');
    AddItem( 'refami'    , 'displease, unpleasant, uncomfortable');
    AddItem( 'refaso'    , 'indelible');
    AddItem( 'refala'    , 'privilege, prerogative');
    AddItem( 'refati'    , 'qualify, rights');
    AddItem( 'resodo'    , 'doubt, skepticism, skeptical');
    AddItem( 'resore'    , 'err, mistake, error, wrong');
    AddItem( 'resomi'    , 'heresy, heretic');
    AddItem( 'resofa'    , 'atheism, atheist');
    AddItem( 'resola'    , 'materialism, materialistic, materialize');
    AddItem( 'resoti'    , 'discourage, discouragement, tedious');
    AddItem( 'relado'    , 'misanthropy, misanthrope');
    AddItem( 'relare'    , 'antagonize, thwart, inconvenience, annoying, bother');
    AddItem( 'relami'    , 'distrust, suspect, beware, skittish');
    AddItem( 'relafa'    , 'impossible, impracticable, impossibility');
    AddItem( 'relaso'    , 'intolerance, severity, rigorism, inflexible, rigid');
    AddItem( 'relati'    , 'regret, sorry, repent, rue');
    AddItem( 'retido'    , 'mistreat, bully, maltreatment, brutality, rudeness');
    AddItem( 'retire'    , 'toughen, hard, hardness, tough');
    AddItem( 'retimi'    , 'spite, sulky, grumpiness, cantankerous, recalcitrant');
    AddItem( 'retifa'    , 'become');
    AddItem( 'retiso'    , 'sadness, gloom, melancholy, sad');
    AddItem( 'retila'    , 'bored, boredom');
    AddItem( 'midore'    , 'sympathy, inclination, sympathize');
    AddItem( 'midomi'    , 'attract, draw, appeal');
    AddItem( 'midofa'    , 'prefer');
    AddItem( 'midoso'    , 'trinity');
    AddItem( 'midola'    , 'reciprocity, mutuality, respective');
    AddItem( 'midoti'    , 'friend, friendly');
    AddItem( 'miredo'    , 'agree');
    AddItem( 'miremi'    , 'beloved, darling, dear');
    AddItem( 'mirefa'    , 'tenderness, kindness, loving, tenderly, fondness, endearment');
    AddItem( 'mireso'    , 'acknowledge, gratitude, recognition');
    AddItem( 'mirela'    , 'devote, focus, dedication');
    AddItem( 'mireti'    , 'link, bond, chain, fasten');
    AddItem( 'mifado'    , 'agreement, pleasure, enjoyment');
    AddItem( 'mifare'    , 'please, like, pleasant, seductive');
    AddItem( 'mifami'    , 'sigh, sighs');
    AddItem( 'mifaso'    , 'raise, elevate, uplift, high');
    AddItem( 'mifala'    , 'desire, wish, want, desirous');
    AddItem( 'mifati'    , 'incorrigible');
    AddItem( 'misomi'    , 'sweetness, pleasantness, smooth');
    AddItem( 'misofa'    , 'benevolence, affability, goodwill');
    AddItem( 'misola'    , 'equal, equality, equally');
    AddItem( 'misoti'    , 'amiability, kindness, kind, pleasant');
    AddItem( 'milado'    , 'appraise, estimate, consider');
    AddItem( 'milare'    , 'confide, confidence');
    AddItem( 'milami'    , 'passion, ardor, ardently');
    AddItem( 'milafa'    , 'idolize, idolatry');
    AddItem( 'milaso'    , 'love');
    AddItem( 'milati'    , 'cherish, lover, enamored');
    AddItem( 'mitido'    , 'fidelity, fathfulness, loyalty');
    AddItem( 'mitire'    , 'durable, stability, permanent, duration');
    AddItem( 'mitimi'    , 'engagement, fianc');
    AddItem( 'mitifa'    , 'marry, marriage, husband, wife, spouse');
    AddItem( 'mitiso'    , 'luck, fortune, fortunate, lucky, fortunately, luckily');
    AddItem( 'mitila'    , 'enjoy, enjoyment, pleasure, delirium, intoxicated, drunk');
    AddItem( 'fadore'    , 'immoral, immorality');
    AddItem( 'fadomi'    , 'letter, missive, epistle');
    AddItem( 'fadofa'    , 'date, era');
    AddItem( 'fadoso'    , 'sign, signature');
    AddItem( 'fadola'    , 'stamp, seal');
    AddItem( 'fadoti'    , 'end, complete, finish, final, finally, termination, achieve');
    AddItem( 'faredo'    , 'contact, address, mailing');
    AddItem( 'faremi'    , 'be, exist');
    AddItem( 'farefa'    , 'go, proceed');
    AddItem( 'fareso'    , 'lengthen, extend, extension, elongation');
    AddItem( 'farela'    , 'mail, post');
    AddItem( 'fareti'    , 'inexactitude, inaccuracy, inexact');
    AddItem( 'famido'    , 'free, exempt, enfranchisement, emancipate');
    AddItem( 'famire'    , 'evil, wicked, malignant, harmful, maleficence');
    AddItem( 'famifa'    , 'stamp, postmark');
    AddItem( 'famiso'    , 'have, possess, own, get');
    AddItem( 'famila'    , 'postman, courier, mail');
    AddItem( 'famiti'    , 'carry, bring, portable');
    AddItem( 'fasodo'    , 'cause, because');
    AddItem( 'fasore'    , 'theism, deism, theist');
    AddItem( 'fasomi'    , 'malice, unkindness, disparaging, malignancy');
    AddItem( 'fasofa'    , 'act, behave, conduct');
    AddItem( 'fasola'    , 'do, make, action, doable, commit');
    AddItem( 'fasoti'    , 'prepare, preparation');
    AddItem( 'falare'    , 'able, can, may, possible, practicability');
    AddItem( 'falami'    , 'means, resources, via');
    AddItem( 'falafa'    , 'understand, intellect');
    AddItem( 'falaso'    , 'disfavor, disrepute, disgrace');
    AddItem( 'falati'    , 'trend, tendency, inclination, disposition');
    AddItem( 'fatido'    , 'walk, step');
    AddItem( 'fatire'    , 'inquire, investigate');
    AddItem( 'fatimi'    , 'divorce, disown, repudiate, unmarry');
    AddItem( 'fatifa'    , 'want');
    AddItem( 'fatiso'    , 'cry, weep, sob, tears');
    AddItem( 'fatila'    , 'decide, determine, resolve, decision, resolution');
    AddItem( 'sodore'    , 'copy, transcribe, transcription');
    AddItem( 'sodomi'    , 'imitate, forge, mimic, emulate, counterfeit');
    AddItem( 'sodofa'    , 'example, model, type');
    AddItem( 'sodoso'    , 'translate, interpret, translation, interpreter');
    AddItem( 'sodola'    , 'comment, commentary, commentator');
    AddItem( 'sodoti'    , 'tradition, traditionally, traditional');
    AddItem( 'soredo'    , 'imply, implied');
    AddItem( 'soremi'    , 'ingratitude, disregarding, ungrateful');
    AddItem( 'sorefa'    , 'shorten, abbreviate, concise, terse, succinct, brief');
    AddItem( 'soreso'    , 'language, idiom, dialect, linguistics, philology');
    AddItem( 'sorela'    , 'dictionary, vocabulary, glossary');
    AddItem( 'soreti'    , 'irregularity, anomaly, irregular');
    AddItem( 'somido'    , 'devil, satan, lucifer, demon, satanic');
    AddItem( 'somire'    , 'verb, noun, adjective, adverb');
    AddItem( 'somifa'    , 'article');
    AddItem( 'somiso'    , 'pronoun');
    AddItem( 'somila'    , 'remember, recollection, commemorate, memorial, souvenir');
    AddItem( 'somiti'    , 'complicate, complication');
    AddItem( 'sofado'    , 'neutral, neutrality');
    AddItem( 'sofare'    , 'delete, erase, efface');
    AddItem( 'sofami'    , 'lower, diminish');
    AddItem( 'sofaso'    , 'intimidate, bully');
    AddItem( 'sofala'    , 'let, allow');
    AddItem( 'sofati'    , 'impatient, irritable, impatience, chafe');
    AddItem( 'solado'    , 'excuse, exonerate, pardon, pardonable, excusable');
    AddItem( 'solare'    , 'tolerate, bear, indulge, indulgence, tolerance');
    AddItem( 'solami'    , 'forgive, pardon, amnesty, mercy, forgiveness');
    AddItem( 'solafa'    , 'favor, favorite, favorable, prefer');
    AddItem( 'solaso'    , 'permit, allow, authorize, permission');
    AddItem( 'solati'    , 'climb, mount, rise, climbing, ascent');
    AddItem( 'sotido'    , 'run, gallop, race, runner');
    AddItem( 'sotire'    , 'thappiness, joy, playfulness, elation, happy, gay, joyful, perky, merrily, cheerfully, happily');
    AddItem( 'sotimi'    , 'misfortune, setback, calamity, disaster, unfortunate');
    AddItem( 'sotifa'    , 'laugh, hilarity, laughing');
    AddItem( 'sotiso'    , 'smile, grin, smiling');
    AddItem( 'sotila'    , 'skip, hop, jump, pounce, frolic');
    AddItem( 'ladore'    , 'college, institution, university, teacher, professor, headmaster');
    AddItem( 'ladomi'    , 'school, student, pupil, disciple');
    AddItem( 'ladofa'    , 'read, reading, reader, readable, legibly');
    AddItem( 'ladoso'    , 'book, volume, tome');
    AddItem( 'ladola'    , 'preface, foreword, preamble, introduction');
    AddItem( 'ladoti'    , 'word, term');
    AddItem( 'laredo'    , 'alphabet, alphabetical, alphabetically');
    AddItem( 'laremi'    , 'spell, spelling');
    AddItem( 'larefa'    , 'syllable, syllabic, monosyllable, monosyllabic');
    AddItem( 'lareso'    , 'lesson, instruction');
    AddItem( 'larela'    , 'phrase, sentence');
    AddItem( 'lareti'    , 'distraction, inattention, inadvertently, inattentive, carelessness, disapplied');
    AddItem( 'lamido'    , 'temporary, fleeting');
    AddItem( 'lamire'    , 'write, writing, writer');
    AddItem( 'lamifa'    , 'pen');
    AddItem( 'lamiso'    , 'forget, overlook, neglect, forgetful, omit');
    AddItem( 'lamila'    , 'ink, inkwell');
    AddItem( 'lamiti'    , 'difficulty, trouble, hardness');
    AddItem( 'lafado'    , 'paper, parchment');
    AddItem( 'lafare'    , 'page, leaf');
    AddItem( 'lafami'    , 'notebook, portfolio, laptop, pc, puter, computer, pad, ipad, phone, smartphone, handy');
    AddItem( 'lafaso'    , 'manuscript, scroll, writing');
    AddItem( 'lafala'    , 'text, textual, verbatim, to    the    letter');
    AddItem( 'lafati'    , 'subject, topic, thesis, theme');
    AddItem( 'lasodo'    , 'irreligion, irreligious, impious, impiety');
    AddItem( 'lasore'    , 'pencil, sketch');
    AddItem( 'lasomi'    , 'inequality, unequal, uneven');
    AddItem( 'lasofa'    , 'undo, undone');
    AddItem( 'lasola'    , 'align, line, row');
    AddItem( 'lasoti'    , 'draw, trace');
    AddItem( 'latido'    , 'point');
    AddItem( 'latire'    , 'fun, delighted, entertainment, amuse');
    AddItem( 'latimi'    , 'penalty, displeasure, tribulation, pain, punishment');
    AddItem( 'latifa'    , 'hesitate, irresolution, indecisive');
    AddItem( 'latiso'    , 'punctuate, punctuation');
    AddItem( 'latila'    , 'punctuation, point, comma');
    AddItem( 'tidore'    , 'dishonor, degrade, defile, humiliate');
    AddItem( 'tidomi'    , 'enmity, hostility, enemy, opponent, antagonist, hostile');
    AddItem( 'tidofa'    , 'start, begin, commence');
    AddItem( 'tidoso'    , 'study');
    AddItem( 'tidola'    , 'progress, progressive');
    AddItem( 'tidoti'    , 'learn, learning, education, apprenticeship');
    AddItem( 'tiredo'    , 'grammar, syntax, grammatically');
    AddItem( 'tiremi'    , 'spell, spelling, orthography, orthographic');
    AddItem( 'tirefa'    , 'accuracy, exactitude, punctuality, assiduity');
    AddItem( 'tireso'    , 'regular, symmetric, symmetry, correct');
    AddItem( 'tirela'    , 'attention, contemplation, careful');
    AddItem( 'tireti'    , 'apply, applied, studious, studiously');
    AddItem( 'timido'    , 'distinguish, differentiate, discern, difference, nuance');
    AddItem( 'timire'    , 'explain, define, demonstrate, expound, definition, explanation');
    AddItem( 'timifa'    , 'meaning, significance');
    AddItem( 'timiso'    , 'simplify, simple, elementary');
    AddItem( 'timila'    , 'facilitate, aid, easy');
    AddItem( 'timiti'    , 'clarify, lighten, elucidate, explicit, clarity, conceivable');
    AddItem( 'tifado'    , 'dictate, dictation');
    AddItem( 'tifare'    , 'fault, defect, failure, defective, failed');
    AddItem( 'tifami'    , 'correct, reform, correction, reformation');
    AddItem( 'tifaso'    , 'wait, patience, patient');
    AddItem( 'tifala'    , 'repeat, recapitulate');
    AddItem( 'tifati'    , 'reiterate, renew, again');
    AddItem( 'tisodo'    , 'compete, contest');
    AddItem( 'tisore'    , 'encourage, embolden, stimulate');
    AddItem( 'tisomi'    , 'inspect, review, inspector, inspection');
    AddItem( 'tisofa'    , 'review, scrutinize, investigate');
    AddItem( 'tisola'    , 'analyze, analysis, analyst, analytical');
    AddItem( 'tisoti'    , 'question, interrogate, query');
    AddItem( 'tilado'    , 'compare, comparison, similarity, similitude, comparable');
    AddItem( 'tilare'    , 'crown, tiara, coronation, crowned');
    AddItem( 'tilami'    , 'hate, detest, abhor, hatred, animosity, aversion');
    AddItem( 'tilafa'    , 'unable, unfit, incapable, inept');
    AddItem( 'tilaso'    , 'down, descent');
    AddItem( 'tilati'    , 'medal, medallion');
    AddItem( 'doredore'  , 'birth, born');
    AddItem( 'doredomi'  , 'body, physical');
    AddItem( 'doredofa'  , 'head, brain, skull');
    AddItem( 'doredoso'  , 'hair, eyebrows, mustache, beard, eyelashes');
    AddItem( 'doredola'  , 'face');
    AddItem( 'doredoti'  , 'cheeks');
    AddItem( 'doremido'  , 'forehead, temple, front');
    AddItem( 'doremire'  , 'eyes, eyelid, eye');
    AddItem( 'doremifa'  , 'nose, nostril, nasal');
    AddItem( 'doremiso'  , 'mouth, lips, tongue, palate');
    AddItem( 'doremila'  , 'teeth, dental, dentition');
    AddItem( 'doremiti'  , 'chin');
    AddItem( 'dorefado'  , 'lavish, wasteful, prodigality');
    AddItem( 'dorefare'  , 'neck, throat, larynx');
    AddItem( 'dorefami'  , 'chest, breast');
    AddItem( 'dorefaso'  , 'shoulder');
    AddItem( 'dorefala'  , 'arm, elbow, wrist');
    AddItem( 'dorefati'  , 'hand, palm');
    AddItem( 'doresodo'  , 'disrupt, disturb, upset, upheaval, disturbance');
    AddItem( 'doresore'  , 'servant, maid, valet, chambermaid');
    AddItem( 'doresomi'  , 'stomach');
    AddItem( 'doresofa'  , 'abdomen, belly');
    AddItem( 'doresola'  , 'intestines, entrails, guts, intestinal');
    AddItem( 'doresoti'  , 'liver, spleen');
    AddItem( 'dorelado'  , 'back, backbone, spine');
    AddItem( 'dorelare'  , 'kidneys');
    AddItem( 'dorelami'  , 'hips, side');
    AddItem( 'dorelafa'  , 'unapproachable, inaccessible');
    AddItem( 'dorelaso'  , 'buttocks, bottom, behind, anus');
    AddItem( 'dorelati'  , 'skin, flesh, epidermis');
    AddItem( 'doretido'  , 'bone, ossification');
    AddItem( 'doretire'  , 'marrow');
    AddItem( 'doretimi'  , 'artery, vein');
    AddItem( 'doretifa'  , 'chyle');
    AddItem( 'doretiso'  , 'blood, bloody');
    AddItem( 'doretila'  , 'circulate, circulation');
    AddItem( 'domidore'  , 'members');
    AddItem( 'domidomi'  , 'thigh, knee');
    AddItem( 'domidofa'  , 'leg, calf');
    AddItem( 'domidoso'  , 'foot, heel');
    AddItem( 'domidola'  , 'finger, thumb');
    AddItem( 'domidoti'  , 'fingernail, nail');
    AddItem( 'domiremi'  , 'see, vision, visual');
    AddItem( 'domirefa'  , 'touch, feel, handle');
    AddItem( 'domireso'  , 'taste');
    AddItem( 'domirela'  , 'smell');
    AddItem( 'domireti'  , 'hear, hearing, auditory, audible');
    AddItem( 'domifado'  , 'humanity, mankind, man, woman, male, female');
    AddItem( 'domifare'  , 'live, exist, be, life, existence');
    AddItem( 'domifami'  , 'breathe, breath, breathing');
    AddItem( 'domifaso'  , 'child, children, jouvenile, adolescent');
    AddItem( 'domifala'  , 'grow, increase, growth, stature, size');
    AddItem( 'domifati'  , 'strength, force, vigour, virility, strong, manly');
    AddItem( 'domisodo'  , 'superiority, supreme, elite');
    AddItem( 'domisore'  , 'power, authority, preeminence, rule');
    AddItem( 'domisomi'  , 'grandeur, magnitude, magnanimity, great');
    AddItem( 'domisofa'  , 'intelligence, capacity, smart, intellect');
    AddItem( 'domisola'  , 'spirit, spiritual');
    AddItem( 'domisoti'  , 'soul');
    AddItem( 'domilado'  , 'talk, utter, speak, speech, speaking, verbal, speaker, verbally, orally');
    AddItem( 'domilare'  , 'pronounce, articulate, pronunciation, articulation');
    AddItem( 'domilami'  , 'argue, reason, confer');
    AddItem( 'domilafa'  , 'rationality, reason, sense, reasonable');
    AddItem( 'domilaso'  , 'discern, discernment, judgement, dialectic, logic, sensible, wisely, logician, judicious');
    AddItem( 'domilati'  , 'insight, perspicacity, sagacity, clairvoyance, discerning, sharpness');
    AddItem( 'domitido'  , 'skill, ability, know-how, competence, clever, adroit, expert, skillful');
    AddItem( 'domitire'  , 'pretend, claim, pretense');
    AddItem( 'domitimi'  , 'liberate, release, free, freedom, independence');
    AddItem( 'domitifa'  , 'energy, energetic, force');
    AddItem( 'domitiso'  , 'firmness, stoicism, strong, unshakeable, fortitude, hardiness');
    AddItem( 'domitila'  , 'confront, courage, bravery, courageous, valiant, intrepid');
    AddItem( 'dofadore'  , 'wait, pending, expectation, until');
    AddItem( 'dofadomi'  , 'hope');
    AddItem( 'dofadofa'  , 'know, knowledge, cognizable');
    AddItem( 'dofadoso'  , 'relating, related, relationships');
    AddItem( 'dofadola'  , 'social, sociability, sociably, sociable');
    AddItem( 'dofadoti'  , 'educate, rear');
    AddItem( 'dofaredo'  , 'stinginess, parsimony, miser, sordid, squalor, scant, avarice');
    AddItem( 'dofaremi'  , 'save, economy, savings, economic, economically');
    AddItem( 'dofarefa'  , 'prosper, prosperity, prosperous, flourishing');
    AddItem( 'dofareso'  , 'ease, comfortable, ease, wellbeing, comfortably');
    AddItem( 'dofarela'  , 'enrich, wealth, rich, fortune, opulence, opulent');
    AddItem( 'dofareti'  , 'luxury, deluxe, magnificence, sumptuous, pompous, luxurious, prosperous');
    AddItem( 'dofamido'  , 'feelings, principles');
    AddItem( 'dofamire'  , 'qualify, qualification, quality, attribute, qualifier');
    AddItem( 'dofamifa'  , 'virtue, virtuous, virtuously');
    AddItem( 'dofamiso'  , 'wisdom, wise, sage, wisely');
    AddItem( 'dofamila'  , 'modesty, chastity, modest, chaste');
    AddItem( 'dofamiti'  , 'purified, purity, sweetness, pure, sweet, intact, irreproachable, blameless');
    AddItem( 'dofasodo'  , 'innocence, innocent, innocently');
    AddItem( 'dofasore'  , 'naivety, candor, simplicity, candid, nave, simply');
    AddItem( 'dofasomi'  , 'beauty, lovely, pretty, beautiful');
    AddItem( 'dofasofa'  , 'grace, graceful, graciously');
    AddItem( 'dofasola'  , 'charm, enchant, attractions, charms, charming, attractive, enchanting, delightful');
    AddItem( 'dofasoti'  , 'truth, reality, true, veracity, truthful, real, effective, really, actually');
    AddItem( 'dofalado'  , 'frankness, honesty, sincerity, frank, sincere, truly');
    AddItem( 'dofalare'  , 'goodness, meekness, benignity');
    AddItem( 'dofalami'  , 'sensitivity, impressionability, sensitive, impressionable');
    AddItem( 'dofalafa'  , 'heartfelt, warmth, effusion, profusely');
    AddItem( 'dofalaso'  , 'generosity, liberality, munificence, liberal, generous');
    AddItem( 'dofalati'  , 'humanity, helpful, humanitarian, humanly');
    AddItem( 'dofatido'  , 'prudence, caution, carefully, circumspection');
    AddItem( 'dofatire'  , 'predict, forecast, foresee, anticipate, foresight');
    AddItem( 'dofatimi'  , 'discretion, discrete, reserved');
    AddItem( 'dofatifa'  , 'consiousness, awareness, conscientous');
    AddItem( 'dofatiso'  , 'honesty, integrity, righteousness, integrity, rectitude, incorruptibility');
    AddItem( 'dofatila'  , 'delicacy, tact, delicate, gently');
    AddItem( 'dosodore'  , 'appeal, evoke');
    AddItem( 'dosodomi'  , 'ask, request, demand');
    AddItem( 'dosodofa'  , 'open, opening, ajar, gaping, opener');
    AddItem( 'dosodoso'  , 'enter, penetrate, entry');
    AddItem( 'dosodola'  , 'inside, in, internal, internally');
    AddItem( 'dosodoti'  , 'essence, innate, fundamentally, intrinsically, inherently');
    AddItem( 'dosoredo'  , 'arrange, arrangement, order, ordered');
    AddItem( 'dosoremi'  , 'experience, experienced');
    AddItem( 'dosorefa'  , 'regulate, regularize, rule, discipline, regularity, statute');
    AddItem( 'dosoreso'  , 'occupy, occupation, absorbed, occupied');
    AddItem( 'dosorela'  , 'ttribution, attributive');
    AddItem( 'dosoreti'  , 'accidental, fortuitous');
    AddItem( 'dosomido'  , 'inferiority, lower, secondary, subordinate');
    AddItem( 'dosomire'  , 'watch, standby, vigil, watchfulness');
    AddItem( 'dosomifa'  , 'light, match, igniter, lighter');
    AddItem( 'dosomiso'  , 'candle, wax');
    AddItem( 'dosomila'  , 'gas, gasometer');
    AddItem( 'dosomiti'  , 'illuminate, lighting, illumination, clear');
    AddItem( 'dosofado'  , 'guilt, culpability, guilty, culprit, reprehensible, blameworthy, reproachable, punishable');
    AddItem( 'dosofare'  , 'awaken, arouse, awake');
    AddItem( 'dosofami'  , 'rise, stand, standing, raised, lifted');
    AddItem( 'dosofaso'  , 'appear, appearance');
    AddItem( 'dosofala'  , 'come, approach, hasten, coming');
    AddItem( 'dosofati'  , 'salute, bow, greet, hello, greetings');
    AddItem( 'dosolado'  , 'listen, hearken, listener, earpiece, earphone');
    AddItem( 'dosolare'  , 'rid, unload, riddance, discharge, unloading');
    AddItem( 'dosolami'  , 'compliment, commend, praise, congratulation, compliment, laud, laudatory');
    AddItem( 'dosolafa'  , 'exaggerate, amplify, magnify, exaggeration, amplification');
    AddItem( 'dosolaso'  , 'flatter, adulate, fawn, flattery, adulation, sycophancy, flattering');
    AddItem( 'dosolati'  , 'complacent, condescend, complacency, condescension, patronizing');
    AddItem( 'dosotido'  , 'advise, urge, council, exhortation, advice');
    AddItem( 'dosotire'  , 'remonstrate, remonstrance');
    AddItem( 'dosotimi'  , 'suggest, insinuate, insinuation, innuendo, allusion');
    AddItem( 'dosotifa'  , 'lead, guide, conduct, drive, mentor, cicerone');
    AddItem( 'dosotiso'  , 'take, carry, bring');
    AddItem( 'dosotila'  , 'hasten, rush, willingness, enthusiasm, readiness, eager');
    AddItem( 'doladore'  , 'invite, summon, call, invitation');
    AddItem( 'doladomi'  , 'accept, acquiesce, adhere, consent, aggree, acceptance');
    AddItem( 'doladofa'  , 'put, lay, affix, put, set, place, apposition');
    AddItem( 'doladoso'  , 'cover');
    AddItem( 'doladola'  , 'table');
    AddItem( 'doladoti'  , 'tablecloth, napkin');
    AddItem( 'dolaredo'  , 'tableware, dishes, plate, dish, cup, vase');
    AddItem( 'dolaremi'  , 'fork');
    AddItem( 'dolarefa'  , 'spoon, scoop, spoonful');
    AddItem( 'dolareso'  , 'cup, glass, goblet, mug, beaker');
    AddItem( 'dolarela'  , 'bottle, jar, carafe, flask');
    AddItem( 'dolareti'  , 'wine, champagne');
    AddItem( 'dolamido'  , 'silence, shutup, hush, taciturn, silently');
    AddItem( 'dolamire'  , 'water, wet');
    AddItem( 'dolamifa'  , 'thirsty, thirst, parched');
    AddItem( 'dolamiso'  , 'drink, drinker');
    AddItem( 'dolamila'  , 'hunger, appetite, starvation');
    AddItem( 'dolamiti'  , 'eat, eating, eater, eatable');
    AddItem( 'dolafado'  , 'hide, conceal, dissimulate, sneaky');
    AddItem( 'dolafare'  , 'feed, nourish, food, substantial, nourishing, feeder, nurse');
    AddItem( 'dolafami'  , 'soup, broth');
    AddItem( 'dolafaso'  , 'pasta, macaroni, vermicelli, &c.');
    AddItem( 'dolafala'  , 'bread, loaf, crumb, crust');
    AddItem( 'dolafati'  , 'fish, oysters, eels, salmon');
    AddItem( 'dolasodo'  , 'meat, steak, beef');
    AddItem( 'dolasore'  , 'roast, roasting, toast, grill, broil');
    AddItem( 'dolasomi'  , 'poultry, pigeon, chicken, duck, turkey');
    AddItem( 'dolasofa'  , 'game, partridge, hare, roe');
    AddItem( 'dolasola'  , 'pastry, cake, meringue, pie');
    AddItem( 'dolasoti'  , 'pork, ham');
    AddItem( 'dolatido'  , 'salt, salted, salting');
    AddItem( 'dolatire'  , 'pepper, spice');
    AddItem( 'dolatimi'  , 'vinegar, vinaigrette');
    AddItem( 'dolatifa'  , 'north, northern, northward');
    AddItem( 'dolatiso'  , 'oil, oily, lubricate');
    AddItem( 'dolatila'  , 'mustard');
    AddItem( 'dotidore'  , 'suffice, sufficiency, sufficient, enough, necessary');
    AddItem( 'dotidomi'  , 'vegetables, peas, asparagus, spinach, artichokes');
    AddItem( 'dotidofa'  , 'salad, lettuce');
    AddItem( 'dotidoso'  , 'season, flavor, dress, seasoning, dressing');
    AddItem( 'dotidola'  , 'juice, sauce, juicy');
    AddItem( 'dotidoti'  , 'eggs, omelette');
    AddItem( 'dotiredo'  , 'dairy, milk, cream, creamy, dairyman');
    AddItem( 'dotiremi'  , 'butter');
    AddItem( 'dotirefa'  , 'cheese');
    AddItem( 'dotireso'  , 'dessert, hazelnuts, nuts, almonds, chestnuts, walnuts, hazel, hazelnut, chestnut, almond');
    AddItem( 'dotirela'  , 'fruit, strawberries, raspberries, currants, cherries, cherry, fruits');
    AddItem( 'dotireti'  , 'grapes, peaches, pears, apples, plums, plum, apple, pear, peach');
    AddItem( 'dotimido'  , 'inability, awkwardness, clumsiness, incompetence, clumsy');
    AddItem( 'dotimire'  , 'pour, spill');
    AddItem( 'dotimifa'  , 'liquors');
    AddItem( 'dotimiso'  , 'spirits, alcohol, brandy, rum, absinthe, wiskey, whisky, cognac, calvados, moonshine');
    AddItem( 'dotimila'  , 'withdraw, disengage, retract, withdrawal');
    AddItem( 'dotimiti'  , 'drunk, drunken, drunkenness, drunkard');
    AddItem( 'dotifado'  , 'imprudence, temerity, recklessness, rashness, indiscretion');
    AddItem( 'dotifare'  , 'beer, wing, wear');
    AddItem( 'dotifami'  , 'refresh, refreshing');
    AddItem( 'dotifaso'  , 'caf, koffieshop');
    AddItem( 'dotifala'  , 'tea, teapot');
    AddItem( 'dotifati'  , 'sweeten, sugar, candy');
    AddItem( 'dotisodo'  , 'chocolate, cocoa');
    AddItem( 'dotisore'  , 'uselessness, redundancy, superfluous, useless');
    AddItem( 'dotisomi'  , 'meal, feast, banquet, lunch, dinner, supper');
    AddItem( 'dotisofa'  , 'cook, cooking, kitchen');
    AddItem( 'dotisola'  , 'provide, supply, fill, supplies, provisions, provider');
    AddItem( 'dotisoti'  , 'food, edible');
    AddItem( 'dotilado'  , 'fragment, cutting, piece, bit, slice');
    AddItem( 'dotilare'  , 'empty, void, dry, blank');
    AddItem( 'dotilami'  , 'appetizer');
    AddItem( 'dotilafa'  , 'reserve, keep, book, conserve, reserved, conservative');
    AddItem( 'dotilaso'  , 'buffet');
    AddItem( 'dotilati'  , 'pantry');
    AddItem( 'redoredo'  , 'clothes, outfit, effects');
    AddItem( 'redoremi'  , 'generalize, universalize, generality, universality, general, universal, generally');
    AddItem( 'redorefa'  , 'shirt');
    AddItem( 'redoreso'  , 'sock, stocking');
    AddItem( 'redorela'  , 'shoes, boots, slippers');
    AddItem( 'redoreti'  , 'pants, knickers, breeches');
    AddItem( 'redomido'  , 'clean, lather, ablution, soap');
    AddItem( 'redomire'  , 'shave, razor');
    AddItem( 'redomifa'  , 'mirror, reflected');
    AddItem( 'redomiso'  , 'comb');
    AddItem( 'redomila'  , 'decrease, subtract, deduct, mitigate, reduce, discounts, reduction');
    AddItem( 'redomiti'  , 'coif, hairdresser, wigmaker, wig');
    AddItem( 'redofado'  , 'unexpected, surprising, unforeseen, unexpectedly');
    AddItem( 'redofare'  , 'move, dislodge, displace, moving');
    AddItem( 'redofami'  , 'tie, necktie, scarf');
    AddItem( 'redofaso'  , 'vest, camisole');
    AddItem( 'redofala'  , 'button, buttonhole');
    AddItem( 'redofati'  , 'clothe, dress, clothing, apparel, costume, suit');
    AddItem( 'redosodo'  , 'coat, overcoat, cloak, jacket');
    AddItem( 'redosore'  , 'garnish, trim, trimmed, lined');
    AddItem( 'redosomi'  , 'fur');
    AddItem( 'redosofa'  , 'glove, gloves, mittens');
    AddItem( 'redosola'  , 'cane, rod, stick, baton');
    AddItem( 'redosoti'  , 'umbrella, parasol');
    AddItem( 'redolado'  , 'toilet, toiletries, personal');
    AddItem( 'redolare'  , 'corset, bodice');
    AddItem( 'redolami'  , 'lace');
    AddItem( 'redolafa'  , 'skirt, underskirt');
    AddItem( 'redolaso'  , 'dress, robe');
    AddItem( 'redolati'  , 'inevitable, unavoidable, immenent, inevitably');
    AddItem( 'redotido'  , 'insufficiency, inadequacy, inufficient');
    AddItem( 'redotire'  , 'adorn, decorate, embelish, prettify, ornament, embellishment');
    AddItem( 'redotimi'  , 'invisibility, imperceptibility, invisible, imperceptible, unseen');
    AddItem( 'redotifa'  , 'ring');
    AddItem( 'redotiso'  , 'bracelet, wristband');
    AddItem( 'redotila'  , 'glasses, eyeglasses, spectacles');
    AddItem( 'remidore'  , 'shawl, scarf');
    AddItem( 'remidomi'  , 'depositary');
    AddItem( 'remidofa'  , 'brooch, clasp, pin, staples');
    AddItem( 'remidoso'  , 'hat, cap, hood, bonnet');
    AddItem( 'remidola'  , 'mask, veil, veiled');
    AddItem( 'remidoti'  , 'elegance, elegant, elegantly');
    AddItem( 'remiredo'  , 'walk, trot, march, hike, walker, hiker');
    AddItem( 'remiremi'  , 'walk, stroll, wander, promenade');
    AddItem( 'remirefa'  , 'car, carriage, coupe, bus');
    AddItem( 'remireso'  , 'visit, tour, visitor, guest');
    AddItem( 'remirela'  , 'ticket');
    AddItem( 'remireti'  , 'society, worldly, mundane');
    AddItem( 'remifado'  , 'defect, defective, fault');
    AddItem( 'remifare'  , 'property, domain, owner');
    AddItem( 'remifami'  , 'concierge, porter, janitor, caretaker, doorman, gatekeeper');
    AddItem( 'remifaso'  , 'rent, rental, lease, renter');
    AddItem( 'remifala'  , 'home, house, hut, cottage, hotel');
    AddItem( 'remifati'  , 'live, reside, housing, residence, inhabitant, resident, residential');
    AddItem( 'remisodo'  , 'sleep, doze, drowsiness, sleeper, asleep, dormant, zzz');
    AddItem( 'remisore'  , 'vestibule, lobby, hall, threshold');
    AddItem( 'remisomi'  , 'door, gate');
    AddItem( 'remisofa'  , 'antechamber');
    AddItem( 'remisola'  , 'room, lounge');
    AddItem( 'remisoti'  , 'office');
    AddItem( 'remilado'  , 'bedroom');
    AddItem( 'remilare'  , 'fireplace, chimney, hearth, flue');
    AddItem( 'remilami'  , 'lamp, torch, chandelier, lamppost, candlestick, light, beacon');
    AddItem( 'remilafa'  , 'window');
    AddItem( 'remilaso'  , 'shutters, blinds, lattice');
    AddItem( 'remilati'  , 'balcony, balustrade, railing');
    AddItem( 'remitido'  , 'cellar, vault');
    AddItem( 'remitire'  , 'cask, barrel');
    AddItem( 'remitifa'  , 'dally, dawdle, linger, slow, stragglers, slowpoke, slowly');
    AddItem( 'remitiso'  , 'staircase, stairs, steps, step');
    AddItem( 'remitila'  , 'floor');
    AddItem( 'refadore'  , 'install');
    AddItem( 'refadomi'  , 'furniture, furnishings');
    AddItem( 'refadofa'  , 'household');
    AddItem( 'refadoso'  , 'furniture, wardrobe');
    AddItem( 'refadola'  , 'drawer');
    AddItem( 'refadoti'  , 'library, librarian');
    AddItem( 'refaredo'  , 'closet, wardrobe, cabinet, cupboard');
    AddItem( 'refaremi'  , 'seat, bench, chair, sofa');
    AddItem( 'refarefa'  , 'sit, seated, sitting');
    AddItem( 'refareso'  , 'bed, cradle, hammock');
    AddItem( 'refarela'  , 'bedding, mattress, featherbed, pillow');
    AddItem( 'refareti'  , 'curtains, drapes');
    AddItem( 'refamido'  , 'die, , death, mortality, fatally');
    AddItem( 'refamire'  , 'stove, radiator, furnace, oven');
    AddItem( 'refamifa'  , 'andiron');
    AddItem( 'refamiso'  , 'breathe, blow, puff');
    AddItem( 'refamila'  , 'fuel, heating, wood, log, pyre');
    AddItem( 'refamiti'  , 'coal, embers');
    AddItem( 'refasodo'  , 'fire, furnace, incendiary');
    AddItem( 'refasore'  , 'burn, kindle, consume');
    AddItem( 'refasomi'  , 'twinkle, sparkle, scintillate, blaze, flame, flamboyant, blazing');
    AddItem( 'refasofa'  , 'pacify, calme, peace, amity, peacemaker, peaceful, peacefully');
    AddItem( 'refasola'  , 'chauffeur, driver');
    AddItem( 'refasoti'  , 'hearth, fireplace, firebox');
    AddItem( 'refalado'  , 'rag, dishtowel, apron, pillowcase, sheet');
    AddItem( 'refalare'  , 'tear, snag, torn, ragged, sparce, tattered');
    AddItem( 'refalami'  , 'mend, patch, mending, mender');
    AddItem( 'refalafa'  , 'sew, sewing, seamstress');
    AddItem( 'refalaso'  , 'border, edge');
    AddItem( 'refalati'  , 'embroider, embroidery, embroiderer');
    AddItem( 'refatido'  , 'thimble');
    AddItem( 'refatire'  , 'case, cover, sheath, box');
    AddItem( 'refatimi'  , 'needles, pins');
    AddItem( 'refatifa'  , 'chisel, cut, carving, scissors');
    AddItem( 'refatiso'  , 'cut, knife, cutter, cutting, sharp');
    AddItem( 'refatila'  , 'work, worker');
    AddItem( 'resodore'  , 'pouch, backpack, knapsack, bag');
    AddItem( 'resodomi'  , 'convenience, ease, convenient, conveniently');
    AddItem( 'resodofa'  , 'hide, bury, stash, cache, hidden, occult, stealthy, secretly');
    AddItem( 'resodoso'  , 'wallet, purse');
    AddItem( 'resodola'  , 'contain, include, containing, capacity');
    AddItem( 'resodoti'  , 'spend, expenditure, fees, disburse, liabilities');
    AddItem( 'resoredo'  , 'master, boss');
    AddItem( 'resoremi'  , 'order, command, precept');
    AddItem( 'resorefa'  , 'send, dispatch, sending, ship, shipment');
    AddItem( 'resoreso'  , 'dismiss, leave, vacant, dismissal');
    AddItem( 'resorela'  , 'develop, improve, development');
    AddItem( 'resoreti'  , 'remove, dismiss, removal, dismissal, resignation');
    AddItem( 'resomido'  , 'impotence, powerlessness, helplessness, impotent');
    AddItem( 'resomire'  , 'sweep, broom, sweeper, sweeping');
    AddItem( 'resomifa'  , 'wax, waxing');
    AddItem( 'resomiso'  , 'brush, dust');
    AddItem( 'resomila'  , 'continue, continuation, continuer, continual, continuously');
    AddItem( 'resomiti'  , 'clean, scrub, degrease, detach, scour, cleaning, cleanliness');
    AddItem( 'resofado'  , 'cunning, ruse, subterfuge, stratagem, crafty, artful, sly');
    AddItem( 'resofare'  , 'bathe, bath, bather');
    AddItem( 'resofami'  , 'timidity, shy, apprehensive, fearful, tentative, timidly');
    AddItem( 'resofaso'  , 'nudity, naked, bare, strip');
    AddItem( 'resofala'  , 'mop, wipe, sponge');
    AddItem( 'resofati'  , 'rub, massage');
    AddItem( 'resolado'  , 'fatigue, weariness, dejection, tired, overwhelmed, tiring');
    AddItem( 'resolare'  , 'undress');
    AddItem( 'resolami'  , 'stretch');
    AddItem( 'resolafa'  , 'shelter, coverage, cover, sheltered, covered');
    AddItem( 'resolaso'  , 'rest, relax, repose');
    AddItem( 'resolati'  , 'dream, dreaming, dreamy, reverie');
    AddItem( 'resotido'  , 'need, require, necessary, useful, indispensable, utile');
    AddItem( 'resotire'  , 'work, task, worker, travail, laborious, laboriously');
    AddItem( 'resotimi'  , 'practice, exercise, practicability, practitioner, practicable, executable');
    AddItem( 'resotifa'  , 'business, occupation, profession, career, worker, craftsman');
    AddItem( 'resotiso'  , 'book, work');
    AddItem( 'resotila'  , 'accustom, familiarize, habit, usage, custom, routine, familiar, habitual');
    AddItem( 'reladore'  , 'wash, launder, bleach, laundry, launderer');
    AddItem( 'reladomi'  , 'starch, stiffen, starched, stiffness');
    AddItem( 'reladofa'  , 'iron, ironing, ironer');
    AddItem( 'reladoso'  , 'wrinkle, crease, pleate, fold');
    AddItem( 'reladola'  , 'chains, shackles, irons');
    AddItem( 'reladoti'  , 'sharpen, whet, hone, ');
    AddItem( 'relaredo'  , 'crack, split, slot, crevasse, cracked');
    AddItem( 'relaremi'  , 'break, brittle, fragile, fragility');
    AddItem( 'relarefa'  , 'break, smash, brokenness, breaker, broken');
    AddItem( 'relareso'  , 'nail, pin, point, sharp, piercing, acute');
    AddItem( 'relarela'  , 'drill, bore, awl, perforate');
    AddItem( 'relareti'  , 'dig, excavate, hole, cavity, hollow');
    AddItem( 'relamido'  , 'stutter, mumble, stammer');
    AddItem( 'relamire'  , 'neighborhood, vicinity, close, neighborly, neighbor');
    AddItem( 'relamifa'  , 'strike, hit, impact, blow');
    AddItem( 'relamiso'  , 'agitate, shake, agitation, jerk, concussion');
    AddItem( 'relamila'  , 'cheap, inexpensive');
    AddItem( 'relamiti'  , 'ring, tinkle, peal, chime, bell, tintinnabulation, doorbell');
    AddItem( 'relafado'  , 'malignancy, wickedness, darkness, mischievous, wicked, maliciously, nastily');
    AddItem( 'relafare'  , 'bend, stoop, curvature, curve, bent');
    AddItem( 'relafami'  , 'tilt, lean, leaning, slope, inclination, penchant, tendency');
    AddItem( 'relafaso'  , 'turn');
    AddItem( 'relafala'  , 'disproportionate, disproportion, disproportionately');
    AddItem( 'relasodo'  , 'embarrass, embarrassment, embarrassing, discomfort');
    AddItem( 'relasore'  , 'throw, launch');
    AddItem( 'relasomi'  , 'find');
    AddItem( 'relasofa'  , 'raise, lift, elevator');
    AddItem( 'relasola'  , 'object');
    AddItem( 'relasoti'  , 'trifel, bagatelle, peccadillo, meticulous, minutiae');
    AddItem( 'relatido'  , 'fill, complete, filling, fullness, filled, full, packed, fully');
    AddItem( 'relatire'  , 'load, burden, cargo, loading');
    AddItem( 'relatimi'  , 'overflow, irruption, invasion');
    AddItem( 'relatifa'  , 'west, western');
    AddItem( 'relatiso'  , 'bottled, cork');
    AddItem( 'relatila'  , 'enclose, confine, contain, inclusion, confined, withdrawn, included, inclusive, enclosed');
    AddItem( 'retidore'  , 'race, breed');
    AddItem( 'retidomi'  , 'genealogy, genealogist, genealogical');
    AddItem( 'retidofa'  , 'ancestors, predecessors, sires');
    AddItem( 'retidoso'  , 'family, consanguinity, kinship, relative');
    AddItem( 'retidola'  , 'grandparent, grandfather, grandmother, granny');
    AddItem( 'retidoti'  , 'father, mother, paternity, maternity, paternal, maternal, dad, mom');
    AddItem( 'retiredo'  , 'cause, engender, generate, procreate, generative, genital');
    AddItem( 'retiremi'  , 'conceive, conception, pregnant, big    with    child');
    AddItem( 'retirefa'  , 'birth, childbirth');
    AddItem( 'retireso'  , 'son, daughter, filial');
    AddItem( 'retirela'  , 'kiss');
    AddItem( 'retireti'  , 'caress, caresses, caressing');
    AddItem( 'retimido'  , 'renounce, abandon, abjure, denunciation, renunciation, abdication, detachment');
    AddItem( 'retimire'  , 'brother, sister, fraternal, fraternally');
    AddItem( 'retimifa'  , 'grandson, granddaughter');
    AddItem( 'retimiso'  , 'uncle, aunt');
    AddItem( 'retimila'  , 'nephew, niece');
    AddItem( 'retimiti'  , 'cousin');
    AddItem( 'retifado'  , 'improvidence, improvident');
    AddItem( 'retifare'  , 'stepfather, stepmother');
    AddItem( 'retifami'  , 'son-in-law, daughter-in-law');
    AddItem( 'retifaso'  , 'stepbrother, stepsister');
    AddItem( 'retifala'  , 'tutelage, tutor');
    AddItem( 'retifati'  , 'pupil');
    AddItem( 'retisodo'  , 'godfather, godmother');
    AddItem( 'retisore'  , 'name, nominal, nominative, denominative');
    AddItem( 'retisomi'  , 'godson, goddaughter');
    AddItem( 'retisofa'  , 'intimacy, intimate, intimately');
    AddItem( 'retisola'  , 'familiarity, familiar, familiarly');
    AddItem( 'retisoti'  , 'camaraderie, fellowship, companion, comrade, peer');
    AddItem( 'retilado'  , 'promise, vow, promised');
    AddItem( 'retilare'  , 'oblige, useful, helpfulness, helpful');
    AddItem( 'retilami'  , 'loan, lend');
    AddItem( 'retilafa'  , 'host, hospitality');
    AddItem( 'retilaso'  , 'ignore, ignorance, illiterate');
    AddItem( 'retilati'  , 'location, place, locality, local');
    AddItem( 'midoredo'  , 'moral, immateriality, spirituality, intangible, spiritualist, spiritualism');
    AddItem( 'midoremi'  , 'origin, source, original, primitive, basic, originally');
    AddItem( 'midorefa'  , 'precede, anteriority, priority, primacy, predecessor, previous, before, antecedent');
    AddItem( 'midoreso'  , 'posterity, descendants');
    AddItem( 'midorela'  , 'clear');
    AddItem( 'midoreti'  , 'disperse, scatter, disseminate, strew, sprinkle, dispersion');
    AddItem( 'midomido'  , 'equip, endow, provide');
    AddItem( 'midomire'  , 'deposit');
    AddItem( 'midomifa'  , 'bequeath, testament');
    AddItem( 'midomiso'  , 'inherit, heritage, heir, heiress, recipient');
    AddItem( 'midomila'  , 'lack, miss, destitution, shortage, gap, famine, dearth, devoid, lacking, without');
    AddItem( 'midomiti'  , 'have, possess, possession, property, asset');
    AddItem( 'midofado'  , 'despair, desolation, desperate, despairing');
    AddItem( 'midofare'  , 'mourn, mourning, bereavement');
    AddItem( 'midofami'  , 'widowhood, widower, widow');
    AddItem( 'midofaso'  , 'orphanage, orphan');
    AddItem( 'midofala'  , 'abandon, leave, abandonment, abandoned, forasaken');
    AddItem( 'midofati'  , 'suffer, distressed, tested, suffering, affliction, grief, pain, anguish, distress');
    AddItem( 'midosodo'  , 'fear, apprehend, apprehension, fearful, apprehensive');
    AddItem( 'midosore'  , 'inconvenience, inconvenient, inconveniently');
    AddItem( 'midosomi'  , 'worry, concern, anxiety, alarm, anxious');
    AddItem( 'midosofa'  , 'complain, groan, whine, murmur, complaint, lamentation');
    AddItem( 'midosola'  , 'destiny, fate, predestination, destined');
    AddItem( 'midosoti'  , 'fatalism, fatalist, inevitably');
    AddItem( 'midolado'  , 'refuse, reject, repel, exclude, dismiss, rejection, refusal, exclusion, rejectable, unacceptable');
    AddItem( 'midolare'  , 'intervene, intercede, mediate, intervention, intercession, mediation, mediator, intercessor');
    AddItem( 'midolami'  , 'influence, influential');
    AddItem( 'midolafa'  , 'reconcile, conciliation, conciliator, conciliatory, reconcilable, accommodative');
    AddItem( 'midolaso'  , 'console, comfort, consolation, comforter, consolable');
    AddItem( 'midolati'  , 'content, satisfied, contentment, satisfaction, glad');
    AddItem( 'midotido'  , 'sacrifice, immolate, offering, consecration');
    AddItem( 'midotire'  , 'refrain, deprive, restriction, without, abstinence, deprivation');
    AddItem( 'midotimi'  , 'improbability, unlikely, incredible, implausible');
    AddItem( 'midotiso'  , 'try, endeavor, strive');
    AddItem( 'midotila'  , 'get, obtain');
    AddItem( 'miredore'  , 'particularize, specialize, particularity, speciality, particular, special, especially, particularly');
    AddItem( 'miredomi'  , 'rarity, rare, rarely, ');
    AddItem( 'miredofa'  , 'initiative');
    AddItem( 'miredoso'  , 'attempt, try, test');
    AddItem( 'miredola'  , 'undertake, undertaking, enterprise, entrepreneur');
    AddItem( 'miredoti'  , 'venture, hazard, risk, adventure, risky');
    AddItem( 'miremido'  , 'blind, blindness, blind, blindly');
    AddItem( 'miremire'  , 'enroll, register, registration, enrollment');
    AddItem( 'miremifa'  , 'document, title');
    AddItem( 'miremiso'  , 'official, formal, officially');
    AddItem( 'miremila'  , 'dedicate, dedication, dedicatory');
    AddItem( 'miremiti'  , 'stanza, verse');
    AddItem( 'mirefado'  , 'say, diction, elocution, speech, speaker');
    AddItem( 'mirefare'  , 'thing, object, stuff');
    AddItem( 'mirefami'  , 'main, essential, mainly, essentially');
    AddItem( 'mirefaso'  , 'seriousness, gravity, serious, grave, seriously');
    AddItem( 'mirefala'  , 'precious, preciously');
    AddItem( 'mirefati'  , 'curiosity');
    AddItem( 'miresodo'  , 'inexperience, inexperienced, unskilled');
    AddItem( 'miresore'  , 'participate, contribute, cooperate, participation, cooperation, participant');
    AddItem( 'miresomi'  , 'propose, plan, premeditate, project, intent, design, designer, intentioned');
    AddItem( 'miresofa'  , 'goal, aim');
    AddItem( 'miresola'  , 'diminish, thinning, thin, slender');
    AddItem( 'miresoti'  , 'destine, destination, destinator');
    AddItem( 'mirelado'  , 'wrong');
    AddItem( 'mirelare'  , 'reproach, blame, debug');
    AddItem( 'mirelami'  , 'scold, reprimand, admonish, admonition, scolding');
    AddItem( 'mirelafa'  , 'quarrel, fight, dispute, quibble, altercation');
    AddItem( 'mirelaso'  , 'insult, dishonor, revile, affront, snub, insolence, insulting, impertinent');
    AddItem( 'mirelati'  , 'incident, occurance, event, incidentally');
    AddItem( 'miretido'  , 'provoke, defy, provocation, defiance, provoker');
    AddItem( 'miretire'  , 'threaten, threat, menace, threatening, formidable');
    AddItem( 'miretimi'  , 'slap, kick, punch');
    AddItem( 'miretifa'  , 'hurt, wound, bruise');
    AddItem( 'miretiso'  , 'anger, angry, upset');
    AddItem( 'miretila'  , 'sulk, resentment, sulky, spiteful');
    AddItem( 'mifadore'  , 'character, nature, characteristic, characterized');
    AddItem( 'mifadomi'  , 'require, requirement');
    AddItem( 'mifadofa'  , 'envy, envious');
    AddItem( 'mifadoso'  , 'aspire, ambition, ambitious, ambitiously');
    AddItem( 'mifadola'  , 'solicit, solicitation, solicitor, insistant');
    AddItem( 'mifadoti'  , 'intrigue');
    AddItem( 'mifaredo'  , 'personality, personal');
    AddItem( 'mifaremi'  , 'incidental, incidentally');
    AddItem( 'mifarefa'  , 'self-esteem, pride');
    AddItem( 'mifareso'  , 'vanity, ostentation, vainglory, conceited, vain');
    AddItem( 'mifarela'  , 'arrogance, pride, haughtiness, arrogant, proud, imperious, haughty');
    AddItem( 'mifareti'  , 'boast, presumption, pride, proud, presumptuous');
    AddItem( 'mifamido'  , 'susceptibility, susceptible');
    AddItem( 'mifamire'  , 'mediocrity, mediocre, poor, common, regular');
    AddItem( 'mifamifa'  , 'laze, loiter, idle, laziness, dawdling, indolence, idleness, inactivity, lazy, loafer, nonchalant, inactive');
    AddItem( 'mifamiso'  , 'neglect, omit, negligence, omission, negligent, negligently');
    AddItem( 'mifamila'  , 'sensuality, intemperance, sensual, voluptuous, carnal');
    AddItem( 'mifamiti'  , 'gluttony, greedy, glutton, greedily');
    AddItem( 'mifasodo'  , 'libertine, prostitute, wantonness, prostitution, vagrant, debauchery');
    AddItem( 'mifasore'  , 'effrontery, impudence, insolence, cheeky, impudent');
    AddItem( 'mifasomi'  , 'corrupt, pervert, debauch, corruption, perversity, depravity, warped, perverted, perverse');
    AddItem( 'mifasofa'  , 'succumb, disarmed, defeated');
    AddItem( 'mifasola'  , 'blunt, blunten, blas');
    AddItem( 'mifasoti'  , 'awful, horrible, odious, hateable, detestable, execrable, abominable, horribly, dreadfully, badly');
    AddItem( 'mifalado'  , 'debase, degradation, abjection, ignominy, low, abject');
    AddItem( 'mifalare'  , 'betray, treason, traitor, treacherous, traitorously');
    AddItem( 'mifalami'  , 'steal, rob, swindle, theft, spoils, thief');
    AddItem( 'mifalafa'  , 'deceive, fool, cheat, guile, deceitful, deception, cheating, imposter, duplicity, misleading');
    AddItem( 'mifalaso'  , 'falsify, forgery, forged');
    AddItem( 'mifalati'  , 'violate, transgress, infringe, offense, contravention, offense, offender, transgressor, violator');
    AddItem( 'mifatido'  , 'adultery, adulterer');
    AddItem( 'mifatire'  , 'jealous, jealousy');
    AddItem( 'mifatimi'  , 'anger, exasperate, indignate, irritation, exasperation, irritated, outraged, exasperated');
    AddItem( 'mifatifa'  , 'fulminate, enraged, ardor, anger, frenzy, fury, rage, furious');
    AddItem( 'mifatiso'  , 'fickleness, lightness, frovolity, inconstant, volatile');
    AddItem( 'mifatila'  , 'revenge, avenge, avenger, vindictive');
    AddItem( 'misodore'  , 'tease, harass, teasing');
    AddItem( 'misodomi'  , 'reassure, calm, appease, calming, tranquility, serenity, serene, quiet, reassuring');
    AddItem( 'misodofa'  , 'annoy, tease, teasing, playfulness, harassing');
    AddItem( 'misodoso'  , 'thoughtlessness, ill-conceived, poorly    considered, thoughtlessly');
    AddItem( 'misodola'  , 'change, vary, variation, mutation, variability, instability, variable');
    AddItem( 'misodoti'  , 'diversify, diversity, variant, variety, variations');
    AddItem( 'misoredo'  , 'person, individual, people, personal, individually, personally');
    AddItem( 'misoremi'  , 'recommend, recommendation');
    AddItem( 'misorefa'  , 'presentation');
    AddItem( 'misoreso'  , 'accept, receive, admit, admission, reception, admissible, receivable');
    AddItem( 'misorela'  , 'meet, get    together, meeting, group, assemble, assembled, grouped');
    AddItem( 'misoreti'  , 'society, corporation, company, member    of    a    society');
    AddItem( 'misomido'  , 'smallness, pettiness, petty');
    AddItem( 'misomire'  , 'attend, to    frequent, attendance');
    AddItem( 'misomifa'  , 'remark, observation');
    AddItem( 'misomiso'  , 'watch, look, glance');
    AddItem( 'misomila'  , 'converse, gossip, chat, conversation, dialogue, conversationalist');
    AddItem( 'misomiti'  , 'answer, reply, respond, refute, response, refutation');
    AddItem( 'misofado'  , 'ugliness, hideous, unattractive');
    AddItem( 'misofare'  , 'attitude, poise, posture');
    AddItem( 'misofami'  , 'turn, pace');
    AddItem( 'misofaso'  , 'propriety, appropriateness, decency, proper, seemly, presentable');
    AddItem( 'misofala'  , 'distinction, distinguished');
    AddItem( 'misofati'  , 'politeness, courtesy, civility, polite');
    AddItem( 'misolado'  , 'unique, eccentric, quirky, original, bizarre, strange');
    AddItem( 'misolare'  , 'capricious, whimsical, moody');
    AddItem( 'misolami'  , 'mania, maniac');
    AddItem( 'misolafa'  , 'simper, pedantry, prudery, affected, prudish, pretentious');
    AddItem( 'misolaso'  , 'coquet, dainty, prettycoquettishly');
    AddItem( 'misolati'  , 'courtship');
    AddItem( 'misotido'  , 'inaugurate, inaguration, inaugural');
    AddItem( 'misotire'  , 'celebrate, celebration, ceremony, ceremonial');
    AddItem( 'misotimi'  , 'holiday, anniversary');
    AddItem( 'misotifa'  , 'indirect, indirectly');
    AddItem( 'misotiso'  , 'ball, ballet');
    AddItem( 'misotila'  , 'dance, dancer, dancing');
    AddItem( 'miladore'  , 'waltz, waltzer');
    AddItem( 'miladomi'  , 'simultaneity, simultaneous, collective, all, simultaneously, collectively');
    AddItem( 'miladofa'  , 'times, sometimes');
    AddItem( 'miladoso'  , 'entertainment, recreation, relaxation, hobby');
    AddItem( 'miladola'  , 'joke, banter, farce, comic, bantering, facetious, funny, :-), :)');
    AddItem( 'miladoti'  , 'parody, mimic, ape, caricature');
    AddItem( 'milaredo'  , 'mock, , ridicule, jeer, sneer, satirize, mockery, satire, scoffer, mocker, :p');
    AddItem( 'milaremi'  , 'ridiculous, ludicrous, preposterous, o_o, o_o, o_o, o_o');
    AddItem( 'milarefa'  , 'criticize, criticism');
    AddItem( 'milareso'  , 'decry, depreciate, disparage, slander, detraction, disparagement, slanderous, :(, :-(');
    AddItem( 'milarela'  , 'slander, slanderer, slanderously');
    AddItem( 'milareti'  , 'compromise, compromising');
    AddItem( 'milamido'  , 'ramble, rave, unreason, rambling, extravagance, illogical, unreasonable');
    AddItem( 'milamire'  , 'secret, mystery, mysterious, secretly, in    secret, mysteriously');
    AddItem( 'milamifa'  , 'confidential, confidence, confidant, confidentially');
    AddItem( 'milamiso'  , 'curiosity, curious, curiously, testing');
    AddItem( 'milamila'  , 'guess, diviner, guessable');
    AddItem( 'milamiti'  , 'disclose, divulge, reveal, disclosure, unveiling, revelation, discloser, revealer');
    AddItem( 'milafado'  , 'insensitivity, apathy, insensitive, apathetic');
    AddItem( 'milafare'  , 'indifference, indifferent, carefree, unconcerned, indifferently');
    AddItem( 'milafami'  , 'coldness, impassivity, cold, impassive, :-|, :|');
    AddItem( 'milafaso'  , 'composure, self-control');
    AddItem( 'milafala'  , 'prejudge, bias, prejudice, partiality, prejudiced, biased');
    AddItem( 'milafati'  , 'coterie, clique');
    AddItem( 'milasodo'  , 'blame, disapprove, reproach, disapproval, reproachful, disapprobation');
    AddItem( 'milasore'  , 'disdain, disrespect, irreverence');
    AddItem( 'milasomi'  , 'defame, lampoon, libel, diatribe, slanderer');
    AddItem( 'milasofa'  , 'challenge, discuss, debate, object, controvert, obejection, controversy, conflict, questionable');
    AddItem( 'milasola'  , 'insolvency');
    AddItem( 'milasoti'  , 'oppose, contradict, contrast, opposition, contradiciton, unlike, opposite');
    AddItem( 'milatido'  , 'congress');
    AddItem( 'milatire'  , 'note, annotate, annotation, remark');
    AddItem( 'milatimi'  , 'review, report, extract');
    AddItem( 'milatifa'  , 'deliberate, deliberation, deliberative');
    AddItem( 'milatiso'  , 'report, relate, reporter, minutes');
    AddItem( 'milatila'  , 'conclude, infer, conclusion, solution, conclusive, final, definitely');
    AddItem( 'mitidore'  , 'appearance, aspect, visibility, apparent, ostensible, visible, apparently');
    AddItem( 'mitidomi'  , 'likelihood, probability, likely, probable, presumable, believable');
    AddItem( 'mitidofa'  , 'seem, seeming, semblance, seems');
    AddItem( 'mitidoso'  , 'pretext, excuse, allege, allegation, subterfuge');
    AddItem( 'mitidola'  , 'dare, aplomb, boldness, audacity, dared, bold, boldly');
    AddItem( 'mitidoti'  , 'imbibe');
    AddItem( 'mitiredo'  , 'remain, stay');
    AddItem( 'mitirefa'  , 'anticipate, anticipation, forestall, premature');
    AddItem( 'mitireso'  , 'assume, suppose, presume, surmise, supposition, conjecture, hypothesis, supposed');
    AddItem( 'mitirela'  , 'sense, augur, foresee, omen, prescience, ominous');
    AddItem( 'mitireti'  , 'predict, prophesy, prediction, prognosis, prophet, divination, oracle, soothsayer, prophetic, divinator');
    AddItem( 'mitimido'  , 'slavery, yoke, subjugation, servility, servitude, slave, serf, servile');
    AddItem( 'mitimire'  , 'depend, dependence, dependent');
    AddItem( 'mitimifa'  , 'responsibility, responsible');
    AddItem( 'mitimiso'  , 'submit');
    AddItem( 'mitimila'  , 'obey, obedience, submission, resignation, obedient');
    AddItem( 'mitimiti'  , 'circumstance, case, circumstantial');
    AddItem( 'mitifado'  , 'indiscretion, indiscreet, obtrusive, indiscretely');
    AddItem( 'mitifare'  , 'sleepwalking, somnambulism, sleepwalker');
    AddItem( 'mitifami'  , 'magnetize, magnetism, magnet, magnetic');
    AddItem( 'mitifaso'  , 'fluid, fluidity');
    AddItem( 'mitifala'  , 'light, aerial, vaporous');
    AddItem( 'mitifati'  , 'spiritism, spiritist');
    AddItem( 'mitisodo'  , 'experiment, test, experimenter, experimental, testing, debug');
    AddItem( 'mitisore'  , 'prove, demonstrate, evidence, demonstration, proven, obvious, demonstrated');
    AddItem( 'mitisomi'  , 'observe, note, finding, observation, found, observed');
    AddItem( 'mitisofa'  , 'persuade, convince, persuation, persuasive, persuaded, convinced');
    AddItem( 'mitisola'  , 'believe, imagine, credulity, credulous');
    AddItem( 'mitisoti'  , 'certainty, surety, infallibility, authenticity, sure, infallible, indubitable, irrifutable, positive, certain, undoubtedly');
    AddItem( 'mitilado'  , 'prestidigitation, conjure, conjuror, magician');
    AddItem( 'mitilare'  , 'phantasmagoria, ghost, spectrum, specter, phantom, ');
    AddItem( 'mitilami'  , 'transform, transmute, transformation, transfigure, transmutation, transmutable');
    AddItem( 'mitilafa'  , 'bewitch, sorcery, witchcraft, sorcerer, magician, enchanting, magic');
    AddItem( 'mitilaso'  , 'partner, accomplice');
    AddItem( 'mitilati'  , 'colusion, accomplice');
    AddItem( 'fadoredo'  , 'nature, naturalist, natural, naturally');
    AddItem( 'fadoremi'  , 'countryside, country, countryman, villager, peasant, shepherd');
    AddItem( 'fadorefa'  , 'landscape, scenery, site, landscape');
    AddItem( 'fadoreso'  , 'picturesque, rugged');
    AddItem( 'fadorela'  , 'surround, gird, area, vicinity, surrounded, around, girded');
    AddItem( 'fadoreti'  , 'farm, farmer');
    AddItem( 'fadomido'  , 'mountain, mount, hill, mountainous, hilly');
    AddItem( 'fadomire'  , 'volcano, volcanic');
    AddItem( 'fadomifa'  , 'pit, chasm, abyss, precipice');
    AddItem( 'fadomiso'  , 'rock, rocky');
    AddItem( 'fadomila'  , 'dry, arid, drought, sterile, infertile, unsuccessful');
    AddItem( 'fadomiti'  , 'desert, deserted, uninhabited');
    AddItem( 'fadofado'  , 'isolate, seclude, isolation, loneliness, isolated, lonely, solitary, alone');
    AddItem( 'fadofare'  , 'withdraw, retire, refuge, asylum, reclusion, recluse');
    AddItem( 'fadofami'  , 'park, grove, forest, forestry');
    AddItem( 'fadofaso'  , 'tree, shrub, bush');
    AddItem( 'fadofala'  , 'foliage, leaves');
    AddItem( 'fadofati'  , 'shade, shadowing, shadowy, shady');
    AddItem( 'fadosodo'  , 'close, closing, closed');
    AddItem( 'fadosore'  , 'clear');
    AddItem( 'fadosomi'  , 'mow, reap, trim');
    AddItem( 'fadosofa'  , 'rake, harrow');
    AddItem( 'fadosola'  , 'dig, spade, pickaxe');
    AddItem( 'fadosoti'  , 'graft, grafter');
    AddItem( 'fadolado'  , 'delete, remove, subtract, prune, extract, deletion, removal');
    AddItem( 'fadolare'  , 'plant, planting, planter, plant');
    AddItem( 'fadolami'  , 'root');
    AddItem( 'fadolafa'  , 'stem, stalk, trunk, stump');
    AddItem( 'fadolaso'  , 'branch, limb, bough');
    AddItem( 'fadolati'  , 'hedge, bush, undergrowth');
    AddItem( 'fadotido'  , 'garden, gardening, horticulture, gardener, horticulturist');
    AddItem( 'fadotire'  , 'garden, grove');
    AddItem( 'fadotimi'  , 'bloom, flower, blossom, florist, floral');
    AddItem( 'fadotifa'  , 'bouquet, flower-seller, florist');
    AddItem( 'fadotiso'  , 'classic, classical, standard, traditional');
    AddItem( 'fadotila'  , 'perfume, embalm, balm, smell, scent, aroma, perfumer, fragrant, odorous, embalmed');
    AddItem( 'faredore'  , 'cultivate, culture, agriculture, cultivator, farmer, agricultural');
    AddItem( 'faredomi'  , 'succeed, successor, after');
    AddItem( 'faredofa'  , 'plow, furrow, till, tillage, husbandman, ploughable');
    AddItem( 'faredoso'  , 'sow, sowing, seed');
    AddItem( 'faredola'  , 'grain, oats, millet, corn');
    AddItem( 'faredoti'  , 'wheat, rye');
    AddItem( 'faremido'  , 'ripen, mature, maturity, ripe, maturely');
    AddItem( 'faremire'  , 'pick, pluck, collect, picking');
    AddItem( 'faremifa'  , 'harvest, reap, collect, reaper, harvester');
    AddItem( 'faremiso'  , 'glean, gleaner');
    AddItem( 'faremila'  , 'ear');
    AddItem( 'faremiti'  , 'field, meadow, prairie, plain, pastoral');
    AddItem( 'farefado'  , 'vegetate, grow, vegetation, growth, plant');
    AddItem( 'farefare'  , 'greening, greenery, herbage, turf, grass');
    AddItem( 'farefami'  , 'valley, dale');
    AddItem( 'farefaso'  , 'hunt, hunting, hunter');
    AddItem( 'farefala'  , 'pursue, pursuit, pursuer');
    AddItem( 'farefati'  , 'savagery, savage, wild');
    AddItem( 'faresodo'  , 'horse, steed');
    AddItem( 'faresore'  , 'bridle, reins, bit');
    AddItem( 'faresomi'  , 'saddle, saddler');
    AddItem( 'faresofa'  , 'stirrup');
    AddItem( 'faresola'  , 'whip, crop');
    AddItem( 'faresoti'  , 'harness');
    AddItem( 'farelado'  , 'animal, quadroped');
    AddItem( 'farelare'  , 'sheep, lamb');
    AddItem( 'farelami'  , 'cattle, cow, bull, beef');
    AddItem( 'farelafa'  , 'pig, pork, sow, boar');
    AddItem( 'farelaso'  , 'goat, kid');
    AddItem( 'farelati'  , 'donkey, ass');
    AddItem( 'faretido'  , 'camel');
    AddItem( 'faretire'  , 'dog, bitch');
    AddItem( 'faretimi'  , 'cat');
    AddItem( 'faretifa'  , 'bird');
    AddItem( 'faretiso'  , 'wings, plumage, feathers');
    AddItem( 'faretila'  , 'fly, flit, flight');
    AddItem( 'famidore'  , 'conscription, recruitment, ');
    AddItem( 'famidomi'  , 'soldier, corporal, brigadier, sergeant');
    AddItem( 'famidofa'  , 'uniform, helmet, tunic');
    AddItem( 'famidoso'  , 'equipment, pouch, belt, harness, knapsack');
    AddItem( 'famidola'  , 'arms, weapons, weapon');
    AddItem( 'famiredo'  , 'staff');
    AddItem( 'famiremi'  , 'officer, lieutenant, captain, major');
    AddItem( 'famirefa'  , 'colonel');
    AddItem( 'famirela'  , 'general, generalship, generalissimo');
    AddItem( 'famireti'  , 'marshal');
    AddItem( 'famifado'  , 'vice, vicious, viciously');
    AddItem( 'famifare'  , 'army, troops');
    AddItem( 'famifami'  , 'activity, vigilance, active, vigilant, alert, actively');
    AddItem( 'famifaso'  , 'brigade, division');
    AddItem( 'famifala'  , 'regiment, legion, battalion, squadron, company');
    AddItem( 'famifati'  , 'zouaves, spahis');
    AddItem( 'famisodo'  , 'deaden');
    AddItem( 'famisore'  , 'infantry, rifleman, infantryman, grenadier, outfielder');
    AddItem( 'famisomi'  , 'cavalry, cavalryman');
    AddItem( 'famisofa'  , 'artillery, gunner');
    AddItem( 'famisola'  , 'reverse, inverse');
    AddItem( 'famisoti'  , 'engineer, engineering');
    AddItem( 'familado'  , 'camp, encampment');
    AddItem( 'familare'  , 'tent');
    AddItem( 'familami'  , 'bivouac');
    AddItem( 'familafa'  , 'itinerant, travelling');
    AddItem( 'familaso'  , 'material, equipment');
    AddItem( 'familati'  , 'flag, banner');
    AddItem( 'famitido'  , 'firearms, musketry');
    AddItem( 'famitire'  , 'bomb, cannon, mortar');
    AddItem( 'famitimi'  , 'percussion, drums');
    AddItem( 'famitifa'  , 'shoot, rifle, shooting, discharge');
    AddItem( 'famitila'  , 'pistol, handgun');
    AddItem( 'fasodore'  , 'gunpowder');
    AddItem( 'fasodomi'  , 'cartridge, capsule');
    AddItem( 'fasodofa'  , 'bullet');
    AddItem( 'fasodoso'  , 'initiate, prime, wick, fuse');
    AddItem( 'fasodola'  , 'sword, saber');
    AddItem( 'fasodoti'  , 'dagger, poniard, stilleto');
    AddItem( 'fasoredo'  , 'covet, lust, greed, insatiability, rapacious, greedy, insatiable');
    AddItem( 'fasoremi'  , 'besiege, siege, besieging');
    AddItem( 'fasorefa'  , 'attack, assail, offensive, assault, aggression, attacker, assailant, aggressor, aggressive');
    AddItem( 'fasoreso'  , 'fight, combat, battle, melee, battlefield');
    AddItem( 'fasorela'  , 'fire, unload, release, shooting, rifleman');
    AddItem( 'fasomido'  , 'stupidity, inability, unable, limited, unintelligent, unintelligently');
    AddItem( 'fasomire'  , 'strategy, tactic, tactical, strategic');
    AddItem( 'fasomifa'  , 'serve, service, servant');
    AddItem( 'fasomiso'  , 'resist, defend, resistance, resistant, unbeaten');
    AddItem( 'fasomila'  , 'inflexibility, invulnerability, invinciblility, indomitable, unshakable, irresistibly, unswervingly');
    AddItem( 'fasomiti'  , 'stubborn, obstinate, stubbornness, obstinately');
    AddItem( 'fasofado'  , 'unsightly, awkward, ugly, ungraceful, unpleasing, clumsy');
    AddItem( 'fasofare'  , 'warring, war, warrior, military, martial, warlike');
    AddItem( 'fasofami'  , 'conquer, overcome, disarm, tame, surmount, victory, winner');
    AddItem( 'fasofaso'  , 'conquer, conquest, feat, conqueror');
    AddItem( 'fasofala'  , 'heroism, heroes, hero, heroin, heroic');
    AddItem( 'fasofati'  , 'triumph, victory, victor, triumphant, triumphal');
    AddItem( 'fasolado'  , 'sea, ocean, gulf, bay');
    AddItem( 'fasolare'  , 'drift, fluctuate, undulate, flow, wave, vague, ripple, floating, rolling, undulating');
    AddItem( 'fasolami'  , 'marine, maritime');
    AddItem( 'fasolafa'  , 'fleet, squadron');
    AddItem( 'fasolaso'  , 'ship, brig');
    AddItem( 'fasolati'  , 'boat, skiff');
    AddItem( 'fasotido'  , 'mast, foremast');
    AddItem( 'fasotire'  , 'sail');
    AddItem( 'fasotimi'  , 'rudder');
    AddItem( 'fasotifa'  , 'mariner, pilot, boatman');
    AddItem( 'fasotiso'  , 'sailor');
    AddItem( 'fasotila'  , 'admirality, admiral, vice-admiral, contre-admiral');
    AddItem( 'faladore'  , 'navigation, navigator, navigable');
    AddItem( 'faladomi'  , 'space, expanse, area, spacious, expansive');
    AddItem( 'faladofa'  , 'kilometer');
    AddItem( 'faladoso'  , 'row, oar, rowing, rower, oarsman');
    AddItem( 'faladola'  , 'sail, sailing');
    AddItem( 'faladoti'  , 'maneuver');
    AddItem( 'falaredo'  , 'access, dock, boardable, accessible');
    AddItem( 'falaremi'  , 'anchor, mooring');
    AddItem( 'falarefa'  , 'seaport, harbor');
    AddItem( 'falareso'  , 'beach, shore, bank');
    AddItem( 'falarela'  , 'coast, seashore, sea-cliff');
    AddItem( 'falareti'  , 'strait, peninsula, isthmus');
    AddItem( 'falamido'  , 'madness, insanity, dementia, crazy, insane');
    AddItem( 'falamire'  , 'lake, lagoon, pond');
    AddItem( 'falamifa'  , 'depth, deep, deeply');
    AddItem( 'falamiso'  , 'fishing, fisherman');
    AddItem( 'falamiti'  , 'bait, lure, fishhook');
    AddItem( 'falafado'  , 'swim, swimmer');
    AddItem( 'falafare'  , 'flood, inundate, swamp, submerge, overwhelm');
    AddItem( 'falafami'  , 'dupe, fooled, duped');
    AddItem( 'falafaso'  , 'shipwreck, shipwrecked, castaway');
    AddItem( 'falafala'  , 'drowned');
    AddItem( 'falafati'  , 'torrent, surge, torrentuous, billowing');
    AddItem( 'falasodo'  , 'creek, stream, brook');
    AddItem( 'falasore'  , 'channel, canal, conduit');
    AddItem( 'falasomi'  , 'well, sump');
    AddItem( 'falasofa'  , 'pump, draw, pour, pumper');
    AddItem( 'falasola'  , 'flowing, gush, gushing, cascade, waterfall');
    AddItem( 'falasoti'  , 'unite, rally, union, rally, federalism, confederation, coalition, league, federal, united');
    AddItem( 'falatido'  , 'continent, mainland, continental');
    AddItem( 'falatire'  , 'europe, european');
    AddItem( 'falatimi'  , 'asia, asian');
    AddItem( 'falatifa'  , 'africa, african');
    AddItem( 'falatiso'  , 'america, american');
    AddItem( 'falatila'  , 'oceania, australia, oceanian, australian');
    AddItem( 'fatidore'  , 'travel, tour, peregrination, pilgrimage, traveler, tourist, pilgrim');
    AddItem( 'fatidomi'  , 'leave, depart');
    AddItem( 'fatidofa'  , 'route');
    AddItem( 'fatidoso'  , 'separating, separation, separable');
    AddItem( 'fatidola'  , 'remoteness, distance, distant, remote');
    AddItem( 'fatidoti'  , 'absence, absent');
    AddItem( 'fatiredo'  , 'railway, railroad');
    AddItem( 'fatiremi'  , 'locomotive, steam');
    AddItem( 'fatirefa'  , 'transport, transfer, transportation, transferable, transportable');
    AddItem( 'fatireso'  , 'baggage, luggage, case, trunk, suitcase');
    AddItem( 'fatirela'  , 'duty, debt, debtor, indebted');
    AddItem( 'fatireti'  , 'roll, wheel, bearing, cog');
    AddItem( 'fatimido'  , 'rushing, lan, spontaneity, suddenness, spontaneous, suddenly, immediately, instantly, spontan');
    AddItem( 'fatimire'  , 'speed, haste, velocity, promptness, quickly, fast');
    AddItem( 'fatimifa'  , 'vivacity, exuberance, impetuosity, dashing, lively, expeditious, strongly, impetuously');
    AddItem( 'fatimiso'  , 'move, movement, oscillation, locomotion, fluctuation, engine, moving');
    AddItem( 'fatimila'  , 'hasten, accelerate, acceleration, accelerator');
    AddItem( 'fatimiti'  , 'advance, move, thrust, forward');
    AddItem( 'fatifare'  , 'perspective');
    AddItem( 'fatifami'  , 'moderate, temper, repress, moderation, restraint, moderate, withheld, reserve, tempered, temperate, moderate');
    AddItem( 'fatifaso'  , 'accident, incident, event');
    AddItem( 'fatifala'  , 'scare, frighten, terrify, scary, frightening, frightful, terrible, dreadful');
    AddItem( 'fatifati'  , 'scaring, shudder, terror, fright, fear, dread, trembling, terrified, scared');
    AddItem( 'fatisodo'  , 'train, wagon');
    AddItem( 'fatisore'  , 'route');
    AddItem( 'fatisomi'  , 'direct, express, directly');
    AddItem( 'fatisofa'  , 'road, trail, way');
    AddItem( 'fatisola'  , 'stop, station, respite, truce, stationary');
    AddItem( 'fatisoti'  , 'bifurcate, branch, bifurcation');
    AddItem( 'fatilado'  , 'south, southern, meridian');
    AddItem( 'fatilare'  , 'east, eastern, orient, oriental');
    AddItem( 'fatilafa'  , 'joining, meeting, rejoining');
    AddItem( 'fatilaso'  , 'arrive, reach, achieve, arrival, finish');
    AddItem( 'fatilati'  , 'station, wharf, pier');
    AddItem( 'sodoredo'  , 'art, artist, artistic, artistically');
    AddItem( 'sodoremi'  , 'theater, theatre, theatrical, theatrically');
    AddItem( 'sodorefa'  , 'represent, representation');
    AddItem( 'sodoreti'  , 'dcor, decorator');
    AddItem( 'sodomido'  , 'opera, musical');
    AddItem( 'sodomire'  , 'comedy, comedian');
    AddItem( 'sodomifa'  , 'tragedy, tragedian, tragic');
    AddItem( 'sodomiso'  , 'drama, dramatist, dramatic, dramatically');
    AddItem( 'sodomila'  , 'author, writer');
    AddItem( 'sodomiti'  , 'actor, actress');
    AddItem( 'sodofado'  , 'recite, recitation, reciter, reciting');
    AddItem( 'sodofare'  , 'declaim, recite, declamation, recitative, declamatory');
    AddItem( 'sodofami'  , 'gesticulate, mime, mimic, gesture, gesticulation, pantomime, mimicry');
    AddItem( 'sodofaso'  , 'tell, recount, describe, narrate, narration, description, raconteur, descriptive');
    AddItem( 'sodofala'  , 'accent, inflection, accentuated');
    AddItem( 'sodofati'  , 'express, enunciate, emit, expression, significant, expressive, expressible');
    AddItem( 'sodosodo'  , 'exit');
    AddItem( 'sodosore'  , 'repertoire');
    AddItem( 'sodosomi'  , 'role, character');
    AddItem( 'sodosofa'  , 'masterpiece');
    AddItem( 'sodosola'  , 'effect, sensation');
    AddItem( 'sodosoti'  , 'succeed, success, successful');
    AddItem( 'sodolado'  , 'dilettantism, dilettante');
    AddItem( 'sodolare'  , 'place, situate, place, spot, position');
    AddItem( 'sodolati'  , 'amphitheater');
    AddItem( 'sodotido'  , 'orchestrate, instrumentation, orchestra');
    AddItem( 'sodotire'  , 'instrument, instrumentalist, instrumental');
    AddItem( 'sodotimi'  , 'violin, viola, violinist');
    AddItem( 'sodotifa'  , 'cello, bass, cellist, contra-bass');
    AddItem( 'sodotiso'  , 'bow');
    AddItem( 'sodotila'  , 'strings');
    AddItem( 'soredore'  , 'harp, harpist');
    AddItem( 'soredomi'  , 'flute, flutist');
    AddItem( 'soredofa'  , 'woodwind, clarinet, bassoon');
    AddItem( 'soredoso'  , 'saxophone, trombone');
    AddItem( 'soredola'  , 'trumpet, bugle, horn');
    AddItem( 'soredoti'  , 'fanfare, sounding');
    AddItem( 'soremido'  , 'perform, performance, performer, performing, executive');
    AddItem( 'soremire'  , 'accompany, accompaniment, accompanist');
    AddItem( 'soremifa'  , 'sing, song, singer, singing');
    AddItem( 'soremiso'  , 'voice, vocal');
    AddItem( 'soremila'  , 'tenor, counter-tenor, countertenor, baritone');
    AddItem( 'soremiti'  , 'soprano, mezzo-soprano, mezzo-contralto, contralto');
    AddItem( 'sorefado'  , 'vocalize, vocalization, vocalized');
    AddItem( 'sorefare'  , 'agility, quickness, dexterity, nimbleness, agile, nimble, quickly, nimbly, swiftly, lightly');
    AddItem( 'sorefami'  , 'flexible, flexibility, versatility, maneuverability, supple');
    AddItem( 'sorefala'  , 'talent, talented');
    AddItem( 'sorefati'  , 'perfect, improve, develop, enhance, improvement, development, improved, perfected, perfectible');
    AddItem( 'soresodo'  , 'idleness, leisure, unoccupied, idle, aimless');
    AddItem( 'soresore'  , 'sentimentality, feeling, sentimental, romantic');
    AddItem( 'soresomi'  , 'interest, concern');
    AddItem( 'soresofa'  , 'touching, moving, pathetic');
    AddItem( 'soresola'  , 'delude, illusion, delusion, chimera, utopia, illusory, chimerical, imaginary, fantastic, misleading');
    AddItem( 'soresoti'  , 'symbolize, figure, symbol, allegory, symbolism, emblem, figurative, symbolic, allegorical, figuratively');
    AddItem( 'sorelado'  , 'surprise, amaze, astonish, astound, marvel, amazing, surprising, unimaginable, striking, stunning, extraordinary, ');
    AddItem( 'sorelare'  , 'wonder, marvel, stupefaction, surprised, amazed, stunned, struck, shocked');
    AddItem( 'sorelami'  , 'fascinate, mesmerize, enchant, delight, excite, inspire, fascination, prestige, fascinating');
    AddItem( 'sorelafa'  , 'feel, experience, impression, feeling, emotion, stir, thrill, impressed, moved, affected');
    AddItem( 'sorelaso'  , 'enthusiasm, excitement, enthusiastic, excited');
    AddItem( 'sorelati'  , 'irresistible, irresistibly');
    AddItem( 'soretido'  , 'captivate, enthrall, fascinate, subjugate, submit');
    AddItem( 'soretire'  , 'attend, audience, listener, spectator, public');
    AddItem( 'soretimi'  , 'applaud, applauds, cheer, applause, cheers, clapping, bravos, cheering, applauders, claps');
    AddItem( 'soretifa'  , 'interrupt, suspend, pause, interruption, suspension, interrupter, suspended, stopped, discontinued, pending');
    AddItem( 'soretiso'  , 'approve, endorse, sanction, ratify, approval, endorsement, vote, sanction, ratification, approving, approvingly');
    AddItem( 'soretila'  , 'unanimous, universal, unianimously');
    AddItem( 'somidore'  , 'compete, rivalry, competition, rival, competitor');
    AddItem( 'somidomi'  , 'emulate, emulator');
    AddItem( 'somidofa'  , 'animation, verve, spirit, nerve, animate');
    AddItem( 'somidoso'  , 'excel, excellence, excellent, excellently');
    AddItem( 'somidola'  , 'shine, dazzle, sparkle, burst, flash, glitter, brilliant, bright, dazzling, blinding, stunning, brilliantly, stunningly, bri');
    AddItem( 'somidoti'  , 'surpass, eclipse, overshadow, outweigh, prevail, dominate, exceed');
    AddItem( 'somiredo'  , 'acoustic');
    AddItem( 'somirefa'  , 'tone, tonality, tonic');
    AddItem( 'somireso'  , 'sharp, chromatic');
    AddItem( 'somirela'  , 'flat');
    AddItem( 'somireti'  , 'natural');
    AddItem( 'somifado'  , 'misconduct, immorality');
    AddItem( 'somifami'  , 'interval, gap');
    AddItem( 'somifaso'  , 'semitone');
    AddItem( 'somifati'  , 'dominant, sub-dominant');
    AddItem( 'somisodo'  , 'compose, composition, composer, maestro');
    AddItem( 'somisore'  , 'music, musician, musical, musically');
    AddItem( 'somisomi'  , 'note');
    AddItem( 'somisofa'  , 'sound, intonation');
    AddItem( 'somisola'  , 'melody, melodist, melodious, melodiously');
    AddItem( 'somisoti'  , 'harmony, harmonist, harmonious, harmoniously');
    AddItem( 'somilado'  , 'organize, organization, organizer');
    AddItem( 'somilare'  , 'concert');
    AddItem( 'somilami'  , 'program, brochure, flyer, bulletin');
    AddItem( 'somilafa'  , 'piano, pianist');
    AddItem( 'somilaso'  , 'organ, harmonium, harmonicorde, organist');
    AddItem( 'somilati'  , 'keyboard');
    AddItem( 'somitido'  , 'lecture, course');
    AddItem( 'somitire'  , 'teach, instruct, educate, profess, professorship, teaching, didactic, teacher, tutor, professor, didactics');
    AddItem( 'somitimi'  , 'method, system, methodist, methodical, systematic, methodically, systematically');
    AddItem( 'somitifa'  , 'immobility, immobile');
    AddItem( 'somitiso'  , 'theory, theorist, theoretician, theoretical');
    AddItem( 'somitila'  , 'problem, issue, challenge, touble, problematic, bug');
    AddItem( 'sofadore'  , 'philotechnic');
    AddItem( 'sofadomi'  , 'philharmonic, philharmonic');
    AddItem( 'sofadofa'  , 'choirmaster');
    AddItem( 'sofadoso'  , 'choir, singer, chorus');
    AddItem( 'sofadoti'  , 'cantata');
    AddItem( 'sofaredo'  , 'concerto');
    AddItem( 'sofaremi'  , 'cavatina');
    AddItem( 'sofarefa'  , 'solo, duo, trio, quartet, quintet');
    AddItem( 'sofareso'  , 'portamento');
    AddItem( 'sofarela'  , 'trill, cadence');
    AddItem( 'sofareti'  , 'fermata, pause');
    AddItem( 'sofamido'  , 'ageing, decay, decrepitude, decrepit');
    AddItem( 'sofamire'  , 'scale');
    AddItem( 'sofamifa'  , 'solfge');
    AddItem( 'sofamiso'  , 'quaver');
    AddItem( 'sofamila'  , 'syncopate, syncopation');
    AddItem( 'sofamiti'  , 'rhythm, measure, rhythmic, rhythmical');
    AddItem( 'sofasodo'  , 'disappear, vanish, disappearance, gone, missing, vanished');
    AddItem( 'sofasore'  , 'score');
    AddItem( 'sofasomi'  , 'prelude');
    AddItem( 'sofasofa'  , 'modulate, modulation');
    AddItem( 'sofasola'  , 'transpose, transposition, transposer');
    AddItem( 'sofasoti'  , 'fugue');
    AddItem( 'sofalado'  , 'arpeggiate, arpeggio');
    AddItem( 'sofalami'  , 'base, foundation, basis, fundamental, fundamentally');
    AddItem( 'sofalafa'  , 'concern, related');
    AddItem( 'sofalaso'  , 'symphony, symphonist');
    AddItem( 'sofalati'  , 'counterpoint');
    AddItem( 'sofatido'  , 'rests');
    AddItem( 'sofatire'  , 'drum, drumming');
    AddItem( 'sofatimi'  , 'resonate, resound, vibrate, sound, resonance, echo, resounding');
    AddItem( 'sofatifa'  , 'whistle');
    AddItem( 'sofatiso'  , 'castanets, tambourine, cymbal');
    AddItem( 'soladore'  , 'museum, gallery');
    AddItem( 'soladomi'  , 'picture, painting, print, engraving, frame');
    AddItem( 'soladofa'  , 'sketch, outline, draft');
    AddItem( 'soladoso'  , 'draw, drawing, line');
    AddItem( 'soladola'  , 'paint, painting, painter, color, shade');
    AddItem( 'soladoti'  , 'miniature');
    AddItem( 'solaredo'  , 'watercolor');
    AddItem( 'solaremi'  , 'pastel');
    AddItem( 'solarefa'  , 'gouache, guache');
    AddItem( 'solareso'  , 'fresco');
    AddItem( 'solarela'  , 'palette');
    AddItem( 'solareti'  , 'brush, paintbrush');
    AddItem( 'solamido'  , 'modelling, model');
    AddItem( 'solamire'  , 'moulder, molder');
    AddItem( 'solamifa'  , 'sculpt, carve, sculpture, carving, sculptor, carver');
    AddItem( 'solamiso'  , 'statue, bust, torso, statuary');
    AddItem( 'solamila'  , 'pedestal, base');
    AddItem( 'solamiti'  , 'relief');
    AddItem( 'solafado'  , 'daguerreotype, daguerre');
    AddItem( 'solafare'  , 'photograph, photograph, photographic');
    AddItem( 'solafami'  , 'reproduce, reflect, reproduction, reflection, reflectivity, reflexivity, reflector');
    AddItem( 'solafaso'  , 'portrait, image, picture, likeness, portrait    painter');
    AddItem( 'solafala'  , 'resemble, like, similarity, resemblance, likeness, analogy, relation, homogeneity, similar, homogeneous');
    AddItem( 'solafati'  , 'conformity, identity, conforms, similar, same, identical');
    AddItem( 'solasodo'  , 'stenography, shorthand, stenographer, stenographic');
    AddItem( 'solasore'  , 'lithography, lithographer, lithographical');
    AddItem( 'solasomi'  , 'engrave, engraving, engraver');
    AddItem( 'solasola'  , 'compass');
    AddItem( 'solasoti'  , 'adjust, specify, tune');
    AddItem( 'solatido'  , 'academy, academician, academic');
    AddItem( 'solatire'  , 'knowledge, science, education, instruction, scholarship, savant, educated, enlightened, erudite, scholar');
    AddItem( 'solatimi'  , 'institute, institution, school');
    AddItem( 'solatifa'  , 'baccalaureat');
    AddItem( 'solatiso'  , 'encyclopedia, encyclopedist, encyclopedic');
    AddItem( 'solatila'  , 'mathematics, mathematician, mathematical, mathematically');
    AddItem( 'sotidore'  , 'literature, writer, literary');
    AddItem( 'sotidomi'  , 'prose');
    AddItem( 'sotidofa'  , 'novel, story, tale, fiction, novelist, romantic, fictitious, fictitiously');
    AddItem( 'sotidoso'  , 'fable, parable, fabulous');
    AddItem( 'sotidola'  , 'poetry, rhyme, versification, rhymer, versifier');
    AddItem( 'sotidoti'  , 'poeticize, idealize, poetry, ideality, poet, bard, poetic, perfect, ideal, poetically');
    AddItem( 'sotiredo'  , 'speaker, orator, oratory');
    AddItem( 'sotiremi'  , 'inspiration, inspiring');
    AddItem( 'sotirefa'  , 'discourse, harangue, speech, address, speaker, talker, orator');
    AddItem( 'sotireso'  , 'eloquence, eloquent, eloquently');
    AddItem( 'sotirela'  , 'remarkable, notable, outstanding, remarkably, notably');
    AddItem( 'sotireti'  , 'pinnacle, zenith, climax, culmination, virtuoso, phoenix, eminent, transcendent, sublime');
    AddItem( 'sotimido'  , 'softening, softeness, weakness, enervation, soft, weakens, effeminate');
    AddItem( 'sotimire'  , 'idea, notion, concept');
    AddItem( 'sotimifa'  , 'metaphysics, ideology, metaphysician, ideologist, idealogical, ideographic');
    AddItem( 'sotimiso'  , 'think, meditate, reflect, thought, meditation, reflection, thinker, thoughtful, meditative, musingly');
    AddItem( 'sotimila'  , 'imagine, imagination, imaginative, imaginatively, imaginable');
    AddItem( 'sotimiti'  , 'abstraction, abstract');
    AddItem( 'sotifado'  , 'dishonesty, disloyalty, dishonest, unfair, dishonestly, unfairly');
    AddItem( 'sotifare'  , 'genius, ingenuity, ingenious, clever, ingeniously');
    AddItem( 'sotifami'  , 'persevere, persist, perseverance, persistence, constancy, persevering, persistent, constant, steady, constantly');
    AddItem( 'sotifaso'  , 'invent, create, invention, creation, inventor, creator, inventive');
    AddItem( 'sotifala'  , 'discover, discovery');
    AddItem( 'sotifati'  , 'innovate, innovation, newness, novelty, innovative, innovator, new, modern, recent, newly, recently, modernly');
    AddItem( 'sotisodo'  , 'fight, fighter, wrestler');
    AddItem( 'sotisore'  , 'propagate, popularize, disseminate, spread, propaganda, popularization');
    AddItem( 'sotisomi'  , 'popularity, popular, accredited, popularly');
    AddItem( 'sotisofa'  , 'reputation, fame, repute, reputation, famous, renowned, reputed, famed');
    AddItem( 'sotisola'  , 'demonstration, celebrity, illustrious, famous');
    AddItem( 'sotisoti'  , 'glory, glorious, gloriously');
    AddItem( 'sotilado'  , 'telegraph, telegraphic, telegraphically');
    AddItem( 'sotilare'  , 'electricity, electric');
    AddItem( 'sotilami'  , 'telephony, telephone, telephonic, internet, internets, smartphone, handy');
    AddItem( 'sotilafa'  , 'indicate, sign, evidence, signal');
    AddItem( 'sotilaso'  , 'correspond, communicate, correspondence, communication, telegram, corresponding');
    AddItem( 'sotilati'  , 'inform, warn, announce, notify, share, notice, advertisement');
    AddItem( 'ladoredo'  , 'industry, industral, industrious, industriously');
    AddItem( 'ladoremi'  , 'manufacture, make, manufacturing, making, manufacturer, factory');
    AddItem( 'ladorefa'  , 'material, substance, materiality, materials, materially');
    AddItem( 'ladoreso'  , 'produce, production, product, producer');
    AddItem( 'ladorela'  , 'selling, sale, seller');
    AddItem( 'ladoreti'  , 'wholesale');
    AddItem( 'ladomido'  , 'establish, institute, erect, found, establishment, foundation, erection, institution, founder');
    AddItem( 'ladomire'  , 'shop, store, boutique, shopkeeper, ebay');
    AddItem( 'ladomifa'  , 'workshop, studio, foreman');
    AddItem( 'ladomiso'  , 'tarnish, dull, dim, matte');
    AddItem( 'ladomila'  , 'agree, stipulate, convention, stipulation, pact, market, clause, conventional, convention');
    AddItem( 'ladomiti'  , 'deal, negotiate, compromise, capitulate, treaty, transaction, capitulation, transactional');
    AddItem( 'ladofado'  , 'trade, traffic, merchant, trading, trafficker, dealer, commercial, commercially');
    AddItem( 'ladofare'  , 'associate, association, copartner, partner, jointly, shared');
    AddItem( 'ladofami'  , 'condition, conditional, conditionally');
    AddItem( 'ladofaso'  , 'fulfilling, achieve, accomplish, realize, execute, achievement, feasible');
    AddItem( 'ladofala'  , 'indivisibility, indissolubility, indivisible, indissoluble, indivisibly');
    AddItem( 'ladofati'  , 'linen, flax');
    AddItem( 'ladosodo'  , 'outside, exterior, external, outwardly, outdoors');
    AddItem( 'ladosore'  , 'spin, thread, yarn, wire, spinner, weaver, spindle, coil, reel, bobbin, distaff');
    AddItem( 'ladosomi'  , 'unwind, unravel, spool');
    AddItem( 'ladosofa'  , 'spinning, spinner');
    AddItem( 'ladosola'  , 'canvas, cretonne, chintz, batiste, cambric');
    AddItem( 'ladosoti'  , 'cotton, calico, percale, jaconet');
    AddItem( 'ladolado'  , 'muslin, organdy, tarlatan, gauze, pancake');
    AddItem( 'ladolare'  , 'wool, flannel, merino, cashmere');
    AddItem( 'ladolami'  , 'sheet, draper');
    AddItem( 'ladolafa'  , 'silk, moire.taft');
    AddItem( 'ladolaso'  , 'satin, brocade');
    AddItem( 'ladolati'  , 'velvet, velvety');
    AddItem( 'ladotido'  , 'fabric, cloth, material');
    AddItem( 'ladotire'  , 'damask, damascus');
    AddItem( 'ladotimi'  , 'indian, persian');
    AddItem( 'ladotifa'  , 'approach');
    AddItem( 'ladotiso'  , 'prepare, making, confection');
    AddItem( 'ladotila'  , 'fashion, style, form, way');
    AddItem( 'laredore'  , 'color, hue, colorist, colored, coloring, dye');
    AddItem( 'laredomi'  , 'dark, brown, tanned, brunet');
    AddItem( 'laredofa'  , 'chestnut, blonde, redhead');
    AddItem( 'laredoso'  , 'brightness, white, whitish');
    AddItem( 'laredola'  , 'buy, purchase, acquire, acquisition, buyer, customer');
    AddItem( 'laredoti'  , 'clarify, transparency, clarity, diaphaneity, clear, transparent, translucent');
    AddItem( 'laremido'  , 'grey, gray, greyish, grayish');
    AddItem( 'laremire'  , 'violet, violetish, purple, purplish, lilac');
    AddItem( 'laremifa'  , 'green, greenish, olive');
    AddItem( 'laremiso'  , 'yellow, yellowish, yellowing');
    AddItem( 'laremila'  , 'blue, bluish, azure');
    AddItem( 'laremiti'  , 'reddening, blush, crimson, reddish, scarlet');
    AddItem( 'larefado'  , 'poor, impoverished');
    AddItem( 'larefare'  , 'offer, propose, proposal, offer, motion, offering, offerer, supplier, nominator');
    AddItem( 'larefami'  , 'opportunity, windfall, chance');
    AddItem( 'larefaso'  , 'series, following, filiation');
    AddItem( 'larefala'  , 'sort, kind, genre, manner');
    AddItem( 'larefati'  , 'choose, elect, choice, option');
    AddItem( 'laresodo'  , 'fold, bend, ply');
    AddItem( 'laresore'  , 'wrap, package, encase, enclose, envelope');
    AddItem( 'laresomi'  , 'tie, link, string, belt, cordon, rope');
    AddItem( 'laresofa'  , 'package, bundle, bouquet');
    AddItem( 'laresola'  , 'take, seize, grab, taking, seizure, capture, catch, booty, loot, spoil, prey');
    AddItem( 'laresoti'  , 'take, remove, monopolize, hoard, hoarding, monopolist, abductor');
    AddItem( 'larelado'  , 'accounting, bookkeeping, cashier, teller, counter, box, office');
    AddItem( 'larelare'  , 'count, counting, enumerate, list, account, enumeration, counter');
    AddItem( 'larelami'  , 'calculate, calculation, arithmetic, algebra, calculator, arithmetician, algebraic, arithmetically, algrebraically, pc, laptop, calculator');
    AddItem( 'larelafa'  , 'quantify, number, numbering, figure, numerical, numeral');
    AddItem( 'larelaso'  , 'combine, coordinate, combination, coordination, combined, coordinated');
    AddItem( 'larelati'  , 'nomenclature, list, collection, catalog, nomenclator');
    AddItem( 'laretido'  , 'assess, appraise, esteem, evaluation, estimate, price');
    AddItem( 'laretire'  , 'worth, cost, value, valuable, valuably');
    AddItem( 'laretimi'  , 'sum');
    AddItem( 'laretifa'  , 'pay, settle, payment, payer, payable');
    AddItem( 'laretiso'  , 'bill, invoice, check');
    AddItem( 'laretila'  , 'receipt, acquitted, received');
    AddItem( 'lamidore'  , 'augment, increase, enlarge, magnify, magnification, extra, added, increased, additional');
    AddItem( 'lamidomi'  , 'abound, flowing, plenty, abundance, exuberance, abundant, copious, exuberant, abundantly, copiously, profusely');
    AddItem( 'lamidofa'  , 'fructify, fertilize, fertility, fertile, fertile, productive');
    AddItem( 'lamidoso'  , 'mix, jumble, muddle, scramble, mixed, jumbled, scrambled, chaos, confused, confusion');
    AddItem( 'lamidola'  , 'tread, trample, knead, stomping, pressing, kneading, trampling');
    AddItem( 'lamidoti'  , 'crush, grind, pulverize, crushing, grinding, mashing, pulverization, crushed, ground');
    AddItem( 'lamiredo'  , 'borrowing, loan, borrower');
    AddItem( 'lamiremi'  , 'credit, creditor, lender');
    AddItem( 'lamirefa'  , 'pawnshop, pawnbrokers');
    AddItem( 'lamireso'  , 'trade, barter, swap, exchange, swapping');
    AddItem( 'lamirela'  , 'guarantee, deposit, bail, bond, pledge');
    AddItem( 'lamireti'  , 'delay, procrastinate, adjourn, prolong, postpone, extension, delayed, belated');
    AddItem( 'lamifado'  , 'indecency, immodesty, impurity, licentiousness, obscenity, conscupiescence, lust, indecent, immodest, lewd, imp');
    AddItem( 'lamifare'  , 'search, browse, research, searchers, researchers');
    AddItem( 'lamifami'  , 'verify, authenticate, verification, evaluation, verifier, expert');
    AddItem( 'lamifaso'  , 'inventory, survey');
    AddItem( 'lamifala'  , 'dividend');
    AddItem( 'lamifati'  , 'sort, class, rank, categorize, ranking, classification, classifier');
    AddItem( 'lamisodo'  , 'expire, deadline, expiration, term');
    AddItem( 'lamisore'  , 'cease, discontinue, termination, cessation, discontinuation, discontinued');
    AddItem( 'lamisomi'  , 'liquidate, liquidation, liquidator');
    AddItem( 'lamisofa'  , 'remaining, residue');
    AddItem( 'lamisola'  , 'shrink, tighten, narrow, shrinking');
    AddItem( 'lamisoti'  , 'totality, total, entire, all, the    amount, full, complete, totally');
    AddItem( 'lamilado'  , 'sate, satiate, satisfy, quench, gorge, satiation, fullness, satiety, satiating, satisfying');
    AddItem( 'lamilami'  , 'doubling, double, duplicate, doubly');
    AddItem( 'lamilafa'  , 'commodity, merchandise, goods');
    AddItem( 'lamilaso'  , 'number, quantity, amount, numerous, many, multiplicity');
    AddItem( 'lamilati'  , 'employ, apply, use, utilise, consume, adapt, application, consumer, applicable, adaptable');
    AddItem( 'lamitido'  , 'engage, enlist, engagement, commitment, enrollment, enroller');
    AddItem( 'lamitire'  , 'register, record, recording, recorder');
    AddItem( 'lamitimi'  , 'disobey, disobedience, insubordination, rebellion, disobedient, rebellious, insubordinate, unruly, indirigible');
    AddItem( 'lamitifa'  , 'result, occur, arise, result, consequence, consequently, accordingly, therefore, because');
    AddItem( 'lamitiso'  , 'toll, assessment, report');
    AddItem( 'lamitila'  , 'budget, statistical');
    AddItem( 'lafadomi'  , 'add, addition, additional');
    AddItem( 'lafadofa'  , 'subtract, subtraction');
    AddItem( 'lafadoso'  , 'multiply, multiplier, multiplication, multiple, multipliable');
    AddItem( 'lafadola'  , 'divide, separate, divider, division, section, divisible');
    AddItem( 'lafadoti'  , 'share, distribute, dispense, sharing, distribution, allocation, distributor, dispenser, distributive');
    AddItem( 'lafaredo'  , 'tarif, tax, rate, fixed, unchanging, invariable');
    AddItem( 'lafaremi'  , 'mark, label, tag, price, index');
    AddItem( 'lafarefa'  , 'quarter, fourth, quart');
    AddItem( 'lafareso'  , 'third');
    AddItem( 'lafarela'  , 'half');
    AddItem( 'lafareti'  , 'complete, supplement, supplementary, entire, whole, completely, entirely, fully, thoroughly');
    AddItem( 'lafamido'  , 'decline, shrinking, diminish, lessen, smallness');
    AddItem( 'lafamire'  , 'monetize');
    AddItem( 'lafamifa'  , 'bill, debt');
    AddItem( 'lafamila'  , 'discount, discounted, discounter');
    AddItem( 'lafamiti'  , 'endorse, endorsement, endorser');
    AddItem( 'lafasodo'  , 'currency, money, cash, monetary');
    AddItem( 'lafasore'  , 'gold');
    AddItem( 'lafasomi'  , 'vulgarity, common, banal, trivial, improper, inappropriate, trivially');
    AddItem( 'lafasofa'  , 'silver, argentiferous');
    AddItem( 'lafasola'  , 'franc, dollar, euro');
    AddItem( 'lafasoti'  , 'centime, cent');
    AddItem( 'lafalado'  , 'measure, gauge, measurement, measurable, measurer, gauger');
    AddItem( 'lafalare'  , 'proportion, calibrate, size, proportion, calibration, proportionate, proportional, proportionally');
    AddItem( 'lafalami'  , 'geometry, geometric, geometrical, geometrically');
    AddItem( 'lafalafa'  , 'liter');
    AddItem( 'lafalaso'  , 'hectare');
    AddItem( 'lafalati'  , 'survey, acre, surveyor');
    AddItem( 'lafatido'  , 'balance, counterbalance');
    AddItem( 'lafatire'  , 'weigh, weight, weighing');
    AddItem( 'lafatimi'  , 'gravity, heaviness, heavy, solid, heavily, massively');
    AddItem( 'lafatifa'  , 'gram');
    AddItem( 'lafatiso'  , 'm3');
    AddItem( 'lafatila'  , 'cube, block, cubic, cubical');
    AddItem( 'lasodore'  , 'deca');
    AddItem( 'lasodomi'  , 'hecto');
    AddItem( 'lasodofa'  , 'kilo');
    AddItem( 'lasodoso'  , 'milli');
    AddItem( 'lasodola'  , 'centi');
    AddItem( 'lasodoti'  , 'deci');
    AddItem( 'lasoremi'  , 'grow, size, big, huge, excess');
    AddItem( 'lasorefa'  , 'meter, metric, metre');
    AddItem( 'lasoreso'  , 'disillusioned, disillusion, disenchantment, disappointment, disillusionment, disenchanted, disappointing');
    AddItem( 'lasorela'  , 'render, restore, restitution');
    AddItem( 'lasoreti'  , 'consolidate, solidify, strengthen, strong, solid, durable, solidly');
    AddItem( 'lasomido'  , 'stupefy, stupidity, foolishness, silliness, stupidity, brute, stupid, stultifying, stupefying');
    AddItem( 'lasomire'  , 'sample, coupon, cloth');
    AddItem( 'lasomifa'  , 'place');
    AddItem( 'lasomiso'  , 'bias, skew, oblique, slanting, transverse, awry, askew, aside, across, obliquely');
    AddItem( 'lasomila'  , 'broaden, width, magnitude, extent, scope, scale, wide, full, spacious, widely');
    AddItem( 'lasomiti'  , 'length, long, longitudinal, longitudinally');
    AddItem( 'lasofado'  , 'show, display, showman');
    AddItem( 'lasofare'  , 'expose, flaunt, exhibit, exposition, exhibition, exhibitor, display');
    AddItem( 'lasofami'  , 'match, set');
    AddItem( 'lasofaso'  , 'importance, important, significant, considerable');
    AddItem( 'lasofati'  , 'patent, patented');
    AddItem( 'lasolado'  , 'business, businessman');
    AddItem( 'lasolare'  , 'except, exception, exceptional, exceptionally, except, besides');
    AddItem( 'lasolami'  , 'solvency, solvent');
    AddItem( 'lasolafa'  , 'guarantee, insure, guarantee, insurance');
    AddItem( 'lasolaso'  , 'pack, packaging, packer');
    AddItem( 'lasolati'  , 'export, exportation, exporter');
    AddItem( 'lasotido'  , 'negotiate, parley, negotiation, negotiator');
    AddItem( 'lasotire'  , 'reclaim, reprise, resume, recapture');
    AddItem( 'lasotimi'  , 'doubt, disbelief, incredulous');
    AddItem( 'lasotifa'  , 'return');
    AddItem( 'lasotiso'  , 'equivalent, equal, equivalence');
    AddItem( 'lasotila'  , 'advantage, advantageous, advantageously');
    AddItem( 'latidore'  , 'fashion, desired, run, fashionable, vogue');
    AddItem( 'latidomi'  , 'lingerie');
    AddItem( 'latidofa'  , 'flange, collar');
    AddItem( 'latidoso'  , 'cuffs, sleeves');
    AddItem( 'latidola'  , 'lace');
    AddItem( 'latidoti'  , 'ribbon, cord, band');
    AddItem( 'latiredo'  , 'display, placard, poster, sign');
    AddItem( 'latiremi'  , 'hosiery, hosier');
    AddItem( 'latirefa'  , 'couple, pair');
    AddItem( 'latireso'  , 'knit, knitting');
    AddItem( 'latirela'  , 'net, mesh, network, fishnet');
    AddItem( 'latireti'  , 'crochet');
    AddItem( 'latimido'  , 'cowardice, coward, cowardly');
    AddItem( 'latimire'  , 'amass, accumulate, pile, stack, heap, accumulation, mass');
    AddItem( 'latimifa'  , 'encumber, congest, congestion, hinder, hindrance');
    AddItem( 'latimiso'  , 'haberdashery');
    AddItem( 'latimiti'  , 'skein');
    AddItem( 'latimila'  , 'pi');
    AddItem( 'latifado'  , 'indelicacy, indelicate, incelicately');
    AddItem( 'latifare'  , 'fair, showman');
    AddItem( 'latifami'  , 'toys, toy, rattles');
    AddItem( 'latifaso'  , 'doll, dolls');
    AddItem( 'latifala'  , 'puppets, marionettes, puppet');
    AddItem( 'latifati'  , 'lottery');
    AddItem( 'latisodo'  , 'roam, wander, adventurer, vagabond');
    AddItem( 'latisore'  , 'charlatan');
    AddItem( 'latisomi'  , 'bribe, venal, mercenary');
    AddItem( 'latisofa'  , 'false, forger');
    AddItem( 'latisola'  , 'hinjure, damage, wrong, detriment, disadvantage, detrimental, damaging');
    AddItem( 'latisoti'  , 'bankruptcy, bankrupt');
    AddItem( 'latilado'  , 'doorkeeper, usher');
    AddItem( 'latilare'  , 'assigne, allocate, assignment');
    AddItem( 'latilami'  , 'call, summon, interpellation, question, summons, questioner');
    AddItem( 'latilafa'  , 'protest');
    AddItem( 'latilaso'  , 'trial, case, lawsuit');
    AddItem( 'latilati'  , 'gain, benefit, profit, gainer, beneficiaries, fruitful, profitable, lucrative, fruitfully, successfully');
    AddItem( 'tidoredo'  , 'city, village, town, hamlet');
    AddItem( 'tidoremi'  , 'fence, border, gate, barrier, suburbs');
    AddItem( 'tidorefa'  , 'district');
    AddItem( 'tidoreso'  , 'neighborhood, area, quarter');
    AddItem( 'tidorela'  , 'opacity, opaque');
    AddItem( 'tidoreti'  , 'street, sidewalk, roadway, road, pavement');
    AddItem( 'tidomido'  , 'square');
    AddItem( 'tidomire'  , 'monument, obelisk, column, pyramid');
    AddItem( 'tidomifa'  , 'building, edifice, castle, manor, palace');
    AddItem( 'tidomiso'  , 'prefecture, prefect, sub-prefect');
    AddItem( 'tidomiti'  , 'municipality, council, municipal');
    AddItem( 'tidofado'  , 'convent, monastery, abbey, monastic');
    AddItem( 'tidofare'  , 'hospital, hospice, infirmary');
    AddItem( 'tidofami'  , 'inn, hotel, pension, hostel, innkeeper');
    AddItem( 'tidofaso'  , 'cabaret, nightclub, tavern, restaurant, restaurateur');
    AddItem( 'tidofala'  , 'caf, cafe');
    AddItem( 'tidofati'  , 'market, bazaar, marketplace');
    AddItem( 'tidosodo'  , 'boulevard, avenue, promenade');
    AddItem( 'tidosore'  , 'repay, reimburse, receipt, reimbursement, repayment, refundable');
    AddItem( 'tidosomi'  , 'monotony, uniformity, monotonous, uniform');
    AddItem( 'tidosofa'  , 'alley, avenue, aisle');
    AddItem( 'tidosola'  , 'passage');
    AddItem( 'tidosoti'  , 'corridor, hallway, maze, labyrinth');
    AddItem( 'tidolado'  , 'suburb, suburban');
    AddItem( 'tidolare'  , 'dyke, dam');
    AddItem( 'tidolami'  , 'river, fluvial');
    AddItem( 'tidolafa'  , 'bridge, pontoon');
    AddItem( 'tidolaso'  , 'arch, archway, arc');
    AddItem( 'tidolati'  , 'fountain');
    AddItem( 'tidotido'  , 'centralize, centralization, center, middle, central, centrally');
    AddItem( 'tidotire'  , 'average, mean');
    AddItem( 'tidotimi'  , 'angle, corner, wedge, angular, angularly');
    AddItem( 'tidotifa'  , 'presence, present');
    AddItem( 'tidotiso'  , 'realism, positivism, realist, positivist, realistic');
    AddItem( 'tidotila'  , 'triangle, triangulation, triangular, triangularly');
    AddItem( 'tiredore'  , 'govern, government, governor, governing, governmental, gubment, gubmental');
    AddItem( 'tiredomi'  , 'force, oblige, require, subdue, compel, constrain, constrained, forced, mandatory, coercive, necessarily, forth');
    AddItem( 'tiredofa'  , 'vote, voting, voter');
    AddItem( 'tiredoso'  , 'elect, election, elected, electoral');
    AddItem( 'tiredola'  , 'detail, detailing');
    AddItem( 'tiredoti'  , 'representative, deputy, delegate');
    AddItem( 'tiremido'  , 'deafness, deaf, deafly');
    AddItem( 'tiremire'  , 'dynasty, dynastic');
    AddItem( 'tiremifa'  , 'empire, emperor, empress, imperial, imperially');
    AddItem( 'tiremiso'  , 'royalty, sovereignty, king, queen, sovereign, monarch, dictator, royal');
    AddItem( 'tiremila'  , 'rule, reign, reigning, throne');
    AddItem( 'tiremiti'  , 'courtier');
    AddItem( 'tirefado'  , 'principality, prince, princess, highness, princely');
    AddItem( 'tirefare'  , 'ennoble, nobility, aristocracy, noble, titled, aristocrat, gentleman, aristocratic, aristocracy, aristocrat');
    AddItem( 'tirefami'  , 'modesty, humility, modest, humble, modestly, humbly');
    AddItem( 'tirefaso'  , 'duke, duchess, marquis, marquise, count, countess, viscount, viscountess, baron, baroness');
    AddItem( 'tirefala'  , 'chivalry, knighthood, knight');
    AddItem( 'tirefati'  , 'blazon, escutcheon');
    AddItem( 'tiresodo'  , 'entitle, entitled, titled, title, qualification');
    AddItem( 'tiresore'  , 'sire');
    AddItem( 'tirelado'  , 'senate, senator, peer');
    AddItem( 'tirelafa'  , 'session, meeting, sitting');
    AddItem( 'tirelaso'  , 'consulate, embassy, consul, ambassador, vice-consul');
    AddItem( 'tirelati'  , 'constitute, establish, form, constitution, formation, constituent, constitutional, constitutionally');
    AddItem( 'tiretido'  , 'diplomacy, diplomat, statesman, diplomatic, diplomatically');
    AddItem( 'tiretire'  , 'politics');
    AddItem( 'tiretire'  , 'plottix');
    AddItem( 'tiretimi'  , 'partisanship, sect, partisan, proselyte, sectarian, follower');
    AddItem( 'tiretifa'  , 'imperialism, royalism, monarchy, legitimacy, imperial, royal, monarchist, royalist, monarchical');
    AddItem( 'tiretiso'  , 'liberalism, liberal');
    AddItem( 'tiretila'  , 'republic, republicanism, republican');
    AddItem( 'timidore'  , 'ministry, minister, ministerial, ministerially');
    AddItem( 'timidomi'  , 'bureaucracy, bureaucrat, bureaucratic, bureaucratically');
    AddItem( 'timidofa'  , 'supernumerary, assistant, second-in-command, chief, boss, head, master');
    AddItem( 'timidoso'  , 'secretary');
    AddItem( 'timidola'  , 'preside, presidency, president, vice-president');
    AddItem( 'timidoti'  , 'commission, comittee');
    AddItem( 'timiredo'  , 'adopt, adoption, adoptive, adoptable');
    AddItem( 'timiremi'  , 'indemnify, indemnity, compensate');
    AddItem( 'timirefa'  , 'compensate, compensation');
    AddItem( 'timireso'  , 'reward, recompense, remunerate, remuneration, payment, remunerative, rewarding');
    AddItem( 'timirela'  , 'pension, pensioned');
    AddItem( 'timireti'  , 'retire, retirement, retired');
    AddItem( 'timifado'  , 'manage, administer, direct, administration, management, direction, leadership, administrator, director, manager, ');
    AddItem( 'timifare'  , 'stewardship, superintendence, steward, superintendent');
    AddItem( 'timifami'  , 'moderate, moderation, temperance, sobriety, frugality, temperate, sober, frugal, soberly');
    AddItem( 'timifaso'  , 'manage, govern, management, governance, manager');
    AddItem( 'timifala'  , 'receive, collection, revenue, collector, receiver');
    AddItem( 'timifati'  , 'reward, salary, compensation, pay');
    AddItem( 'timisodo'  , 'darkness, obscurity, dark, shadowy, darkly');
    AddItem( 'timisore'  , 'soil, tarnish, stain, dirt, spot, dirtiness, spotted, messy, dirty, unclean, dirtily');
    AddItem( 'timisomi'  , 'position, location');
    AddItem( 'timisofa'  , 'employment, job, function, responsibility, mission, employed, employee, staff, agent, agency');
    AddItem( 'timisola'  , 'short, shortened, cropped, curtailed');
    AddItem( 'timisoti'  , 'supplement, assistant');
    AddItem( 'timilado'  , 'journalism, journal, gazette, journalist');
    AddItem( 'timilare'  , 'write, writing, expression, writer');
    AddItem( 'timilami'  , 'insert, interpose, insertion, intercalation, intercalary');
    AddItem( 'timilafa'  , 'article, report, statement, expos');
    AddItem( 'timilaso'  , 'news, novelist');
    AddItem( 'timilati'  , 'publish, promulgate, proclaim, publication, promulgation, proclamation, publicist, public, publicly');
    AddItem( 'timitido'  , 'history, chronicle, record, hsitorian, chronicler, annalist, historic, historically');
    AddItem( 'timitimi'  , 'opinion, sentiment, feeling');
    AddItem( 'timitifa'  , 'retrograde, return, recede, decline, degenerate, reflux, decay, degeneration, retrograde, retrospec');
    AddItem( 'timitiso'  , 'contemporaneity, modern, contemporary');
    AddItem( 'timitila'  , 'brotherhood, brother, colleague');
    AddItem( 'tifadore'  , 'dominate, master, enslave, domination, dominator, dominant');
    AddItem( 'tifadomi'  , 'usurp, appropriate, seize, invade, usurpation, invasion, appropriation, usurper, invader');
    AddItem( 'tifadofa'  , 'oppress, tyrannize, oppression, tyranny, absolutism, despotism, autocracy');
    AddItem( 'tifadoso'  , 'inexpressible, indescribable, ineffable');
    AddItem( 'tifadola'  , 'revoke, abolish, annul, repeal, dethrone, revocation, abolishment, annulation, dethronement, retraction');
    AddItem( 'tifadoti'  , 'exile, banish, outlaw, expatriation, exile, banishment, proscription, expatriation, exiled, banished, proscribed');
    AddItem( 'tifaredo'  , 'nationalize, naturalize, nationality, nation, kingdom, state, territory, national, nationally');
    AddItem( 'tifaremi'  , 'fatherland, homeland, motherland, country, compatriot, countryman');
    AddItem( 'tifarefa'  , 'civilize, humanize, civilization, civilizing, civilized');
    AddItem( 'tifareso'  , 'department, departmental');
    AddItem( 'tifarela'  , 'province, provincial');
    AddItem( 'tifareti'  , 'foreigner, foreign, alien');
    AddItem( 'tifamido'  , 'weakening, weakness, debility, feebleness, weak, frail, sickly, thin, puny, feebly');
    AddItem( 'tifamire'  , 'international');
    AddItem( 'tifamifa'  , 'unknown');
    AddItem( 'tifamiso'  , 'bourgeoisie, commoners, beurgeois, common, vulgar');
    AddItem( 'tifamila'  , 'charge, commission, delegation');
    AddItem( 'tifamiti'  , 'proxy, relay');
    AddItem( 'tifasodo'  , 'indicate, designate, indication, designation, indicator, indicative, designative');
    AddItem( 'tifasore'  , 'traverse, cross, traversal, across');
    AddItem( 'tifasomi'  , 'rudeness, discourtesy, incivility, rusticity, boorish, rude, discourteous, uncivil, uncouth, impolite, grossly');
    AddItem( 'tifasofa'  , 'follow, following, next');
    AddItem( 'tifasola'  , 'escort, accompany, escort, procession');
    AddItem( 'tifasoti'  , 'gather, assemble, rally, crowd, multitude, crowds');
    AddItem( 'tifalado'  , 'pass, passing');
    AddItem( 'tifalare'  , 'meet, meeting, consultation');
    AddItem( 'tifalami'  , 'aknowledge, recognizable');
    AddItem( 'tifalafa'  , 'rendezvous');
    AddItem( 'tifalaso'  , 'difference, dissimilarity, disparity, different, dissimilar, diverse, heterogeneous, differently, diversely');
    AddItem( 'tifalati'  , 'club');
    AddItem( 'tifatido'  , 'casino');
    AddItem( 'tifatire'  , 'play, game');
    AddItem( 'tifatimi'  , 'billiards, pool, ball, skittles');
    AddItem( 'tifatifa'  , 'cards');
    AddItem( 'tifatiso'  , 'age, antiquity, antiques, antique, gothic, ancient, anciently, formerly, once');
    AddItem( 'tifatila'  , 'chess, checkers, dominos, lotto, chip, card');
    AddItem( 'tisodore'  , 'scholarship, scholar');
    AddItem( 'tisodomi'  , 'bank, banker');
    AddItem( 'tisodofa'  , 'place, investment');
    AddItem( 'tisodoso'  , 'fail, failure, defeat, disappointment');
    AddItem( 'tisodola'  , 'stockbroker');
    AddItem( 'tisodoti'  , 'exchange');
    AddItem( 'tisoredo'  , 'lower, drop, decrease, low');
    AddItem( 'tisoremi'  , 'annuity, income, revenue, annuitant');
    AddItem( 'tisorefa'  , 'coupons');
    AddItem( 'tisoreso'  , 'shares, shareholder');
    AddItem( 'tisorela'  , 'bonds');
    AddItem( 'tisoreti'  , 'increase, bull, high, top');
    AddItem( 'tisomido'  , 'treasury');
    AddItem( 'tisomire'  , 'finance, financial');
    AddItem( 'tisomifa'  , 'speculate, speculation, speculator, speculative');
    AddItem( 'tisomiso'  , 'capitalize, capital, capitalist');
    AddItem( 'tisomila'  , 'interest');
    AddItem( 'tisomiti'  , 'usury, usurer, stockbroker');
    AddItem( 'tisofado'  , 'lie, falsehood, hypocrisy, liar, false, hypocritical, falsely, hypocritically');
    AddItem( 'tisofare'  , 'disguise, masquerade, mask, carnival');
    AddItem( 'tisofami'  , 'scream, yell, cry, shouting, crier');
    AddItem( 'tisofaso'  , 'stunning, daze, astound, noise, tumult, explosion, hubbub, noisy, tumultuous, deafening, tumultuously');
    AddItem( 'tisofala'  , 'shock, offend, scandal, shocking, offensive, scandalous');
    AddItem( 'tisofati'  , 'abuse, misuse, abuse, excess, abusive, abusively');
    AddItem( 'tisolado'  , 'populate, people, population, people, plebeian, proletarian, deplorable, deplorables');
    AddItem( 'tisolare'  , 'democracy, democrat, democratic');
    AddItem( 'tisolami'  , 'conspire, plot, hatch, intrigue, weave, conspiracy, plot, cabal, conspirator, plotter');
    AddItem( 'tisolafa'  , 'disunity, misunderstanding, trouble, dissension, discord, anarchy, lawlessness, disruptor, anarchist, anarchic');
    AddItem( 'tisolaso'  , 'detune, distort, disagreement, dissonance, dissonant, wrong');
    AddItem( 'tisolati'  , 'insurge, revolutionize, uprising, insurrection, riot, rebellion, revolution, seditious, factious, insurgent, revolt');
    AddItem( 'tisotido'  , 'damage, harm, spoil, plunder, vandalize, destroy, squander, ravage, looting, pillaging, devastation, pillard, devastation');
    AddItem( 'tisotire'  , 'ruin, destitution, ruin, rubble, debris, remnants, remains');
    AddItem( 'tisotimi'  , 'uncertainty, uncertain, unreliable, vague, ambiguous');
    AddItem( 'tisotifa'  , 'destroy, kill, annihilate, slaughter, massacre, murder, destruction, killing, extermination, carnage, destroyer, killer');
    AddItem( 'tisotiso'  , 'demolish, destroy, demolition, reversal');
    AddItem( 'tisotila'  , 'fall, collapse, crumble, capsize, subsidence, landslide, collapse');
    AddItem( 'tiladore'  , 'avoiding, dodge, evade, circumvent, elude, escape, avoidable');
    AddItem( 'tiladomi'  , 'discontent, dissatisfied, discontentment');
    AddItem( 'tiladofa'  , 'flee, escape, desert, fugitives, deserter');
    AddItem( 'tiladoso'  , 'watch, spy, ambush, lookout, watcher, emissary, spy, alert, watchful');
    AddItem( 'tiladola'  , 'catch, trap, snare');
    AddItem( 'tiladoti'  , 'arrest, arresting');
    AddItem( 'tilaredo'  , 'customs');
    AddItem( 'tilarefa'  , 'storehouse, warehouse');
    AddItem( 'tilareso'  , 'monopolize, monopoly, taxation, tax, tribute, contribution, taxpayer, contributory, tributary');
    AddItem( 'tilarela'  , 'forbid, prohibit, interdiction, prohibition, forbidden, illegal, prohibited');
    AddItem( 'tilareti'  , 'poach, defraud, cheat, poaching, smuggling, fraud, poacher, smuggler, fraudulent, fraudulently');
    AddItem( 'tilamido'  , 'police, policeman');
    AddItem( 'tilamifa'  , 'constabulary');
    AddItem( 'tilamila'  , 'guardhouse, post');
    AddItem( 'tilamiti'  , 'monitor, watch, faction, supervisor, guard, sentinel, sentry, patrol, safe, secure');
    AddItem( 'tilafado'  , 'inhumanity, cruelty, barbarism, savagery, inhuman, cruel, barbaric, savage, bloodthirsty, tormentor');
    AddItem( 'tilasola'  , 'import, importation, importer');
    AddItem( 'tilatire'  , 'decorate, decoration');
    AddItem( 'tilatila'  , 'lose, loss, deficit, loser, lost, irremediable, irreparable');
    AddItem( 'dodo'      , 'imperfect');
    AddItem( 'rere'      , 'pluperfect');
    AddItem( 'mimi'      , 'future');
    AddItem( 'fafa'      , 'conditional');
    AddItem( 'soso'      , 'imperative');
    AddItem( 'dodore'    , 'earth, world, globe, land');
    AddItem( 'dodomi'    , 'seasons');
    AddItem( 'dodofa'    , 'winter');
    AddItem( 'dodoso'    , 'spring');
    AddItem( 'dodola'    , 'summer');
    AddItem( 'dodoti'    , 'autumn, fall');
    AddItem( 'dorere'    , 'january');
    AddItem( 'domimi'    , 'february');
    AddItem( 'dofafa'    , 'march');
    AddItem( 'dososo'    , 'april');
    AddItem( 'dolala'    , 'may');
    AddItem( 'dotiti'    , 'june');
    AddItem( 'reredo'    , 'july');
    AddItem( 'reremi'    , 'august');
    AddItem( 'rerefa'    , 'september');
    AddItem( 'rereso'    , 'october');
    AddItem( 'rerela'    , 'november');
    AddItem( 'rereti'    , 'december');
    AddItem( 'redodo'    , '1, one, first, single, unit, only');
    AddItem( 'remimi'    , '2, two, second, secondly');
    AddItem( 'refafa'    , '3, three, third, thirdly, ');
    AddItem( 'resoso'    , '4, four, fourth, fourthly');
    AddItem( 'relala'    , '5, five, fifth, fifthly');
    AddItem( 'retiti'    , '6, six, sixth, sixthly');
    AddItem( 'mimido'    , '7, seven, seventh');
    AddItem( 'mimire'    , '8, eight, eighth, octave');
    AddItem( 'mimifa'    , '9, nine, ninth');
    AddItem( 'mimiso'    , '10, ten, tenth');
    AddItem( 'mimila'    , '11, eleven, eleventh');
    AddItem( 'mimiti'    , '12, twelve, twelfth');
    AddItem( 'midodo'    , '13, thirteen, thirteenth');
    AddItem( 'mirere'    , '14, fourteen, fourteenth');
    AddItem( 'mifafa'    , '15, fifteenth, fifteenth');
    AddItem( 'misoso'    , '16, sixteen, sixteenth');
    AddItem( 'milala'    , '17, seventeen, seventeenth');
    AddItem( 'mititi'    , '18, eighteen, eighteenth');
    AddItem( 'fafado'    , '19, nineteen, nineteenth');
    AddItem( 'fafare'    , '20, twenty, twentieth');
    AddItem( 'fafami'    , '30, thirty, thirtieth');
    AddItem( 'fafaso'    , '40, fourty, fourtieth');
    AddItem( 'fafala'    , '50, fifty, fiftieth');
    AddItem( 'fafati'    , '60, sixty, sixtieth');
    AddItem( 'fadodo'    , '80, eighty, eightieth');
    AddItem( 'farere'    , '100, hundred, hundredth');
    AddItem( 'famimi'    , '100, thousand, thousandth');
    AddItem( 'fasoso'    , '1e6, million, millionth');
    AddItem( 'falala'    , '1e9, billion, billionth');
    AddItem( 'fatiti'    , '1e12, trillion, trillionth');
    AddItem( 'sosodo'    , 'monday');
    AddItem( 'sosore'    , 'tuesday');
    AddItem( 'sosomi'    , 'wednesday');
    AddItem( 'sosofa'    , 'thursday');
    AddItem( 'sosola'    , 'friday');
    AddItem( 'sosoti'    , 'saturday');
    AddItem( 'sododo'    , 'sunday');
    AddItem( 'sorere'    , 'former, past');
    AddItem( 'somimi'    , 'present');
    AddItem( 'sofafa'    , 'future');
    AddItem( 'solala'    , 'hour');
    AddItem( 'sotiti'    , 'minute, second, moment, instant');
    AddItem( 'lalado'    , 'dawn, twilight');
    AddItem( 'lalare'    , 'day, daylight, light');
    AddItem( 'lalami'    , 'meteorology, phenomenon, meteor');
    AddItem( 'lalafa'    , 'morning, mogge');
    AddItem( 'lalaso'    , 'evening, twilight');
    AddItem( 'lalati'    , 'night, nini, gn');
    AddItem( 'ladodo'    , 'temperature, climate');
    AddItem( 'larere'    , 'fog, mist, foggy, misty');
    AddItem( 'lamimi'    , 'humidity, wet, moisten, soak, humid');
    AddItem( 'lafafa'    , 'snow, hail, sleet');
    AddItem( 'lasoso'    , 'freeze, frost, ice, icy');
    AddItem( 'latiti'    , 'cold, chilly, frigid');
    AddItem( 'titido'    , 'rain, shower, downpour, rainy');
    AddItem( 'titire'    , 'wind, breeze, zephyr, gust, hurricane');
    AddItem( 'titimi'    , 'cloud, cloudy');
    AddItem( 'titifa'    , 'lightning, flash');
    AddItem( 'titiso'    , 'thunder, storm, stormy');
    AddItem( 'titila'    , 'heat, warmth, warm, caloric, warmly');
    AddItem( 'tidodo'    , 'atmosphere, air');
    AddItem( 'tirere'    , 'star, planet, comet');
    AddItem( 'timimi'    , 'moon, lunar, !moon');
    AddItem( 'tifafa'    , 'sun, solar');
    AddItem( 'tisoso'    , 'shine, radiate, radiance, radiation');
    AddItem( 'tilala'    , 'sky, firmament, heavenly');
    AddItem( 'dodoredo'  , 'clergy, priesthood, ecclesiastic, priest, abbot, priestly');
    AddItem( 'dodoremi'  , 'vicariate, vicar');
    AddItem( 'dodorefa'  , 'pastor, presbytery, presbyterian');
    AddItem( 'dodoreso'  , 'bishop, episcopal');
    AddItem( 'dodorela'  , 'archbishop, metropolitan');
    AddItem( 'dodoreti'  , 'cardinal');
    AddItem( 'dodomido'  , 'crucifix, cross');
    AddItem( 'dodomire'  , 'redeem, redemption');
    AddItem( 'dodomifa'  , 'resurrect, resurrection');
    AddItem( 'dodomiso'  , 'paradise, eden');
    AddItem( 'dodomila'  , 'angel, cherub, archangel, seraphim, angelic, seraphic');
    AddItem( 'dodomiti'  , 'perfection, completeness, fulfillment, perfect, completed, flawless');
    AddItem( 'dodofado'  , 'papacy, pontificate, pope, holiness, papal, pontifically');
    AddItem( 'dodofare'  , 'consecrate, dedicate, sacred, dedicated, inviolable, inviolably');
    AddItem( 'dodofami'  , 'bless, blessing');
    AddItem( 'dodofaso'  , 'regenerate, regeneration, regenerated, rebirth');
    AddItem( 'dodofala'  , 'apostolate, apostleship, apostle, disciple');
    AddItem( 'dodofati'  , 'mission');
    AddItem( 'dodosodo'  , 'convert, converting, proselytism');
    AddItem( 'dodosore'  , 'loosen, relax, release, relaxation, releasing, categorization');
    AddItem( 'dodosomi'  , 'evangelize, gospel, bible, evangelist, evangelical, evangelically');
    AddItem( 'dodosofa'  , 'catechize, catechism, catechist');
    AddItem( 'dodosola'  , 'christianity, christian, christianly');
    AddItem( 'dodosoti'  , 'dogma, doctrine, dogmatic, dogmatically');
    AddItem( 'dodolado'  , 'theology, theological, theologian, theoglogically');
    AddItem( 'dodolare'  , 'catholicism, catholicity, catholic    , catholic');
    AddItem( 'dodolami'  , 'protestantism, calvinism, lutheranism, protestant, calvinist, lutheran');
    AddItem( 'dodolafa'  , 'muhammedanism, mohammed, mohammedan, muslim');
    AddItem( 'dodolaso'  , 'judaize, judaism, jewish, israelite, judaic');
    AddItem( 'dodolati'  , 'paganism, pagan, gentile, heathen');
    AddItem( 'dodotido'  , 'church, temple, cathedral');
    AddItem( 'dodotire'  , 'steeple');
    AddItem( 'dodotimi'  , 'chapel, chaplain');
    AddItem( 'dodotifa'  , 'altar, sanctuary, tabernacle');
    AddItem( 'dodotiso'  , 'pulpit, platform');
    AddItem( 'dodotila'  , 'preach, preaching, sermon, preacher');
    AddItem( 'doreredo'  , 'diocese, diocesan');
    AddItem( 'doreremi'  , 'parish, parishioner');
    AddItem( 'dorerefa'  , 'churchwarden');
    AddItem( 'dorereso'  , 'sacristy, sacristan');
    AddItem( 'dorerela'  , 'begging, collection, seeker, beggar');
    AddItem( 'dorereti'  , 'contribute, contribution');
    AddItem( 'domimido'  , 'rite, ritual');
    AddItem( 'domimire'  , 'procession');
    AddItem( 'domimifa'  , 'mass');
    AddItem( 'domimiso'  , 'vesper, vespers, compline');
    AddItem( 'domimiti'  , 'angelus');
    AddItem( 'dofafado'  , 'easter');
    AddItem( 'dofafare'  , 'ascension');
    AddItem( 'dofafami'  , 'pentecost');
    AddItem( 'dofafati'  , 'christmas');
    AddItem( 'dososodo'  , 'sacrament, sacramental');
    AddItem( 'dososore'  , 'baptize, baptism, baptismal, baptist');
    AddItem( 'dososomi'  , 'sacrament');
    AddItem( 'dososola'  , 'confirm, confirmation');
    AddItem( 'dolalado'  , 'litany, litanies');
    AddItem( 'dolalare'  , 'psalmody, psalmist, psalm, canticle');
    AddItem( 'dolalami'  , 'plainchant, plainsong');
    AddItem( 'dolalafa'  , 'lectern');
    AddItem( 'dolalaso'  , 'master, choirboy');
    AddItem( 'dolalati'  , 'cense, incense, censer');
    AddItem( 'dotitido'  , 'seminar, seminarian');
    AddItem( 'dotitire'  , 'novitiate, novice');
    AddItem( 'dotitimi'  , 'converted, convert, catechumen, neophyte');
    AddItem( 'dotitifa'  , 'zeal, fervor, zealous, fervent');
    AddItem( 'dotitiso'  , 'rosary, chaplet');
    AddItem( 'dotitila'  , 'breviary, missal');
    AddItem( 'doredodo'  , 'jesuitism, jesuit, jesuitical');
    AddItem( 'doremimi'  , 'congregation, congregational');
    AddItem( 'dorefafa'  , 'canon, canonical');
    AddItem( 'doresoso'  , 'carthusian');
    AddItem( 'dorelala'  , 'monk, capuchin, trappist');
    AddItem( 'doretiti'  , 'beg, begging, mendicity, beggar, mendicant');
    AddItem( 'domidodo'  , 'mortification, maceration, austerity, austere');
    AddItem( 'domirere'  , 'cilice, hairshirt, haircloth, discipline');
    AddItem( 'domifafa'  , 'fast, fasting, faster');
    AddItem( 'domisoso'  , 'jubilee');
    AddItem( 'domilala'  , 'lent');
    AddItem( 'domititi'  , 'superstition, superstitious, superstitiously');
    AddItem( 'dofadodo'  , 'bigotry, hypocrisy, bigot, hypocrite');
    AddItem( 'dofarere'  , 'fanaticism, fanatic');
    AddItem( 'dofamimi'  , 'behind, back, rear, backward');
    AddItem( 'dofasoso'  , 'persecute, persecution, perecutor, persecutory');
    AddItem( 'dofalala'  , 'curse, damn, excommunicate, anathematize, curse, damnation, excommunication, anathema, imprecation');
    AddItem( 'dofatiti'  , 'flog, torment, torture, flagellation, martyrdom, torture');
    AddItem( 'dosododo'  , 'punish, chastise, punishment, chastisement, correctional, corrections');
    AddItem( 'dosorere'  , 'damned, excommunicated, damnable, damnably');
    AddItem( 'dosomimi'  , 'tempt, try, temptation, tempter, trying, tempting');
    AddItem( 'dosofafa'  , 'sin, sinner, sinful');
    AddItem( 'dosolala'  , 'repent, repentance, contrition, remorse, repentant, contrite');
    AddItem( 'dosotiti'  , 'kneel, genuflection, prostration, kneeling, prostrate');
    AddItem( 'doladodo'  , 'admit, admission');
    AddItem( 'dolarere'  , 'confess, confession, penitent');
    AddItem( 'dolamimi'  , 'confessional');
    AddItem( 'dolafafa'  , 'reconcile, mend, repatriate, reconciliation, reconciling');
    AddItem( 'dolasoso'  , 'deliver, save, rescue, deliverer, saviour');
    AddItem( 'dolatiti'  , 'protect, patronize, protection, patronage, protectorate, auspices, protector, patron, patroness');
    AddItem( 'dotidodo'  , 'glorify, glorification');
    AddItem( 'dotirere'  , 'edify, edification, edifying, exemplary, exemplarily');
    AddItem( 'dotimimi'  , 'after, later');
    AddItem( 'dotifafa'  , 'beatify, beatification, beatitude, beatific');
    AddItem( 'dotisoso'  , 'admire, rave, admiration, ecstasy, admirer, fan, ecstatic, amazed, appreciative');
    AddItem( 'dotilala'  , 'respect, revere, venerate, respect, honor, reverence, veneration, respectful, reverential, respectfully, reverently');
    AddItem( 'reredore'  , 'build, construct, construction, architecture, builder, architect, architectural');
    AddItem( 'reredomi'  , 'plan, perspective');
    AddItem( 'reredofa'  , 'scaffold, scaffolding');
    AddItem( 'reredoso'  , 'stilts, stake, picket, pole, cane, spear');
    AddItem( 'reredola'  , 'build, masonry, building, mason');
    AddItem( 'reredoti'  , 'walling, partition');
    AddItem( 'reremido'  , 'plaster, plasterer');
    AddItem( 'reremire'  , 'limestone, quicklime');
    AddItem( 'reremifa'  , 'stone, rock, stoney, rocky');
    AddItem( 'reremiso'  , 'sand, sandblast, sand, sandy');
    AddItem( 'reremila'  , 'asphalt, bitumen, bituminous');
    AddItem( 'reremiti'  , 'petrol, petroleum, oil, naphtha');
    AddItem( 'rerefado'  , 'pave, macadamize, paving, macadam, paver, paved');
    AddItem( 'rerefare'  , 'tile, brick');
    AddItem( 'rerefami'  , 'courtyard, court');
    AddItem( 'rerefaso'  , 'vault, tunnel');
    AddItem( 'rerefala'  , 'post, pillar');
    AddItem( 'rerefati'  , 'fortify, fortification, rampart, parapet');
    AddItem( 'reresodo'  , 'fortress, citadel, fort, bastion');
    AddItem( 'reresore'  , 'barrack, barracks');
    AddItem( 'reresomi'  , 'garrisoned, stationed, garrison');
    AddItem( 'reresofa'  , 'terrace, plateau, estrade, esplanade');
    AddItem( 'reresola'  , 'tower, turret, minaret, donjon, kiosk');
    AddItem( 'reresoti'  , 'crenellate');
    AddItem( 'rerelado'  , 'dome, cupola');
    AddItem( 'rerelare'  , 'top, summit, peak, apex');
    AddItem( 'rerelami'  , 'attic, garret, mansard');
    AddItem( 'rerelafa'  , 'paneling');
    AddItem( 'rerelaso'  , 'roof, roofing');
    AddItem( 'reretido'  , 'spire, pole, bar');
    AddItem( 'reretire'  , 'gutter');
    AddItem( 'reretimi'  , 'trench, ditch, moat, basin, tank');
    AddItem( 'reretifa'  , 'marsh, swamp, wetland, marshy, muddy, boggy');
    AddItem( 'reretiso'  , 'mud, sludge, mire, silt, slime, dung, muddy, miry, slimy');
    AddItem( 'reretila'  , 'sewer, drain, cesspit, sink');
    AddItem( 'redodore'  , 'cave, grotto, cavern, lair, quarry, catacomb');
    AddItem( 'redodomi'  , 'palisade, espalier, fence');
    AddItem( 'redodofa'  , 'arbor, grove, copse');
    AddItem( 'redodoso'  , 'nursery, nurseryman');
    AddItem( 'redodola'  , 'vineyard, winemaker, vintner');
    AddItem( 'redodoti'  , 'greenhouse');
    AddItem( 'remimido'  , 'glaze, glazing, glazier, glass, windows, glassmaker, glass');
    AddItem( 'remimire'  , 'grill, lattice, trellis, grating, mesh, grille');
    AddItem( 'remimifa'  , 'mill, miller');
    AddItem( 'remimiso'  , 'barn, stable, shed, cowshed');
    AddItem( 'remimila'  , 'forage, grazing, fodder, hay, grass, straw, thatch, clover, alfalfa');
    AddItem( 'remimiti'  , 'manure, fertilizer, dung');
    AddItem( 'refafado'  , 'gear, meshed, gears');
    AddItem( 'refafare'  , 'pulley');
    AddItem( 'refafami'  , 'lock, locksmith, padlock');
    AddItem( 'refafaso'  , 'antidote, counter-poison');
    AddItem( 'refafala'  , 'clef, key');
    AddItem( 'refafati'  , 'bolt, hinge');
    AddItem( 'resosodo'  , 'woodwork, woodworker, woodman');
    AddItem( 'resosore'  , 'wood, timber, woodwork');
    AddItem( 'resosomi'  , 'plane, shave, level, flatten, buff, smooth, equalize, even, flattening, polished, sleek');
    AddItem( 'resosofa'  , 'flatten, flattening, flat');
    AddItem( 'resosola'  , 'horizontality, horizon, horizontal, horizontally');
    AddItem( 'resosoti'  , 'verticality, perpendicularity, vertical, perpendicular, vertically, perpendicularly');
    AddItem( 'relalado'  , 'cabinetmaking, cabinetmaker, cabinet');
    AddItem( 'relalare'  , 'inlay, marquetry, inlaid');
    AddItem( 'relalami'  , 'surface, top, superficial, superficially');
    AddItem( 'relalafa'  , 'ebony, rosewood, mahogany, walnut');
    AddItem( 'relalaso'  , 'carpentry, carpenter');
    AddItem( 'relalati'  , 'plank, flooring, board, floor');
    AddItem( 'retitido'  , 'mechanics, machine, device, mechanic, machinist, mechanical, mechanically');
    AddItem( 'retitire'  , 'lever');
    AddItem( 'retitimi'  , 'cylinder, roller, cylindrical');
    AddItem( 'retitifa'  , 'tube, pipe, conduit, piping');
    AddItem( 'retitiso'  , 'bar, block, line');
    AddItem( 'retitila'  , 'rails, tracks');
    AddItem( 'redorere'  , 'tank, reservoir, pool, container');
    AddItem( 'redomimi'  , 'valve');
    AddItem( 'redofafa'  , 'health, disposed, healthy');
    AddItem( 'redososo'  , 'nut');
    AddItem( 'redolala'  , 'screw');
    AddItem( 'redotiti'  , 'injustice, favor, iniquity, unfair, unjust, iniquitous, unfairly, unjustly');
    AddItem( 'remidodo'  , 'turning, lathe, turner');
    AddItem( 'remirere'  , 'rounding, ball, round');
    AddItem( 'remifafa'  , 'encircle, circle, contour, outline, circumference, around, circular, round, circularly');
    AddItem( 'remisoso'  , 'bulgoing, prominence, prominent, salient');
    AddItem( 'remilala'  , 'oval, oblong');
    AddItem( 'remititi'  , 'squaring, squareing, square');
    AddItem( 'refadodo'  , 'profane, desecrate, profanity, desecration, sacrilege, defiler, profane');
    AddItem( 'refarere'  , 'forge, forging, blacksmith');
    AddItem( 'refamimi'  , 'farrier, shoer, shoeing');
    AddItem( 'refasoso'  , 'horseshoe');
    AddItem( 'refalala'  , 'anvil');
    AddItem( 'refatiti'  , 'clamp, vise');
    AddItem( 'resododo'  , 'tighten, press, clench, condense, squeezing, pressure, squeeze, condensation, compactness, density, tight, press');
    AddItem( 'resorere'  , 'pliers');
    AddItem( 'resomimi'  , 'pinch, pinching');
    AddItem( 'resofafa'  , 'hammer, mallet, hammerman');
    AddItem( 'resolala'  , 'push, press, pressure, force, thrust, pushing');
    AddItem( 'resotiti'  , 'sink, sinking');
    AddItem( 'reladodo'  , 'balance, equilibrium, balancer, upright, balanced');
    AddItem( 'relarere'  , 'spring, elasticity, stretch, elastic');
    AddItem( 'relamimi'  , 'side, edge');
    AddItem( 'relafafa'  , 'right');
    AddItem( 'relasoso'  , 'hang, hook, hooked, hanged');
    AddItem( 'relatiti'  , 'hook, stud');
    AddItem( 'retidodo'  , 'equip, tool, equipped, equipment');
    AddItem( 'retirere'  , 'handle, grip');
    AddItem( 'retimimi'  , 'blade');
    AddItem( 'retifafa'  , 'ax, hatchet');
    AddItem( 'retisoso'  , 'saw, sawyer, sawdust');
    AddItem( 'mimidore'  , 'but, although, only');
    AddItem( 'mimidomi'  , 'however, yet, nevertheless');
    AddItem( 'mimidofa'  , 'whereas');
    AddItem( 'mimidoso'  , 'if');
    AddItem( 'mimidola'  , 'while');
    AddItem( 'mimidoti'  , 'till, even, up');
    AddItem( 'mimiredo'  , 'under, beneath, below, underneath');
    AddItem( 'mimiremi'  , 'as, like, alike');
    AddItem( 'mimirefa'  , 'join, junction, adherence, adhesion, joint, adherent, contiguous, joining, connecting, adjoining, beside');
    AddItem( 'mimireso'  , 'between, among, amid, amongst, amidst');
    AddItem( 'mimirela'  , 'through');
    AddItem( 'mimireti'  , 'on, upon, over, above');
    AddItem( 'mimifado'  , 'before');
    AddItem( 'mimifare'  , 'facing, opposite, across');
    AddItem( 'mimifami'  , 'frontal, abreast');
    AddItem( 'mimifaso'  , 'everywhere, throughout, anywhere, wherever');
    AddItem( 'mimifala'  , 'beyond, further, farther');
    AddItem( 'mimifati'  , 'elsewhere, moreover');
    AddItem( 'mimisodo'  , 'where');
    AddItem( 'mimisore'  , 'in, among, amongst');
    AddItem( 'mimisomi'  , 'here, hence');
    AddItem( 'mimisofa'  , 'there, thence');
    AddItem( 'mimisoti'  , 'end, extremity');
    AddItem( 'mimilado'  , 'frequency, frequent, frequently, often');
    AddItem( 'mimilare'  , 'lengthily');
    AddItem( 'mimilami'  , 'before');
    AddItem( 'mimilafa'  , 'after');
    AddItem( 'mimilaso'  , 'occasionally, sometimes');
    AddItem( 'mimitido'  , 'early, earlier');
    AddItem( 'mimitimi'  , 'now');
    AddItem( 'mimitifa'  , 'soon');
    AddItem( 'mimitiso'  , 'immediately, asap');
    AddItem( 'midodoso'  , 'more, further, again, greater, increased, plus');
    AddItem( 'midodola'  , 'exceed, overstep, excess, surplus, excessive, immoderate, unrestrained, immoderately');
    AddItem( 'midodoti'  , 'with');
    AddItem( 'mireredo'  , 'any, anything, everything, whatever');
    AddItem( 'mireremi'  , 'nevermind, nm');
    AddItem( 'mirerefa'  , 'only');
    AddItem( 'mirereso'  , 'gradually, imperceptibly, gradual');
    AddItem( 'mirereti'  , 'good');
    AddItem( 'mifafado'  , 'about, apropos');
    AddItem( 'mifafati'  , 'finally');
    AddItem( 'misosodo'  , 'therefore, thus, consequently, so, hence');
    AddItem( 'misosore'  , 'asperity, harshness, roughness, rough, gnarled, rugged');
    AddItem( 'misosomi'  , 'so');
    AddItem( 'misosola'  , 'mainly, mostly, especially');
    AddItem( 'milalado'  , 'since, because');
    AddItem( 'milalare'  , 'regarding');
    AddItem( 'milalami'  , 'when');
    AddItem( 'milalafa'  , 'some, many, several, plurality, plural');
    AddItem( 'milalaso'  , 'despite');
    AddItem( 'milalati'  , 'although, albeit, while');
    AddItem( 'mititido'  , 'already');
    AddItem( 'mititire'  , 'without');
    AddItem( 'mititimi'  , 'in');
    AddItem( 'mititifa'  , 'following, next');
    AddItem( 'mititiso'  , 'toward, towards');
    AddItem( 'mititila'  , 'since');
    AddItem( 'midorere'  , 'approximation, approximate, approximately, almost, about, virtually');
    AddItem( 'midofafa'  , 'much');
    AddItem( 'midososo'  , 'also, too');
    AddItem( 'midolala'  , 'same');
    AddItem( 'miredodo'  , 'more, also, further');
    AddItem( 'miremimi'  , 'so');
    AddItem( 'mirefafa'  , 'for, because');
    AddItem( 'mirelala'  , 'perhaps, possibly, maybe');
    AddItem( 'miretiti'  , 'moreover, besides');
    AddItem( 'mifarere'  , 'alternate, alternative, alternatviely');
    AddItem( 'mifalala'  , 'anyone, anybody, everyone, whoever');
    AddItem( 'mifatiti'  , 'unrelated');
    AddItem( 'misorere'  , 'early');
    AddItem( 'milarere'  , 'well?');
    AddItem( 'mitidodo'  , 'first, firstly, primarily');
    AddItem( 'mitimimi'  , 'more');
    AddItem( 'mitifafa'  , 'alas!, alas');
    AddItem( 'mitisoso'  , 'extreme');
    AddItem( 'mitilala'  , 'last, finally');
    AddItem( 'fafadore'  , 'pained, hurt, ache, indisposition, illness, malady, ill, indisposed, unwell, sick, sickly');
    AddItem( 'fafadomi'  , 'consult, consultation, consultant, consultative, advisory');
    AddItem( 'fafadofa'  , 'medicine, doctor, medic, medical');
    AddItem( 'fafadoso'  , 'surgery, sergeon, sergical');
    AddItem( 'fafadola'  , 'dentistry, dentist');
    AddItem( 'fafadoti'  , 'oculist');
    AddItem( 'fafaredo'  , 'prescribe, prescription, enjoin');
    AddItem( 'fafaremi'  , 'dose, lot');
    AddItem( 'fafarefa'  , 'pill');
    AddItem( 'fafareso'  , 'pharmacie, drugstore, pharmacist, apothecary, druggist, pharmaceutical');
    AddItem( 'fafarela'  , 'allopathy');
    AddItem( 'fafareti'  , 'homeopathy, homeopath, homeopathic');
    AddItem( 'fafamido'  , 'herbalism, herbalist, herbs');
    AddItem( 'fafamire'  , 'medicate, drug, medicine, medicinal');
    AddItem( 'fafamifa'  , 'remedy, correct, obviate, remediable');
    AddItem( 'fafamiso'  , 'react, reaction, reactionary, reactive');
    AddItem( 'fafamila'  , 'plaster, cauterize, blister');
    AddItem( 'fafamiti'  , 'ointment');
    AddItem( 'fafasore'  , 'lard');
    AddItem( 'fafasomi'  , 'greasing, greaseing, oiling, grease, oil, oiliness, greasy, oily, unctuous');
    AddItem( 'fafasofa'  , 'poultice, cataplasm, compress');
    AddItem( 'fafasola'  , 'flaxseed, linseed');
    AddItem( 'fafasoti'  , 'thicken, clot, coagulate, congeal, curdle, thickening, coagulation, congealing, curdled, thick, coagulate');
    AddItem( 'fafalado'  , 'dilute, diluting');
    AddItem( 'fafalare'  , 'left');
    AddItem( 'fafalami'  , 'cook, bake, cooking, cooked, burn, sting');
    AddItem( 'fafalafa'  , 'boil, bubble, ferment, boiling, fermentation, effervescence, bubbling, effervescent');
    AddItem( 'fafalaso'  , 'melt, dissolve, liquefy, melting, dissolvent, melted, molten, dissolved, liquefied, dissolvable, soluble');
    AddItem( 'fafalati'  , 'aromatize, aromatization, aromatic');
    AddItem( 'fafatido'  , 'evaporate, exhale, vaporize, evaporation, exhalation, volatility, vaporisation, volatilisation, vapor, evaporation');
    AddItem( 'fafatire'  , 'distill, distillation, distiller, distillery');
    AddItem( 'fafatimi'  , 'refine, quintessence, quintessential');
    AddItem( 'fafatifa'  , 'oil');
    AddItem( 'fafatiso'  , 'lemonade');
    AddItem( 'fafatila'  , 'fizz, froth, effervescence, foam, bubbly, fizzy, foamy, frothy');
    AddItem( 'fadodomi'  , 'divert, deter, dissuade');
    AddItem( 'fadodofa'  , 'prevent, stop, impediment, interference, hindering, obstacle, barrier, obstruction');
    AddItem( 'fadodoso'  , 'care, maintenance, careful, carefully');
    AddItem( 'fadodola'  , 'repair, restore, repair, reparation, rapairer, reparable');
    AddItem( 'fadodoti'  , 'caution, delicately');
    AddItem( 'fareredo'  , 'preserve, safeguard, preservative');
    AddItem( 'fareremi'  , 'vaccinate, vaccine');
    AddItem( 'farerefa'  , 'breastfeed, lactate, breastfeeding, lactation');
    AddItem( 'farereso'  , 'regime, diet, plan, scheme, treatment, therapy, treating');
    AddItem( 'farerela'  , 'heal, cure, healing, curable, remedial');
    AddItem( 'farereti'  , 'reborn, rebirth, renewal, renaissance, revival');
    AddItem( 'famimido'  , 'convalescence, convalescent');
    AddItem( 'famimire'  , 'temperament, complexion');
    AddItem( 'famimifa'  , 'body, organism, organs, organic');
    AddItem( 'famimiso'  , 'health, healthy, healthily');
    AddItem( 'famimila'  , 'clean, sanitize, sanitation, hygiene, sanitary, safe, hygenic');
    AddItem( 'famimiti'  , 'effectiveness, effective, wholesome, effectively');
    AddItem( 'fasosodo'  , 'fatten, overweight, fat, chubby, plump, pudgy, rotund');
    AddItem( 'fasosore'  , 'freshness, fresh, new, freshly, newly');
    AddItem( 'fasosomi'  , 'digest, digestion, digestive');
    AddItem( 'fasosofa'  , 'sneeze, sneezing, sternutatory');
    AddItem( 'fasosola'  , 'handkerchief');
    AddItem( 'fasosoti'  , 'yawn, yawning');
    AddItem( 'falalado'  , 'hoarseness, aphonia, hoarse');
    AddItem( 'falalare'  , 'cold, flu, whooping, covid, crona, corona');
    AddItem( 'falalami'  , 'cough, coughing');
    AddItem( 'falalafa'  , 'salivate, drool, spit, expectorate, saliva, sputum, expectoration, expectorant');
    AddItem( 'falalaso'  , 'asthma, asthmatic');
    AddItem( 'falalati'  , 'phthisis, tuberculosis, consumption, hectic, consumptive, tubercular, pulmonary');
    AddItem( 'fatitido'  , 'lukewarm, tepidity, cooling');
    AddItem( 'fatitire'  , 'pleurisy, pleuritic');
    AddItem( 'fatitimi'  , 'laryngitis, bronchitis, larynx, bronchi');
    AddItem( 'fatitifa'  , 'tonsilitis, tonsils, angina');
    AddItem( 'fatitiso'  , 'gland, glandular');
    AddItem( 'fatitila'  , 'goiter, goitrous');
    AddItem( 'fadorere'  , 'infuse, macerate, infusion, maceration');
    AddItem( 'fadomimi'  , 'gargle');
    AddItem( 'fadofafa'  , 'swallow, gulp');
    AddItem( 'fadolala'  , 'syrup, syrupy');
    AddItem( 'fadotiti'  , 'soften, softening, alleviation, mitigation, softener');
    AddItem( 'faredodo'  , 'gum, gummy');
    AddItem( 'faremimi'  , 'disjoin, separate, disjunction');
    AddItem( 'farefafa'  , 'insipid, fading, blandness, insignificance, insipidity, bland, tasteless, insignificant, vapid');
    AddItem( 'faresoso'  , 'pungency, acrid, corrosive');
    AddItem( 'farelala'  , 'harshness, bitterness, harsh, bitter, harshly, bitterly');
    AddItem( 'faretiti'  , 'acidify, acidity, acid');
    AddItem( 'famidodo'  , 'embitter, sourness, acrimony, tart, sour');
    AddItem( 'famirere'  , 'camphor');
    AddItem( 'famifafa'  , 'narcotic, drug, soporific, opium, henbane, belladonna');
    AddItem( 'famisoso'  , 'chloroform');
    AddItem( 'familala'  , 'etherize, etherisation, ether, ethereal');
    AddItem( 'famititi'  , 'deny, negate, negation, denial');
    AddItem( 'fasododo'  , 'slide, slider, sliding');
    AddItem( 'fasorere'  , 'sprain');
    AddItem( 'fasomimi'  , 'twist, strain, contusion, lesion, bruise');
    AddItem( 'fasofafa'  , 'injury, wound, wounded');
    AddItem( 'fasolala'  , 'dress, bandage, plaster');
    AddItem( 'fasotiti'  , 'minority, minor, minimum');
    AddItem( 'faladodo'  , 'operate, operation, operator');
    AddItem( 'falarere'  , 'scar, cicatrize, scar, cicatrix, cicatrizing');
    AddItem( 'falamimi'  , 'bandage, ligature, headband');
    AddItem( 'falafafa'  , 'strap, sling, garter');
    AddItem( 'falasoso'  , 'cripple, maim, disable, handicap, disability, invalid, disabled, crippled');
    AddItem( 'falatiti'  , 'deformity, deformed');
    AddItem( 'fatidodo'  , 'limp, lameness, wobbly, limping');
    AddItem( 'fatimimi'  , 'hunchback, humpback, counterfeit, disguise');
    AddItem( 'fatifafa'  , 'rickets, rachitis, rickety, rachitic');
    AddItem( 'fatisoso'  , 'orthopedics, orthopedist, orthopedic');
    AddItem( 'fatilala'  , 'anatomize, anatomy, anatomist, anatomical, anatomically');
    AddItem( 'sosodore'  , 'joints, articulations, articular');
    AddItem( 'sosodomi'  , 'nerves, nervous');
    AddItem( 'sosodofa'  , 'muscle, fiber, filament, ligament, muscular, brawny, fibrous, filamentous, ligamentous');
    AddItem( 'sosodoso'  , 'bile, humor, bilious');
    AddItem( 'sosodola'  , 'mucus');
    AddItem( 'sosodoti'  , 'lymph, phlegm, lymphatic, phlegmatic');
    AddItem( 'sosoredo'  , 'migraine, headache, neuralgia, neuralgic');
    AddItem( 'sosoremi'  , 'vomit, vomiting, vomitive, emetic');
    AddItem( 'sosorefa'  , 'sweat, transpire, transpiration');
    AddItem( 'sosoreso'  , 'dropsy, edema, hydrops, dropsical, edematous, hydropic');
    AddItem( 'sosorela'  , 'swell, inflate, swelling, inflation, swollen');
    AddItem( 'sosoreti'  , 'tighten, distend, tension, distension');
    AddItem( 'sosomido'  , 'fever, feverish');
    AddItem( 'sosomire'  , 'tremble, shiver, shaking, chills, shivering, trembling');
    AddItem( 'sosomifa'  , 'pulse, pulsation, beat, beating, throbbing');
    AddItem( 'sosomiso'  , 'aneurysm, aneurism');
    AddItem( 'sosomila'  , 'palpitate, palpitation, palpitating');
    AddItem( 'sosomiti'  , 'heart');
    AddItem( 'sosofado'  , 'gout, sciatica, amaurosis');
    AddItem( 'sosofare'  , 'pain, ache, throbbing, twinges');
    AddItem( 'sosofami'  , 'rheumatism, rheumatic');
    AddItem( 'sosofaso'  , 'paralysis, paralytic');
    AddItem( 'sosofala'  , 'catalepsy, cataleptic');
    AddItem( 'sosofati'  , 'epilepsy, lockjaw, tetanus, epileptic, epilepsy');
    AddItem( 'sosolare'  , 'bladder');
    AddItem( 'sosolami'  , 'genitals, urethra, dick');
    AddItem( 'sosolafa'  , 'probe, sensor, prober');
    AddItem( 'sosolaso'  , 'matrix');
    AddItem( 'sosolati'  , 'menses, period, menstrual');
    AddItem( 'sosotido'  , 'periodicity, period, periodical, periodically');
    AddItem( 'sosotire'  , 'bleed, bleeding, hemorrhage');
    AddItem( 'sosotimi'  , 'apoplexy, stroke, apoplectic');
    AddItem( 'sosotifa'  , 'hemorrhoids, hemorrhoidal');
    AddItem( 'sosotiso'  , 'flow, creep, stream, flowed, discharge, loss, flowing, streaming, liquid, liquidity');
    AddItem( 'sosotila'  , 'leucorrhoea, leukorrhea');
    AddItem( 'sododore'  , 'whitlow');
    AddItem( 'sododomi'  , 'lesser, lower, minor, less, least');
    AddItem( 'sododofa'  , 'chilblain, frostbite');
    AddItem( 'sododoso'  , 'chapping, chap, crack');
    AddItem( 'sododola'  , 'prick, sting, pierce, injection, shot, jab, piquant, biting, stinging');
    AddItem( 'sododoti'  , 'scratch, abrasion, rawness');
    AddItem( 'soreredo'  , 'buttons');
    AddItem( 'soreremi'  , 'inoculate, vaccinate, inoculation, vaccination, inoculator');
    AddItem( 'sorerefa'  , 'irreparable, irreparably');
    AddItem( 'sorereso'  , 'smallpox');
    AddItem( 'sorerela'  , 'scabies, mange, ringworm, mangy');
    AddItem( 'sorereti'  , 'leprosy, leprous');
    AddItem( 'somimido'  , 'scab, scabby, scurf');
    AddItem( 'somimire'  , 'itch, pruritus, itching, prurigo');
    AddItem( 'somimifa'  , 'scrape, scraper, scratch');
    AddItem( 'somimiso'  , 'inflame, inflammation, inflammable');
    AddItem( 'somimila'  , 'gangrene, gangrene, gangrenous');
    AddItem( 'somimiti'  , 'decompose, decomposition');
    AddItem( 'sofafado'  , 'tumor, abscess');
    AddItem( 'sofafare'  , 'scrofula, scrofulous');
    AddItem( 'sofafami'  , 'scurvy, scorbutic');
    AddItem( 'sofafaso'  , 'cancer, cancerous');
    AddItem( 'sofafala'  , 'ulcer, fistula, canker, chancroid');
    AddItem( 'sofafati'  , 'inveterate, chronic, rooted');
    AddItem( 'solalado'  , 'inject, syringe, injection, enema, syringe');
    AddItem( 'solalare'  , 'purge, purify, purgation, purgative, cathartic, depurative');
    AddItem( 'solalami'  , 'senna');
    AddItem( 'solalafa'  , 'evacuate, expel, evacuation, expulsion');
    AddItem( 'solalaso'  , 'excrement, poop, poo, diarrhea, diarrheic, excremental');
    AddItem( 'solalati'  , 'urinate, pee, piss, urine, piss, pee');
    AddItem( 'sotitido'  , 'latrines, toilet, loo, restroom, bathroom, wc');
    AddItem( 'sotitire'  , 'sewage, emptier');
    AddItem( 'sotitimi'  , 'infect, stink, infection, stench, reek, fetidness, foul, disgusting, repulsive, apalling, stinking, fetid, rank');
    AddItem( 'sotitifa'  , 'rot, rotten, corrupt, stale');
    AddItem( 'sotitiso'  , 'worm, maggot, verminous');
    AddItem( 'sotitila'  , 'sicken');
    AddItem( 'sodorere'  , 'fluxion, defluxion, inflammation, ophthalmia');
    AddItem( 'sodomimi'  , 'erysipelas');
    AddItem( 'sodofafa'  , 'measles');
    AddItem( 'sodolala'  , 'typhoid');
    AddItem( 'sodotiti'  , 'virus, pox');
    AddItem( 'soredodo'  , 'contagion, contagious');
    AddItem( 'soremimi'  , 'epidemic');
    AddItem( 'sorefafa'  , 'plague, cholera, choleraic');
    AddItem( 'soresoso'  , 'colic, tenesmus, pangs');
    AddItem( 'sorelala'  , 'spasm, convulsion, spastic, convulsive, cramp');
    AddItem( 'soretiti'  , 'poison, poisoning, poisoner, poison, venom, poisonous, venomous');
    AddItem( 'somidodo'  , 'hell, infernal');
    AddItem( 'somirere'  , 'pale, fade, discolor, pallor, discoloration, pale, faded, wasted, pallid');
    AddItem( 'somifafa'  , 'faint, stagger, fainting, staggering, unconscious, unaware');
    AddItem( 'somisoso'  , 'weaken, languish, wither, wilt, overexertion, weakening, languor, etiolation, depletion, languishing');
    AddItem( 'somilala'  , 'slim, thinning, emaciation, thinness, consumption, lean, emaciated, gaunt, bony');
    AddItem( 'somititi'  , 'loiter, mope, drag, dragging, loitering');
    AddItem( 'sofadodo'  , 'asphyxiate, asphyxia');
    AddItem( 'sofarere'  , 'burden, benumb, heaviness, numbness, heavy, burdened');
    AddItem( 'sofamimi'  , 'nowhere');
    AddItem( 'sofasoso'  , 'lethargy, lethargic');
    AddItem( 'sofalala'  , 'danger, peril, pitfall, dangerous, perilous');
    AddItem( 'sofatiti'  , 'victim');
    AddItem( 'soladodo'  , 'choke, strangle, strangulation, strangler, choker');
    AddItem( 'solarere'  , 'oppress, stifle, oppression, suffocation, choking, stifling');
    AddItem( 'solamimi'  , 'crisis');
    AddItem( 'solafafa'  , 'dying');
    AddItem( 'solasoso'  , 'bury, entombment, buryer');
    AddItem( 'solatiti'  , 'shroud');
    AddItem( 'sotidodo'  , 'coffin, bier');
    AddItem( 'sotirere'  , 'funeral, obsequies, funerary');
    AddItem( 'sotimimi'  , 'bury, burial');
    AddItem( 'sotifafa'  , 'tomb, sepulcher, mausoleum, sepulchral');
    AddItem( 'sotisoso'  , 'gravedigger');
    AddItem( 'sotilala'  , 'cemetary');
    AddItem( 'laladore'  , 'tabacco, snuffbox');
    AddItem( 'laladomi'  , 'snuff, snort');
    AddItem( 'laladofa'  , 'smoke, fumes, smoker');
    AddItem( 'laladoso'  , 'cigar');
    AddItem( 'laladola'  , 'pipe');
    AddItem( 'laladoti'  , 'tavern, bar, estaminet');
    AddItem( 'lalaremi'  , 'subscribe, subscription, subscriber');
    AddItem( 'lalarela'  , 'bookstore, bookshop, bookseller');
    AddItem( 'lalareti'  , 'edit, edition, editor');
    AddItem( 'lalamido'  , 'print, press, printing, printer');
    AddItem( 'lalamila'  , 'format, size');
    AddItem( 'lalamiti'  , 'cardboard, box, carton');
    AddItem( 'lalafado'  , 'watchmaking, clock, watch, watchmaker');
    AddItem( 'lalafare'  , 'gild, gilding, gilder');
    AddItem( 'lalafami'  , 'silverplate, silverware, silversmith, silvery');
    AddItem( 'lalafaso'  , 'safety, security, safe, secure');
    AddItem( 'lalafala'  , 'goldsmithing, goldsmith');
    AddItem( 'lalafati'  , 'hardware');
    AddItem( 'lalasodo'  , 'jewelry, jewellery, jewels, jeweler');
    AddItem( 'lalasore'  , 'gather, assemble, group, assembly, collection');
    AddItem( 'lalasomi'  , 'chain, necklace, link');
    AddItem( 'lalasofa'  , 'ring, covenant, band');
    AddItem( 'lalasola'  , 'pearl, pearls');
    AddItem( 'lalasoti'  , 'jet, lignite');
    AddItem( 'lalatido'  , 'gems, ruby, diamond');
    AddItem( 'lalatire'  , 'emerald');
    AddItem( 'lalatimi'  , 'topaz');
    AddItem( 'lalatifa'  , 'garnet, coral, cornelian, coralline');
    AddItem( 'lalatiso'  , 'amethyst');
    AddItem( 'lalatila'  , 'opal');
    AddItem( 'ladodore'  , 'agate');
    AddItem( 'ladodomi'  , 'onyx');
    AddItem( 'ladodofa'  , 'sardonyx');
    AddItem( 'ladodoso'  , 'turquoise');
    AddItem( 'ladodola'  , 'sapphire');
    AddItem( 'lareredo'  , 'metallurgy, metal, metallurgist, metallurgical');
    AddItem( 'lareremi'  , 'iron, ferruginous');
    AddItem( 'larerefa'  , 'incurability, incurable, untreatable, terminal, unhealable');
    AddItem( 'larereso'  , 'steel, marcasite');
    AddItem( 'larerela'  , 'copper');
    AddItem( 'larereti'  , 'bronze, brass, brazen');
    AddItem( 'lamimido'  , 'tin, pewter, tinplate, tinsmith');
    AddItem( 'lamimire'  , 'zinc');
    AddItem( 'lamimifa'  , 'bismuth');
    AddItem( 'lamimiso'  , 'lead');
    AddItem( 'lamimila'  , 'mercury, quicksilver');
    AddItem( 'lamimiti'  , 'platinum');
    AddItem( 'lafafado'  , 'enamel, enameller');
    AddItem( 'lafafare'  , 'faience, earthenware, potter');
    AddItem( 'lafafami'  , 'porcelain');
    AddItem( 'lafafaso'  , 'mosaic');
    AddItem( 'lafafala'  , 'varnish, lacquer');
    AddItem( 'lafafati'  , 'paste, glue, adhesive, sticky, viscous, glutinous');
    AddItem( 'lasosodo'  , 'scale, scaly');
    AddItem( 'lasosore'  , 'nacre, mother-of-pearl');
    AddItem( 'lasosomi'  , 'shell, hull');
    AddItem( 'lasosofa'  , 'horn, antler, horny, horned, buffalo');
    AddItem( 'lasosola'  , 'ivory');
    AddItem( 'lasosoti'  , 'cork');
    AddItem( 'latitido'  , 'curry, tan, currier, tanner');
    AddItem( 'latitire'  , 'leather, coriaceous, leathery');
    AddItem( 'latitimi'  , 'sheepskin, hide');
    AddItem( 'latitifa'  , 'fur, pelt, mink, squirrel, astrakhan, marten');
    AddItem( 'latitiso'  , 'rubber, elastic');
    AddItem( 'latitila'  , 'whale, whalebone');
    AddItem( 'ladorere'  , 'broker');
    AddItem( 'ladomimi'  , 'exploit, exploitation, exploitable');
    AddItem( 'ladososo'  , 'opportunity, opportune, timely');
    AddItem( 'ladolala'  , 'urge, insist, stress, persistance, pressing, insistent');
    AddItem( 'ladotiti'  , 'replace, substitute, replacement, substitution, replacing, substituting');
    AddItem( 'laredodo'  , 'cover, tapestry, upholstery, upholsterer, carpet, rug, mat, canvas');
    AddItem( 'laremimi'  , 'stuff, pad, fill, wad, stuffing, wadding, filling');
    AddItem( 'larefafa'  , 'hanging, curtains, drapery');
    AddItem( 'laresoso'  , 'slipcover, cover');
    AddItem( 'larelala'  , 'fringe');
    AddItem( 'laretiti'  , 'tassel');
    AddItem( 'lamidodo'  , 'lace-making, trimmings, lace-maker');
    AddItem( 'lamirere'  , 'braid, plait, twine, weave, enlace');
    AddItem( 'lamifafa'  , 'rings, curls, ringed, curled');
    AddItem( 'lamisoso'  , 'garland');
    AddItem( 'lamilala'  , 'tie, knot, bow');
    AddItem( 'lamititi'  , 'crossing, crosses, crosswise');
    AddItem( 'lafadodo'  , 'dye, dyeing, dyer');
    AddItem( 'lafamimi'  , 'shimmer, glisten, glitter, gleam, sparkle, variegation, variegated, multi-colored');
    AddItem( 'lafasoso'  , 'felt, felted');
    AddItem( 'lafalala'  , 'plush');
    AddItem( 'lafatiti'  , 'sanding, rubbing');
    AddItem( 'lasododo'  , 'grocery, grocer');
    AddItem( 'lasorere'  , 'allspice, clove, cinnamon, ginger');
    AddItem( 'lasomimi'  , 'honey, honeyed, sweet');
    AddItem( 'lasofafa'  , 'rice');
    AddItem( 'lasolala'  , 'barley, gruel, grits');
    AddItem( 'lasotiti'  , 'flour, cereal, cornstarch');
    AddItem( 'latidodo'  , 'bake, bakery, baker');
    AddItem( 'latirere'  , 'dough, paste, batter, doughy, mushy, pasty');
    AddItem( 'latimimi'  , 'leaven, yeast, ferment, leavening');
    AddItem( 'latifafa'  , 'basketry, basket');
    AddItem( 'latisoso'  , 'filter, sift, filtration, filter, sieve, strainer, screen, colander');
    AddItem( 'latilala'  , 'utensils, kitchenware, cookware, pots, pans, saucepan, cauldron, tub, vat, boiler');
    AddItem( 'titidore'  , 'justicy, equity, fairness, impartiality, fair, equitable, impartial, just, judiciary, legal, justly, fairly, impartially');
    AddItem( 'titidomi'  , 'decree, law, code, legislator, jurist, legislation, legal');
    AddItem( 'titidofa'  , 'jurisprudence, law, legislation, lawyer');
    AddItem( 'titidoso'  , 'magistracy, magistrate');
    AddItem( 'titidola'  , 'chancellor');
    AddItem( 'titimido'  , 'testify, demonstrate, testimony, evidence, demonstration, witness');
    AddItem( 'titimire'  , 'declare, denounce, declaration, statement, denunciation, declarer, denouncer, informer');
    AddItem( 'titimifa'  , 'affirm, attest, certify, affirmation, attestation, certificate, affirmative, certified, authentic, attested');
    AddItem( 'titimiso'  , 'protest, swear, oath, swearing');
    AddItem( 'titimila'  , 'accuse, indict, charge, inculpate, accusation, inculpation, accuser, accusatory');
    AddItem( 'titifado'  , 'lawyer, advocate, counsel');
    AddItem( 'titifare'  , 'plead, pleading, plea');
    AddItem( 'titifaso'  , 'justify, exculpate, exonerate, defend, justification, apologia, defender, supporting');
    AddItem( 'titisodo'  , 'judge');
    AddItem( 'titisore'  , 'jury');
    AddItem( 'titisomi'  , 'judge, judgement, sentence, verdict, judge, arbitrator, referee, jurisdictional');
    AddItem( 'titisofa'  , 'majority, maximu, major');
    AddItem( 'titisola'  , 'absolve, absolution');
    AddItem( 'titisoti'  , 'condemn, condemnation');
    AddItem( 'titilado'  , 'murder, assassinate, kill, assassination, killing, homicide, assassin, murderer, murderous');
    AddItem( 'titilare'  , 'crime, misdemeanor, criminal, evildoer, criminally');
    AddItem( 'titilami'  , 'infamy, turpitude, enormity, atrocity, abomination, villainy, monstrosity, monster, infamous, atrocious, monstrous');
    AddItem( 'titilafa'  , 'imprison, incarcerate, imprisonment, incarceration, captiviy, prisoner, detainee, captive, prison, detention, jail');
    AddItem( 'titilaso'  , 'galley');
    AddItem( 'titilati'  , 'scaffold, gallows, gibbet');
    AddItem( 'tidodomi'  , 'proceed, procedure, process');
    AddItem( 'tidodofa'  , 'fine, finable');
    AddItem( 'tidodola'  , 'registry, registrar');
    AddItem( 'tidodoti'  , 'auctioning, auction');
    AddItem( 'tireredo'  , 'auctioneer');
    AddItem( 'tirereso'  , 'accountant');
    AddItem( 'tirerela'  , 'membership, dependency, belonging');
    AddItem( 'tirereti'  , 'suit, suitable');
    AddItem( 'timimido'  , 'notary, notarial');
    AddItem( 'timimire'  , 'prosecute, prosecution, prosecutor, attorney, representative');
    AddItem( 'timimifa'  , 'mortgage');
    AddItem( 'timimiso'  , 'contract, contractor, contractual');
    AddItem( 'tifafado'  , 'record, folder, file');
    AddItem( 'tifafare'  , 'archives, archivist');
    AddItem( 'tifafami'  , 'attribution, attributive');
    AddItem( 'tifafaso'  , 'control, montitor, registry, controller');
    AddItem( 'tifafala'  , 'expropriate, expropriation');
    AddItem( 'tifafati'  , 'possession, property');
    AddItem( 'tisosodo'  , 'petition, petitioner');
    AddItem( 'tisosore'  , 'claim, demand');
    AddItem( 'tisosomi'  , 'address, memoir, brief');
    AddItem( 'tisosofa'  , 'apply, application, candidate, applicant');
    AddItem( 'tisosola'  , 'annoy, disturb, bother, obsess, irritate, obsession, unwelcome, boring, annoying, tedious, provoke, unbearable');
    AddItem( 'tisosoti'  , 'transfer, transmit, cession, handover, transmission, transferor, transmitted, transferable');
    AddItem( 'tilalado'  , 'bailiff');
    AddItem( 'tilalare'  , 'chamberlain');
    AddItem( 'tilalami'  , 'butler, majordomo');
    AddItem( 'tilalafa'  , 'huntsman');
    AddItem( 'tilalaso'  , 'esquire, page');
    AddItem( 'tiredodo'  , 'cadastre, cadaster');
    AddItem( 'tiremimi'  , 'physics, physicist');
    AddItem( 'tirefafa'  , 'collaboration, cooperation, collaborator, associate, partner');
    AddItem( 'tiresoso'  , 'chemistry, chemist');
    AddItem( 'tirelala'  , 'botany, botanist');
    AddItem( 'tiretiti'  , 'astronomy, astronomer, astronomical');
    AddItem( 'timidodo'  , 'imperfection, imperfect, imperfectly');
    AddItem( 'timirere'  , 'geography, geographer, geographic');
    AddItem( 'timifafa'  , 'globe');
    AddItem( 'timisoso'  , 'colony, tribe, settler, colonist, colonial');
    AddItem( 'timilala'  , 'naturalize, naturalization');
    AddItem( 'timititi'  , 'repress, repression, repressive');
    AddItem( 'tifadodo'  , 'optics, optician, optical');
    AddItem( 'tifarere'  , 'telescope');
    AddItem( 'tifamimi'  , 'thermometer');
    AddItem( 'tifasoso'  , 'barometer');
    AddItem( 'tifalala'  , 'balloon, aerostat, aeronaut');
    AddItem( 'tifatiti'  , 'parachute');
    AddItem( 'tisododo'  , 'arsenal');
    AddItem( 'tisorere'  , 'sword, cutlass, scimitar');
    AddItem( 'tisomimi'  , 'fence, fencing, fencer');
    AddItem( 'tisofafa'  , 'guarding, protect');
    AddItem( 'tisolala'  , 'gymnasium, gym, gymnastics');
    AddItem( 'tisotiti'  , 'mange, cirque, hippodrome, racecourse, racetrack');
    AddItem( 'tiladodo'  , 'equitation, riding, horse    riding, rider, horseman, horsewoman, equestrian');
    AddItem( 'tilarere'  , 'cavalcade');
    AddItem( 'tilamimi'  , 'joust, jouster');
    AddItem( 'tilafafa'  , 'wager, bet, better, gambler');
    AddItem( 'tilasoso'  , 'pyrotechnics, pyrotechnic');
    AddItem( 'tilatiti'  , 'fireworks, firecracker, rocket');
    AddItem( 'doredoredo', 'gazelle');
    AddItem( 'doredoremi', 'roe');
    AddItem( 'doredoreso', 'deer, stag, hart');
    AddItem( 'doredorela', 'hare');
    AddItem( 'doredoreti', 'rabbit');
    AddItem( 'doredomido', 'squirrel');
    AddItem( 'doredomifa', 'marmot');
    AddItem( 'doredomiso', 'mole');
    AddItem( 'doredomila', 'weasel');
    AddItem( 'doredomiti', 'badger');
    AddItem( 'doredofado', 'fox');
    AddItem( 'doredofare', 'wolf');
    AddItem( 'doredofami', 'sable');
    AddItem( 'doredofaso', 'marten');
    AddItem( 'doredofala', 'polecat');
    AddItem( 'doredofati', 'ermine');
    AddItem( 'doredosodo', 'weasel');
    AddItem( 'doredosore', 'rat');
    AddItem( 'doredosomi', 'mouse');
    AddItem( 'doredosofa', 'fieldmouse');
    AddItem( 'doredosola', 'bat');
    AddItem( 'doredosoti', 'taguan');
    AddItem( 'doredolado', 'hedgehog');
    AddItem( 'doredolare', 'porcupine');
    AddItem( 'doredolami', 'caracal');
    AddItem( 'doredolafa', 'lynx');
    AddItem( 'doredolaso', 'jackal');
    AddItem( 'doredolati', 'hyena');
    AddItem( 'doredotido', 'zebra');
    AddItem( 'doredotire', 'chamois');
    AddItem( 'doredotimi', 'bear');
    AddItem( 'doredotifa', 'rhino, rhinoceros');
    AddItem( 'doredotiso', 'giraffe');
    AddItem( 'doredotila', 'elephant');
    AddItem( 'doremidore', 'lion, lioness');
    AddItem( 'doremidomi', 'mane, hair, coat');
    AddItem( 'doremidofa', 'tail');
    AddItem( 'doremidoso', 'tiger, tigress');
    AddItem( 'doremidola', 'panther');
    AddItem( 'doremidoti', 'leopard');
    AddItem( 'doremiredo', 'amphibian, amphibians');
    AddItem( 'doremiremi', 'hippo, hippopotamus');
    AddItem( 'doremirefa', 'beaver');
    AddItem( 'doremireso', 'crocodile');
    AddItem( 'doremirela', 'turtle');
    AddItem( 'doremireti', 'seal');
    AddItem( 'dorefadore', 'cricket');
    AddItem( 'dorefadomi', 'earwig');
    AddItem( 'dorefadofa', 'ant, anthill');
    AddItem( 'dorefadoso', 'fly, midge');
    AddItem( 'dorefadola', 'wasp, hornet, bumblebee, drone');
    AddItem( 'dorefadoti', 'bee');
    AddItem( 'dorefaredo', 'butterfly');
    AddItem( 'dorefaremi', 'chafer, beetle');
    AddItem( 'dorefarefa', 'cicada, dragonfly');
    AddItem( 'dorefareso', 'grasshopper');
    AddItem( 'dorefarela', 'aphid, cochineal');
    AddItem( 'dorefareti', 'spider');
    AddItem( 'dorefamido', 'silkworm');
    AddItem( 'dorefamire', 'caterpillar');
    AddItem( 'dorefamifa', 'lizard');
    AddItem( 'dorefamiso', 'snake, serpent, reptile, asp, viper');
    AddItem( 'dorefamila', 'crawl, creep, crawling');
    AddItem( 'dorefamiti', 'squirm, twist, wiggle, kinking, twisting, twisted, squirming, winding');
    AddItem( 'dorefasodo', 'bite');
    AddItem( 'doresodore', 'lark');
    AddItem( 'doresodomi', 'warbler');
    AddItem( 'doresodofa', 'tarin');
    AddItem( 'doresodoso', 'goldfinch');
    AddItem( 'doresodola', 'canary');
    AddItem( 'doresodoti', 'nightingale');
    AddItem( 'doresoredo', 'chaffinch');
    AddItem( 'doresoremi', 'linnet');
    AddItem( 'doresorefa', 'bullfinch');
    AddItem( 'doresoreso', 'ortolan');
    AddItem( 'doresorela', 'thrush');
    AddItem( 'doresoreti', 'blackbird');
    AddItem( 'doresomido', 'starling');
    AddItem( 'doresomire', 'sparrow');
    AddItem( 'doresomifa', 'wren');
    AddItem( 'doresomiso', 'tit');
    AddItem( 'doresomila', 'wagtail');
    AddItem( 'doresomiti', 'hummingbird');
    AddItem( 'doresofado', 'hummingbird');
    AddItem( 'doresofami', 'swallow');
    AddItem( 'doresofaso', 'swift');
    AddItem( 'doresofala', 'twitter, chirp, warble, twittering');
    AddItem( 'doresofati', 'parrot, parakeet');
    AddItem( 'doresolado', 'rooster, hen');
    AddItem( 'doresolare', 'turkey');
    AddItem( 'doresolami', 'pheasant');
    AddItem( 'doresolaso', 'pigeon');
    AddItem( 'doresolati', 'peacock');
    AddItem( 'doresotido', 'dove');
    AddItem( 'doresotire', 'partridge');
    AddItem( 'doresotimi', 'quail');
    AddItem( 'doresotifa', 'magpie');
    AddItem( 'doresotiso', 'jay');
    AddItem( 'doresotila', 'cuckoo');
    AddItem( 'doreladore', 'goose');
    AddItem( 'doreladomi', 'swan');
    AddItem( 'doreladofa', 'duck');
    AddItem( 'doreladoso', 'stork');
    AddItem( 'doreladola', 'heron');
    AddItem( 'doreladoti', 'crane, ibis');
    AddItem( 'dorelaredo', 'teal');
    AddItem( 'dorelaremi', 'woodcock');
    AddItem( 'dorelarefa', 'pelican');
    AddItem( 'dorelareso', 'gull');
    AddItem( 'dorelarela', 'seagull');
    AddItem( 'dorelareti', 'penguin');
    AddItem( 'dorelamido', 'raven');
    AddItem( 'dorelamire', 'falcon');
    AddItem( 'dorelamifa', 'owl');
    AddItem( 'dorelamiso', 'vulture');
    AddItem( 'dorelamila', 'kite, bird');
    AddItem( 'dorelamiti', 'eagle');
    AddItem( 'dorelafado', 'perch, roost, hutch');
    AddItem( 'dorelafare', 'brood, nest');
    AddItem( 'dorelafami', 'down');
    AddItem( 'dorelafaso', 'feather, feathered, plumage');
    AddItem( 'dorelafala', 'hoopoe, egret');
    AddItem( 'dorelafati', 'paw');
    AddItem( 'dorelasodo', 'male, female');
    AddItem( 'doretidore', 'anchovy');
    AddItem( 'doretidomi', 'smelt');
    AddItem( 'doretidofa', 'sardine');
    AddItem( 'doretidoso', 'herring');
    AddItem( 'doretidola', 'whiting');
    AddItem( 'doretidoti', 'mackerel');
    AddItem( 'doretiredo', 'shrimp');
    AddItem( 'doretiremi', 'crab');
    AddItem( 'doretirefa', 'lobster');
    AddItem( 'doretireso', 'mussel');
    AddItem( 'doretirela', 'oyster');
    AddItem( 'doretireti', 'brill');
    AddItem( 'doretimido', 'sole, flounder, plaice');
    AddItem( 'doretimire', 'ray');
    AddItem( 'doretimifa', 'mullet');
    AddItem( 'doretimiso', 'salmon');
    AddItem( 'doretimila', 'turbot');
    AddItem( 'doretimiti', 'tuna');
    AddItem( 'doretifado', 'dolphin');
    AddItem( 'doretifami', 'porpoise');
    AddItem( 'doretifaso', 'shark');
    AddItem( 'doretifala', 'orca');
    AddItem( 'doretifati', 'whale');
    AddItem( 'doretisodo', 'gudgeon');
    AddItem( 'doretisore', 'tench');
    AddItem( 'doretisomi', 'pike');
    AddItem( 'doretisofa', 'carp');
    AddItem( 'doretisola', 'shad');
    AddItem( 'doretisoti', 'eel');
    AddItem( 'doretilado', 'leech');
    AddItem( 'doretilare', 'crayfish, crawfish');
    AddItem( 'doretilami', 'snail');
    AddItem( 'doretilafa', 'frog');
    AddItem( 'doretilaso', 'toad');
    AddItem( 'domidoredo', 'mushroom, agaric');
    AddItem( 'domidoremi', 'truffle, truffles');
    AddItem( 'domidorefa', 'potatoes');
    AddItem( 'domidoreso', 'pumpkin');
    AddItem( 'domidorela', 'melon, cantaloupe');
    AddItem( 'domidoreti', 'cucumber');
    AddItem( 'domidomido', 'gherkin');
    AddItem( 'domidomire', 'radish');
    AddItem( 'domidomifa', 'turnip');
    AddItem( 'domidomiso', 'carrot');
    AddItem( 'domidomila', 'parsnip');
    AddItem( 'domidomiti', 'beet');
    AddItem( 'domidofado', 'artichoke');
    AddItem( 'domidofare', 'bean');
    AddItem( 'domidofami', 'lentil');
    AddItem( 'domidofaso', 'peas');
    AddItem( 'domidofati', 'asparagus');
    AddItem( 'domidosodo', 'celery');
    AddItem( 'domidosore', 'spinach');
    AddItem( 'domidosomi', 'sorrel');
    AddItem( 'domidosofa', 'chicory');
    AddItem( 'domidosola', 'lettuce');
    AddItem( 'domidolado', 'escarole');
    AddItem( 'domidolare', 'chervil');
    AddItem( 'domidolami', 'parsley');

    // And then some untranslated french is still here

    AddItem( 'domidolafa', 'mache, bourset');
    AddItem( 'domidolaso', 'tarragon');
    AddItem( 'domidolati', 'burnet');
    AddItem( 'domidotido', 'cabbage, kale');
    AddItem( 'domidotire', 'cauliflower');
    AddItem( 'domidotimi', 'salsify');
    AddItem( 'domisodore', 'sureau');
    AddItem( 'domisodomi', 'aulne');
    AddItem( 'domisodofa', 'osier');
    AddItem( 'domisodoso', 'coudrier');
    AddItem( 'domisodola', 'bouleau');
    AddItem( 'domisodoti', 'hetre');
    AddItem( 'domisoredo', 'frne');
    AddItem( 'domisoremi', 'orme, ormeau');
    AddItem( 'domisorefa', 'saule');
    AddItem( 'domisoreso', 'tremble');
    AddItem( 'domisorela', 'peuplier');
    AddItem( 'domisoreti', 'sycomore');
    AddItem( 'domisomido', 'platane');
    AddItem( 'domisomire', 'sapin');
    AddItem( 'domisomifa', 'rsine, rsineux');
    AddItem( 'domisomiso', 'pin');
    AddItem( 'domisomila', 'santale');
    AddItem( 'domisomiti', 'acajou');
    AddItem( 'domisofado', 'nier, bene');
    AddItem( 'domisofare', 'cerisie, cerise');
    AddItem( 'domisofami', 'mrisier, merise');
    AddItem( 'domisofaso', 'murier');
    AddItem( 'domisofala', 'pommier, pomme, reinette, canada');
    AddItem( 'domisofati', 'poirier, poire');
    AddItem( 'domisolado', 'marronnier, chtaignier, marron, chtaigne');
    AddItem( 'domisolami', 'prunier, prune, violette, reiine-claude, mirabelle');
    AddItem( 'domisolafa', 'pcher, pche, brugnon');
    AddItem( 'domisolaso', 'abricotier, abricot');
    AddItem( 'domisolati', 'amandier, amande');
    AddItem( 'domisotido', 'noyer, noix');
    AddItem( 'domisotire', 'noisetier, noisette');
    AddItem( 'domisotimi', 'cormier, corme');
    AddItem( 'domisotifa', 'coignassier, coing');
    AddItem( 'domisotiso', 'oranger, orange');
    AddItem( 'domisotila', 'citronnier, citron');
    AddItem( 'domiladore', 'grenadier, grenade');
    AddItem( 'domiladomi', 'dattier, palmier, latanier, datte, palme');
    AddItem( 'domiladofa', 'ananas');
    AddItem( 'domiladoso', 'cocotier, coco');
    AddItem( 'domiladola', 'bananier, banane');
    AddItem( 'domiladoti', 'jujubier, jujube');
    AddItem( 'domilaredo', 'olivier, olive');
    AddItem( 'domilaremi', 'noyau');
    AddItem( 'domilarefa', 'pistachier, pistache');
    AddItem( 'domilareso', 'trbinthe, trbenthine');
    AddItem( 'domilarela', 'mucilage, mucilagineux');
    AddItem( 'domilareti', 'glu');
    AddItem( 'domilamido', 'groseillier, groseille');
    AddItem( 'domilamire', 'framboisier, framboise');
    AddItem( 'domilamifa', 'fraisier, fraise');
    AddItem( 'domitidore', 'laurier');
    AddItem( 'domitidomi', 'myrthe, buis');
    AddItem( 'domitidofa', 'bruyere');
    AddItem( 'domitidoso', 'gent');
    AddItem( 'domitidola', 'lilas');
    AddItem( 'domitidoti', 'chvrefeuille');
    AddItem( 'domitiredo', 'rosier, ');
    AddItem( 'domitiremi', 'glantier, eglantine');
    AddItem( 'domitirefa', 'dalhia');
    AddItem( 'domitireso', 'jasmin');
    AddItem( 'domitirela', 'narcisse');
    AddItem( 'domitireti', 'lys');
    AddItem( 'domitimido', 'oeillet');
    AddItem( 'domitimire', 'jonquille');
    AddItem( 'domitimifa', 'tulipe');
    AddItem( 'domitimiso', 'tubreuse');
    AddItem( 'domitimila', 'marguerite');
    AddItem( 'domitimiti', 'paquerette');
    AddItem( 'domitifado', 'hyacinthe');
    AddItem( 'domitifare', 'girofle');
    AddItem( 'domitifami', 'iris');
    AddItem( 'domitifaso', 'renoncule');
    AddItem( 'domitifala', 'anemone');
    AddItem( 'domitifati', 'souci');
    AddItem( 'domitisodo', 'violette');
    AddItem( 'domitisore', 'myosotis');
    AddItem( 'domitisomi', 'pense');
    AddItem( 'domitisofa', 'muguet');
    AddItem( 'domitisola', 'perce-neige');
    AddItem( 'domitilado', 'rsda');
    AddItem( 'domitilare', 'mousse');
    AddItem( 'domitilami', 'lierre');
    AddItem( 'domitilafa', 'clmatite');
    AddItem( 'domitilaso', 'verveine');
    AddItem( 'dofadoredo', 'minralogie, minral, minralogiste, minralogique');
    AddItem( 'dofadoremi', 'mine, carriere, mineur');
    AddItem( 'dofadoreso', 'terre, argile, calcair');
    AddItem( 'dofadomiso', 'roches');
    AddItem( 'dofadofado', 'quartz, quartzeux');
    AddItem( 'dofadofami', 'aventurine');
    AddItem( 'dofadofati', 'cornaline, calcdoine');
    AddItem( 'dofadosodo', 'fedspath');
    AddItem( 'dofadosore', 'ptunse');
    AddItem( 'dofadosomi', 'kaolin');
    AddItem( 'dofadosofa', 'talc');
    AddItem( 'dofadosola', 'mica');
    AddItem( 'dofadolado', 'amiante, abeste, asbestos');
    AddItem( 'dofadolare', 'houillere, houille');
    AddItem( 'dofadolami', 'goudronner, goudron');
    AddItem( 'dofadolafa', 'anthracite');
    AddItem( 'dofadotido', 'souffre, souffr');
    AddItem( 'dofadotifa', 'oxigene, oxign');
    AddItem( 'dofadotiso', 'chlore');
    AddItem( 'dofaredore', 'hydrogene, hydrogn');
    AddItem( 'dofaredomi', 'carbone');
    AddItem( 'dofaredofa', 'azote, azot');
    AddItem( 'dofaredoso', 'iode');
    AddItem( 'dofaredola', 'phophore, phosphorique');
    AddItem( 'dofaredoti', 'arsenic');
    AddItem( 'dofaremire', 'potasse');
    AddItem( 'dofaremifa', 'ammoniaque');
    AddItem( 'dofaremiso', 'alumine, alumineux');
    AddItem( 'dofaremila', 'alun');
    AddItem( 'dofaremiti', 'selgemme');
    AddItem( 'dofarefado', 'magnsie');
    AddItem( 'dofaresodo', 'gologie, gologique');
  end;


  function  ParseWord( const aWord: string): TSolReSolSymbols;
  begin
    if   not WordList.TryGetValue( aWord, Result)
    then Result := [ SymEr  , SymBk]
    else Result := Result + [ SymBk];
  end;


  function  ParseEnglish( const aValue: string): TSolReSolSymbols;
  var
    aWords   : TStringList;
    aSymbols : TSolReSolSymbols;
    i        : Integer;
    j        : Integer;
    p        : Integer;
    s        : Integer;
  begin
    Result := [];
    aWords := Explode( aValue, ' ');

    try
      for i := 0 to aWords.Count - 1
      do begin
        aSymbols := ParseWord( aWords[ i]);
        p        := Length( Result  );
        s        := Length( aSymbols);
        SetLength( Result, p + s);

        for j := 0 to s - 1
        do Result[ p + j] := aSymbols[ j];
      end;
    finally
      FreeAndNil( aWords);
    end;
  end;



initialization

{$ifdef TEST_SOLRESOLS}

  InitSolReSols;

  try
    ParseEnglish( 'some words to test');
  finally
    DoneSolReSols;
  end;

{$endif}

finalization

  FreeAndNil( WordList);  // Clean up if such had n ot been done ...

end.

